! srend.F90
! July 9, 2017 : replaces earlier 2014-16 version with MPI

! --------------------------------------------------------------------
! Copyright (c) 2016 Ted Wetherbee
!
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation files 
! (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge,
! publish, distribute, sublicense, and/or sell copies of the Software,
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
! BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
! ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
! CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
! SOFTWARE.
! --------------------------------------------------------------------

! ---- portability notes : storage size ------------------------------
! Srend has only been tested with Gnu and Intel compilers on Intel
! and similar processors.  Storage size is crucial for srend.  It 
! happens that KIND=4 means 4 bytes for Gnu and Intel, and
! KIND=8 means 8 bytes for Gnu and Intel.  However, srend now uses
! integer(int32), integer(int16), and real(real32) to nail down
! what were integer*4, integer*2, and real*4 types using non-standard
! star notation.  This requires USEing ISO_FORTRAN_ENV for the srend
! module.  
!
! Fortran equivalence has been replaced by using TRANSFER and
! derived types.  This replacement is awkward in some places.
! There is a mild performance hit as well as some loss of clarity
! because TRANSFER and "workarounds" are not really equivalent to
! user flexibility from using equivalence.
! --------------------------------------------------------------------

! --------------------------------------------------------------------
! Gnu:
!   gfortran srend.F90 -fopenmp -O3 -c
!   gfortran app.F90 srend.o -fopenmp -O3 -o app ! for example app
!   ./app 
!   !! debugging flags: -fbacktrace -fbounds-check
!   !! vectorization encourage: -O3 -ffast-math -march=native
!   !! Show vectorization:
!      -fopt-info-optimized          ! shows lots of info
!      -ftree-vectorizer-verbose=1   ! 0-6: 0 nothing, 6 everything, 
!                                    ! 1 just loops vectorized
!   !! flag unused variables: -Wunused
!   !! performance (things that usually work well)
!             -Ofast (includes: -O3 + -ffast-math,-fno-protect-params,-fstack-arrays)
! ifort:
!   ifort srend.F90 -qopenmp -O3 -c
!   ifort app.F90 srend.o -qopenmp -O3 -o app ! for example app
!   ./app 
!   !! debugging flags: -traceback -check all
!   !! vectorize, encourage: -O3, and use SIMD directives for 16-pack loops
!   !! show vectorization:   -O3 -openmp -vec-report1 -c
!   !! flag unused variables: -Wunused
!   !! performance (things that usually work well)
!             -Ofast (includes: -ipo, -O3, -no-prec-div, -static, and -xHost)
! KNL/ifort:
!   ifort srend.F90 -xMIC-AVX512 -align array64byte -03 -qopenmp -c
!   ifort app.F90 srend.o -xMIC-AVX512 -align array64byte -03 -qopenmp -o app
!   ! ulimit -s unlimited
!   ! KMP_ directives for placing threads can be valuable, here typical:
!     ulimit -s unlimited ! up stack size, important for openmp
!     unset OMP_NUM_THREADS ! in case OMP_NUM_THREADS was set
!     export KMP_STACKSIZE=512k                      ! 16m recommended, stack per thread
!     export KMP_AFFINITY=compact,granularity=thread ! pack oversub threads consec
!                        =scatter                    ! thread per core first
!     export KMP_BLOCKTIME=infinite                  ! time for thread to wait before sleeping
!     export KMP_LIBRARY=turnaround                  ! default=throughput may be better, depends
!     export KMP_PLACE_THREADS=8C,4T                 ! c=num cores, T=num threads/core; 32 threads here
! System limits to check:
!   ulimit -s unlimited      ! openmp usually needs this
!   export OMP_NUM_THREADS=4 ! sometime needed, not if KMP_PLACE_THREADS is used though
!   export OMP_STACKSIZE=40M ! was needed once     ! stack per thread
!
! MPI shortcuts if SREND_TEST_YIELDING is not ifort's sleepqq(1):
!   go to the "#ifdef SREND_MPI_SHORTCUTS" section at the end of srend.F90
!   copy the C code to file: sleep.c
!   compile: gcc sleep.c -c || icc sleep.c -c
!   link this in with the app compile/link step: sleep.o
! --------------------------------------------------------------------

! ---- Right-handed schemes: any one will do -------------------------
!           ^
!           y                   This is nice because it retains
!           y                   the x-y plane in a familiar position:
!           y                   x is right, y is up, and z is
!           + x x x x x x >     toward you.
!         z
!       z
!     L
!
! There is no special orientation for vectors so long as U x V != 0 
! (cross product of Up vector and View vector is non-zero,
! hence non-colinear).
!
!           ^                        ^           /
!           z  Mathematicians seem   z         y   Geography people
!           z  to like this.         z       y     like this for
!           z                        z     y       long=x, lat=y,
!           + y y y y y y >          z   y         and z=elevation.
!         x                          z y
!       x                            + x x x x x x x >
!     L
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

! ****************** routines within module srend *******************
! ---- render wrappers:
! srend_render_image()  PUBLIC  ! calls srend_render(), calls srend_write_image()
! srend_render_flush()  PUBLIC  ! calls srend_render()
! srend_render_load()   PUBLIC  ! calls srend_render()
! ---- compose routines:
! srend_compose_image() PUBLIC  ! calls srend_compose_base(), calls srend_write_image()
! srend_compose_flush() PUBLIC  ! calls srend_compose_base()
! srend_compose_load()  PUBLIC
!
! srend_render()        PUBLIC  ! base rendering routine, called by render wrappers
! CONTAINS: 
!    srend_heapsort()   PRIVATE ! orders image sampling rays bound info from team of threads
!    srend_set_cotab()  PRIVATE ! interpolates keys to cotab (color table)
!
! srend_compose_base()  PRIVATE ! base compositing routine
!
! srend_write_image()   PRIVATE ! called by srend_render_write() & srend_compose_write()
! srend_pack_16bit()    PRIVATE ! called by srend_render() & srend_compose_base()
! srend_unpack_16bit()  PRIVATE ! called by srend_compose_load()
! ---- utilities:
! srend_image2disk      PUBLIC  ! writes image at mem to disk filename fname
! srend_n()             PUBLIC  ! returns number of loaded rendered planes
! srend_mem_info()      PUBLIC  ! prints out info on CHAD or .ppm image in memory
! srend_memlen()        PUBLIC  ! returns the length in bytes of a CHAD or image (PPM,PAM,BMP)
! srend_CHAD_concat()   PUBLIC  ! concatenates CHAD2 after CHAD1: to a SCHAD
! srend_CHAD_allocatable_concat() PUBLIC  ! same as above, but chad1 is allocatable and is extended as needed
! srend_CHAD_allocatable_clear()  PUBLIC  ! clears chad setting leading 'CHAD' to 'blnk'
! srend_CHAD_nV_set()   PUBLIC  ! changes Bh_nV in CHAD header to passed nV
! srend_reset_cotab()   PUBLIC  ! resets B(nV)%cotab_set to .FALSE., can rebuild cotab for nV
! srend_MPIsend()       PUBLIC  ! send CHAD
! srend_MPIisend()      PUBLIC  ! non-blocking send of a CHAD
! srend_MPItest()       PUBLIC  ! complete srend_MPIisend() of a CHAD
! srend_MPIrecv()       PUBLIC  ! recv a CHAD and load it using srend_compose_load()

! ####################################################################
! #### srend definitions for cpp #####################################

! #### SREND_VECTORIZE : define exactly ONE of these -----------------
!#define SREND_VECTORIZE !gfortran
!#define SREND_VECTORIZE !$OMP SIMD
#define SREND_VECTORIZE !DIR$ SIMD
!#define SREND_VECTORIZE !DEC$ VECTOR ALWAYS

! #### deallocate B(nV)%cube(__)%im()  and im() as use ends ----------
#define SREND_IM_AUTO_DEALLOCATION

! #### threading options for rendering -------------------------------
#define SREND_TEAM_RENDER 1
! 1: for calling with multiple threads in team to render together
! 0: otherwise: threads renders separately, no barriers or group dir.
! --- LEAVE THESE ALONE BELOW for proper setting of openmp directives
#if SREND_TEAM_RENDER==1
#define SREND_SAVE ,SAVE
#else
#define SREND_SAVE
#endif
! --------------------------------------------------------------------

! #### limits --------------------------------------------------------
! ---- MAX_V and MAX_LOAD initial values; will increase if DYNAMIC set
#define SREND_MAX_V 16
#define SREND_MAX_LOAD 1024
! ---- DO NOT CHANGE BHEADER_LEN without changing B_header_type
#define SREND_BHEADER_LEN 128
! ---- Don't use B_DYNAMIC_V if TEAM_RENDER=0 and multiple threads
!#define SREND_B_DYNAMIC_V
! ---- works in all modes
#define SREND_B_DYNAMIC_LOAD

! #### set image background RGB
! black is (/0,0,0/) and default if 24-bit (RGBA not defined below)
! no effect if RGBA 32-bit is defined below
! - - - - - - - - - - - - - - - - - - -- - - - -- - - - - - - - - - --
!#define SREND_IMAGE_BACKGROUND (/1.0_4,1.0_4,1.0_4/)

! #### output image selection : 4 choices-----------------------------
!    PPM  :   ----   ---
!    PAM  :   RGBA   ---
!    BMP24:   ----   BMP
!    BMP32:   RGBA   BMP
!#define SREND_RGBA
!#define SREND_BMP

! #### image header lengths : informational only, DO NOT CHANGE ------
#define SREND_PPM_HEADER_LEN 20
#define SREND_PAM_HEADER_LEN 80
#define SREND_BMP24_HEADER_LEN 54
#define SREND_BMP32_HEADER_LEN 122

! #### stereo common clipping sphere ---------------------------------
#define SREND_STEREO
#define SREND_TOFF
#define SREND_DT_EPS 0.01

! #### timing option  ------------------------------------------------
!#define SREND_TIMING
!#define SREND_FLOPS
!#define SREND_COMPOSE_TIMING

! #### color table: Uncomment precisely one --------------------------
! ---- 1) HARDCODED: values are set within the set_cotab routine 
#define SREND_COTAB_HARDCODED
! +: This must be also defined if SREND_COTAB_HARDCODED
#define SREND_COTAB_KEYS 200
! ---- 2) FILENAME: designed for desktop apps usage
!#define SREND_COTAB_FILENAME 'cotab.txt'
! ---- 3) PASSNAMELIST: CM1 and WRF
!#define SREND_COTAB_PASSNAMELIST
! - - - - - -- --- - - - - -- - - - - - - - - - - -- - --- -- - - - --
! ---- REBUILD_EACH_PASS: can make sense for FILENAME or PASSNAMELIST
!#define SREND_COTAB_REBUILD_EACH_PASS

! #### AMR -----------------------------------------------------------
!#define SREND_WOODWARD_AMR

! #### CHAD compression ---------------------------------------------
#define SREND_16BIT
#define SREND_RLE0

! #### octree sort ---------------------------------------------------
#define SREND_OCTREE_SORT

! #### solid rendering -----------------------------------------------
!#define SREND_SOLID_RENDERING
!#define SREND_EARLY_TERMINATION .999

! #### clipping switches ---------------------------------------------
! if off, values have to be passed but calculations are not done
#define SREND_POINT_RADIUS_CLIP
#define SREND_CLIPPING_PLANE

! #### MPI shortcuts -------------------------------------------------
!#define SREND_MPI_SHORTCUTS
! ---- define this to have a separate communicator for srend MPI
!#define SREND_MPI_COMM_DUP
!
! Toward the end of this srend.F90 source code, there are "sleep"
! options to be selected and setup if the isend-test MPI shortcuts
! are used for non-blocking send.

! #### constants -----------------------------------------------------
! real32 precision is always used by Srend: DO NOT CHANGE 
#define SREND_PI 3.141593_real32

! #### end: srend #define cpp ########################################
! ####################################################################

! ********************************************************************
! ********************************************************************
! *********** module srend w *****************************************
! ********************************************************************
! ********************************************************************
module srend
   USE ISO_FORTRAN_ENV
! ********************************************************************
! **** global variables/structures for Srend *************************
! ********************************************************************

   type B_header_type ! header for metadata, info to tack in front of each result
      SEQUENCE
      character(LEN=4) :: leader = 'CHAD'   !001-004 'CHAD', will be 'P6**' for PPM image
      integer(int32)   :: tlen              !005-008 CHAD/SCHAD length
      integer(int32)   :: nchad = 1         !009-012 num CHADs
      integer(int32)   :: plen              !013-016 data length following header
      integer(int16)   :: bit16 = 0         !017-018 compression flag, 32 to 16 bit real RGBA
      integer(int16)   :: RLE0 = 0          !019-020 compression flag, rul length encoding of zeros
      integer(int32)   :: wh(1:4)           !021-036 plane bounds: w0:w1 h0:h1
      integer(int32)   :: xoff, yoff, zoff  !037-048 offset in Vdim
      integer(int32)   :: XN, YN, ZN        !049-060 data dimensions
      integer(int32)   :: Vdim              !061-064 volume dimension in cells
      integer(int32)   :: W, H              !065-072 image WxH in pixels
      integer(int32)   :: nV                !073-076 view index
      integer(int32)   :: perspective = 0   !077-080 1 on 0 off
      integer(int32)   :: Alpha, Beta       !081-088 horiz & vert span degrees
      real(real32)     :: E(1:3)            !089-100 eye position
      real(real32)     :: V(1:3)            !101-112 view vector
      real(real32)     :: U(1:3)            !113-124 up vector
      integer(int32)   :: header_len = 128  !124-128 length of this header
   end type B_header_type
! data structure used by rendering and compositing routines ----------
   type srend_cube_type ! these are the target arrays for rendering
      real(real32),allocatable             :: im(:,:,:,:)         ! loaded rendering results for compositing
!DIR$ ATTRIBUTES ALIGN:64 :: im
   end type srend_cube_type

   type srend_block_type
      logical                              :: used  =.FALSE.      ! True after first load
      type(B_header_type),allocatable      :: Bh(:)        ! loaded common parameters
!DIR$ ATTRIBUTES ALIGN:64 :: Bh
      type(srend_cube_type),allocatable    :: cube(:)             ! target arrays: B(nV)%cube(:)%im
      integer                              :: n=0                 ! initial value = 0, num blocks
      logical                              :: cotab_set = .FALSE. ! True after initialize, once on first pass
      real(real32),allocatable             :: cotab(:,:)          ! color table cache
!DIR$ ATTRIBUTES ALIGN:64 :: cotab
   end type srend_block_type

! whether to allow dynamic resizing of the B structure when nV > size(B); size(B) = SREND_MAX_V initially
#ifndef SREND_B_DYNAMIC_V
   type(srend_block_type),save             :: B(1:SREND_MAX_V)    ! static
#else
   logical,save                            :: B_used = .FALSE.    ! True after B(:) alloc
   type(srend_block_type),allocatable,save :: B(:)                ! alloc on first pass
   type(srend_block_type),allocatable      :: B_temp(:)           ! temp for resizing B(:), if dynamic
#endif

! whether to allow resizing of target im and Bheader arrays per nV in shared structure B
   integer(int32),save                     :: B_max_LOAD = SREND_MAX_LOAD ! running max if dynamic
#ifdef SREND_B_DYNAMIC_LOAD
   type(B_header_type),allocatable         :: Bh_temp(:)           ! temp for resizing
!DIR$ ATTRIBUTES ALIGN:64 :: Bh_temp
   type(srend_cube_type),allocatable       :: cube_temp(:)                ! temp for resizing
#endif

! ---- MPI shortcuts -------------------------------------------------
#ifdef SREND_MPI_SHORTCUTS
! these are global values.
   logical, save    :: srend_MPI_INIT = .FALSE. ! initial value, check this to set communicator
   integer, save    :: srend_COMM               ! is own if SREND_comm_dup is called, MPI_COMM_WORLD otherwise
#endif

! ---- PRIVATE vs. PUBLIC --------------------------------------------
   PRIVATE                     ! default, hides shared structure and internal subroutines
   PUBLIC :: srend_image2disk  ! writes image at mem to disk fname
   PUBLIC :: srend_reset_cotab ! resets to .FALSE., use a new color table choice for nV 
   PUBLIC :: srend_n           ! returns B(nV)%n from shared structure, loaded results
   PUBLIC :: srend_compose_flush, srend_compose_image, srend_compose_load ! composite
   PUBLIC :: srend_render      ! render base: -1=image, 0=flush, 1=load
   PUBLIC :: srend_render_image, srend_render_flush, srend_render_load    ! render wrappers
   PUBLIC :: srend_mem_info    ! utility to examine memory for CHAD or PPM image
   PUBLIC :: srend_memlen      ! function to return CHAD tlen
   PUBLIC :: srend_CHAD_concat ! concatenate CHADs
   PUBLIC :: srend_CHAD_allocatable_concat ! same as above, but chad1 is allocatable and extended as needed
   PUBLIC :: srend_CHAD_allocatable_clear  ! sets leading 'CHAD' to 'blnk'
   PUBLIC :: srend_CHAD_nV_set ! changes Bh_nV at (73:76) in CHAD header to passed nV
#ifdef SREND_MPI_SHORTCUTS
   PUBLIC :: srend_MPIsend, srend_MPIisend, srend_MPItest, srend_MPIrecv
#endif

! --------- BMP header: 24 bit RGB and 32 bit RGBA; used for creating BMP images
! NOTE: BMP headers are what they are: a leading 2 byte 'BM" followed by image meta data.
!       A structure (now, a derived type) requires SEQUENCE in order to pack these contiguously
!       in the correct order, but that leading 'BM' offsets metadata so that it is not aligned, 
!       and compilers will issue a warning.  A C packed struct would do the same thing here.
! Solution: Add the extra 2-byte offset as below to eliminate compiler warnings, then
!       transfer a BMP_header_type to a character(1:124) variable, then write/copy 
!       character(3:56) to make the 54-byte BMP24 header,
!       and character(3:124) to make the 122-byte BMP32 header.
! Comment: This was formerly done using equivalence: no errors or warnings, straightforward,
!       flexible, clear.  However, equivalence was declared obsolete, and it appears impossible 
!       to restore performance and clarity with suggested replacements.
#ifdef SREND_BMP
   type BMP_header_type
      SEQUENCE
      character(LEN=2) :: offset_2char_o_crap   ! force alignment to avoid warnings
      character(LEN=2) :: ID = 'BM'             ! 1- 2 ='BM'
      integer(int32)   :: bytes                 ! 3- 6 ?=54+3*W*H or 54+4*W*H
      integer(int32)   :: zero4 = 0             ! 7-10 =0000 - reserved
      integer(int32)   :: data_offset = 54      !11-14 =54, size of header in bytes
      integer(int32)   :: DIB_header_bytes = 40 !15-18 =40, DIB header bytes
      integer(int32)   :: W                     !19-22 ?=W, Width of data
      integer(int32)   :: H                     !23-26 ?=H, Height of data
      integer(int16)   :: planes = 1            !27-28 =1
      integer(int16)   :: bits                  !29-30 ?=24 or 32
      integer(int32)   :: compression = 0       !31-34 =0 24-bit, =3 32-bit (none)
      integer(int32)   :: data_bytes            !35-38 ?=3*W*H or 4*W*H
      integer(int32)   :: Wres = 2834           !39-42 =2834, resolution in horz
      integer(int32)   :: Hres = 2834           !43-46 =2834, resolution in vert
      integer(int32)   :: palette = 0           !47-50 =0, none
      integer(int32)   :: colors = 0            !51-54 =0, means all
!#ifdef SREND_RGBA
      integer(int32)   :: mask_RGBA(1:4)        !55-70 0000FF00,00FF0000,FF000000,000000FF
      character(LEN=4) :: color_space = 'BGRs'  !71-74 LCS color space, little endian
      integer(int32)   :: zero12(1:12) = (/0,0,0,0,0,0, 0,0,0,0,0,0/) !75-122 unused
!#endif
   end type BMP_header_type
#endif
! ****************** end: module var declarations common to all ******

   CONTAINS
   
! ******* render wrappers: image,flush, load *************************
   
! ********************************************************************
! **** srend_render_image ********************************************
! ********************************************************************
! ! returns image in mem address passed; mem must be large enough
   subroutine srend_render_image(                                     &
      nV, Vdim_in,XN_in,YN_in,ZN_in, Bd,                              &
      xoff_in,yoff_in,zoff_in, dt_in,                                 &
#ifdef SREND_WOODWARD_AMR
      xoffth,yoffth,zoffth,                                           &
      indxbq,indxbqm,indxbqf,vox,                                     &
#else
      bytedata,                                                       &
#endif
      Alpha_in,Beta_in, W_in,H_in, perspective_in,icotab,             &
#ifdef SREND_COTAB_PASSNAMELIST
      n_alpha_knot, alpha_knot, n_rgb_knot, rgb_knot,                 &
#endif
      E_in, V_in, U_in, EyeRight_in, EyeConvD_in,                     &
      clip0_in, clip1_in,                                             &
      clipx0_in,clipx1_in,clipy0_in,clipy1_in,clipz0_in,clipz1_in,    &
      pt_in, pt_r_in, cp_p_in, cp_n_in,                               & 
#ifdef SREND_SOLID_RENDERING
      Sd, S, Srgb,                                                    &
#endif
      mem, memlen)
      implicit NONE
! ----------- incoming parameters ------------------------------------
      INTEGER,intent(IN)            :: nV, Vdim_in, XN_in,YN_in,ZN_in,Bd,xoff_in,yoff_in,zoff_in
      REAL,intent(IN)               :: dt_in
#ifdef SREND_WOODWARD_AMR
      INTEGER,intent(IN)            :: xoffth,yoffth,zoffth
      INTEGER,intent(IN)            :: indxbq(5,-2:Vdim_in/4+3,-2:Vdim_in/4+3,-2:Vdim_in/4+3)
      INTEGER,intent(IN)            :: indxbqm(5,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6)
      INTEGER,intent(IN)            :: indxbqf(5,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12)
      CHARACTER,intent(IN)          :: vox(:,:)
#else
      CHARACTER,intent(IN)          :: bytedata(1-Bd:XN_in+Bd,1-Bd:YN_in+Bd,1-Bd:ZN_in+Bd)
#endif
      REAL,intent(IN)               :: Alpha_in, Beta_in
      INTEGER,intent(IN)            :: W_in, H_in, perspective_in, icotab
#ifdef SREND_COTAB_PASSNAMELIST
      INTEGER,intent(IN)            :: n_alpha_knot
      CHARACTER(LEN=256),intent(IN)      :: alpha_knot(1:n_alpha_knot)
      INTEGER,intent(IN)            :: n_rgb_knot
      CHARACTER(LEN=256),intent(IN)      :: rgb_knot(1:n_rgb_knot)
#endif
      REAL,intent(IN)               :: E_in(1:3), V_in(1:3), U_in(1:3)
      REAL,intent(IN)               :: EyeRight_in, EyeConvD_in,clip0_in,clip1_in
      REAL,intent(IN)               :: clipx0_in, clipx1_in,clipy0_in, clipy1_in,clipz0_in, clipz1_in
      REAL,intent(IN)               :: pt_in(1:3),pt_r_in,cp_p_in(1:3), cp_n_in(1:3)
#ifdef SREND_SOLID_RENDERING
      INTEGER,intent(IN)            :: Sd(0:1,1:3) ! 0=min:1=max,1=x 2=y 3=z
      CHARACTER,intent(IN)          :: S(Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(1,3))
      CHARACTER,intent(IN)          :: Srgb(1:3,Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(0,3))
#endif
      CHARACTER,intent(INOUT)       :: mem(:)  ! results out
      INTEGER,optional,intent(OUT)  :: memlen  ! optional
      integer                       :: memlen_dum
! end variable declarations ------------------------------------------

      call srend_render(-1,                                                & ! pass -1 for image
         nV, Vdim_in, XN_in,YN_in,ZN_in,Bd,xoff_in,yoff_in,zoff_in, dt_in, &
#ifdef SREND_WOODWARD_AMR
         xoffth,yoffth,zoffth,indxbq,indxbqm,indxbqf,vox(:,:),             &
#else
         bytedata,                                                         &
#endif
         Alpha_in, Beta_in, W_in, H_in, perspective_in, icotab,            &
#ifdef SREND_COTAB_PASSNAMELIST
         n_alpha_knot,alpha_knot,n_rgb_knot,rgb_knot,                      &
#endif
         E_in, V_in, U_in,EyeRight_in, EyeConvD_in,clip0_in,clip1_in,      &
         clipx0_in, clipx1_in,clipy0_in, clipy1_in,clipz0_in, clipz1_in,   &
         pt_in,pt_r_in,cp_p_in, cp_n_in,                                   &
#ifdef SREND_SOLID_RENDERING
         Sd,S,Srgb,                                                        &
#endif
         mem, memlen_dum)
         
         if(present(memlen)) memlen = memlen_dum
         
   end subroutine srend_render_image
! ********************************************************************

! ********************************************************************
! **** srend_render_flush ********************************************
! ********************************************************************
! ! returns image in mem address passed; mem must be large enough
   subroutine srend_render_flush(                                     &
      nV, Vdim_in,XN_in,YN_in,ZN_in, Bd,                              &
      xoff_in,yoff_in,zoff_in, dt_in,                                 &
#ifdef SREND_WOODWARD_AMR
      xoffth,yoffth,zoffth,                                           &
      indxbq,indxbqm,indxbqf,vox,                                     &
#else
      bytedata,                                                       &
#endif
      Alpha_in,Beta_in, W_in,H_in, perspective_in,icotab,             &
#ifdef SREND_COTAB_PASSNAMELIST
      n_alpha_knot, alpha_knot, n_rgb_knot, rgb_knot,                 &
#endif
      E_in, V_in, U_in, EyeRight_in, EyeConvD_in,                     &
      clip0_in, clip1_in,                                             &
      clipx0_in,clipx1_in,clipy0_in,clipy1_in,clipz0_in,clipz1_in,    &
      pt_in, pt_r_in, cp_p_in, cp_n_in,                               & 
#ifdef SREND_SOLID_RENDERING
      Sd, S, Srgb,                                                    &
#endif
      mem, memlen)
      implicit NONE
! ----------- incoming parameters ------------------------------------
      INTEGER,intent(IN)            :: nV, Vdim_in, XN_in,YN_in,ZN_in,Bd,xoff_in,yoff_in,zoff_in
      REAL,intent(IN)               :: dt_in
#ifdef SREND_WOODWARD_AMR
      INTEGER,intent(IN)            :: xoffth,yoffth,zoffth
      INTEGER,intent(IN)            :: indxbq(5,-2:Vdim_in/4+3,-2:Vdim_in/4+3,-2:Vdim_in/4+3)
      INTEGER,intent(IN)            :: indxbqm(5,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6)
      INTEGER,intent(IN)            :: indxbqf(5,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12)
      CHARACTER,intent(IN)          :: vox(:,:)
#else
      CHARACTER,intent(IN)          :: bytedata(1-Bd:XN_in+Bd,1-Bd:YN_in+Bd,1-Bd:ZN_in+Bd)
#endif
      REAL,intent(IN)               :: Alpha_in, Beta_in
      INTEGER,intent(IN)            :: W_in, H_in, perspective_in, icotab
#ifdef SREND_COTAB_PASSNAMELIST
      INTEGER,intent(IN)            :: n_alpha_knot
      CHARACTER(LEN=256),intent(IN)      :: alpha_knot(1:n_alpha_knot)
      INTEGER,intent(IN)            :: n_rgb_knot
      CHARACTER(LEN=256),intent(IN)      :: rgb_knot(1:n_rgb_knot)
#endif
      REAL,intent(IN)               :: E_in(1:3), V_in(1:3), U_in(1:3)
      REAL,intent(IN)               :: EyeRight_in, EyeConvD_in,clip0_in,clip1_in
      REAL,intent(IN)               :: clipx0_in, clipx1_in,clipy0_in, clipy1_in,clipz0_in, clipz1_in
      REAL,intent(IN)               :: pt_in(1:3),pt_r_in,cp_p_in(1:3), cp_n_in(1:3)
#ifdef SREND_SOLID_RENDERING
      INTEGER,intent(IN)            :: Sd(0:1,1:3) ! 0=min:1=max,1=x 2=y 3=z
      CHARACTER,intent(IN)          :: S(Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(1,3))
      CHARACTER,intent(IN)          :: Srgb(1:3,Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(0,3))
#endif

      CHARACTER,intent(INOUT)       :: mem(:) ! results out
      INTEGER,optional,intent(OUT)  :: memlen ! optional
      integer                       :: memlen_dum
! end variable declarations ------------------------------------------

      call srend_render(0,                                                 &
         nV, Vdim_in, XN_in,YN_in,ZN_in,Bd,xoff_in,yoff_in,zoff_in, dt_in, &
#ifdef SREND_WOODWARD_AMR
         xoffth,yoffth,zoffth,indxbq,indxbqm,indxbqf,vox(:,:),             &
#else
         bytedata,                                                         &
#endif
         Alpha_in, Beta_in, W_in, H_in, perspective_in, icotab,            &
#ifdef SREND_COTAB_PASSNAMELIST
         n_alpha_knot,alpha_knot,n_rgb_knot,rgb_knot,                      &
#endif
         E_in, V_in, U_in,EyeRight_in, EyeConvD_in,clip0_in,clip1_in,      &
         clipx0_in, clipx1_in,clipy0_in, clipy1_in,clipz0_in, clipz1_in,   &
         pt_in,pt_r_in,cp_p_in, cp_n_in,                                   &
#ifdef SREND_SOLID_RENDERING
         Sd,S,Srgb,                                                        &
#endif
         mem, memlen_dum)
         
         if(present(memlen)) memlen = memlen_dum
         
   end subroutine srend_render_flush
! ********************************************************************

! ********************************************************************
! **** srend_render_load *********************************************
! ********************************************************************
! ! returns image in mem address passed; mem must be large enough
   subroutine srend_render_load(                                      &
      nV, Vdim_in,XN_in,YN_in,ZN_in, Bd,                              &
      xoff_in,yoff_in,zoff_in, dt_in,                                 &
#ifdef SREND_WOODWARD_AMR
      xoffth,yoffth,zoffth,                                           &
      indxbq,indxbqm,indxbqf,vox,                                     &
#else
      bytedata,                                                       &
#endif
      Alpha_in,Beta_in, W_in,H_in, perspective_in,icotab,             &
#ifdef SREND_COTAB_PASSNAMELIST
      n_alpha_knot, alpha_knot, n_rgb_knot, rgb_knot,                 &
#endif
      E_in, V_in, U_in, EyeRight_in, EyeConvD_in,                     &
      clip0_in, clip1_in,                                             &
      clipx0_in,clipx1_in,clipy0_in,clipy1_in,clipz0_in,clipz1_in,    &
      pt_in, pt_r_in, cp_p_in, cp_n_in,                               & 
#ifdef SREND_SOLID_RENDERING
      Sd, S, Srgb,                                                    &
#endif
      memlen) ! not: mem,memlen), and memlen passed is optional
      implicit NONE
! ----------- incoming parameters ------------------------------------
      INTEGER,intent(IN)            :: nV, Vdim_in, XN_in,YN_in,ZN_in,Bd,xoff_in,yoff_in,zoff_in
      REAL,intent(IN)               :: dt_in
#ifdef SREND_WOODWARD_AMR
      INTEGER,intent(IN)            :: xoffth,yoffth,zoffth
      INTEGER,intent(IN)            :: indxbq(5,-2:Vdim_in/4+3,-2:Vdim_in/4+3,-2:Vdim_in/4+3)
      INTEGER,intent(IN)            :: indxbqm(5,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6)
      INTEGER,intent(IN)            :: indxbqf(5,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12)
      CHARACTER,intent(IN)          :: vox(:,:)
#else
      CHARACTER,intent(IN)          :: bytedata(1-Bd:XN_in+Bd,1-Bd:YN_in+Bd,1-Bd:ZN_in+Bd)
#endif
      REAL,intent(IN)               :: Alpha_in, Beta_in
      INTEGER,intent(IN)            :: W_in, H_in, perspective_in, icotab
#ifdef SREND_COTAB_PASSNAMELIST
      INTEGER,intent(IN)            :: n_alpha_knot
      CHARACTER(LEN=256),intent(IN)      :: alpha_knot(1:n_alpha_knot)
      INTEGER,intent(IN)            :: n_rgb_knot
      CHARACTER(LEN=256),intent(IN)      :: rgb_knot(1:n_rgb_knot)
#endif
      REAL,intent(IN)               :: E_in(1:3), V_in(1:3), U_in(1:3)
      REAL,intent(IN)               :: EyeRight_in, EyeConvD_in,clip0_in,clip1_in
      REAL,intent(IN)               :: clipx0_in, clipx1_in,clipy0_in, clipy1_in,clipz0_in, clipz1_in
      REAL,intent(IN)               :: pt_in(1:3),pt_r_in,cp_p_in(1:3), cp_n_in(1:3)
#ifdef SREND_SOLID_RENDERING
      INTEGER,intent(IN)            :: Sd(0:1,1:3) ! 0=min:1=max,1=x 2=y 3=z
      CHARACTER,intent(IN)          :: S(Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(1,3))
      CHARACTER,intent(IN)          :: Srgb(1:3,Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(0,3))
#endif

      INTEGER,optional,intent(OUT)  :: memlen      ! results out : optional
      
      character,dimension(1:16)     :: mem_dum     ! pass this something to ignore for load=T
      integer                       :: memlen_dum  ! pass this, then check for optional argument memlen
! end variable declarations ------------------------------------------

      call srend_render(1,                                                 &
         nV, Vdim_in, XN_in,YN_in,ZN_in,Bd,xoff_in,yoff_in,zoff_in, dt_in, &
#ifdef SREND_WOODWARD_AMR
         xoffth,yoffth,zoffth,indxbq,indxbqm,indxbqf,vox(:,:),             &
#else
         bytedata,                                                         &
#endif
         Alpha_in, Beta_in, W_in, H_in, perspective_in, icotab,            &
#ifdef SREND_COTAB_PASSNAMELIST
         n_alpha_knot,alpha_knot,n_rgb_knot,rgb_knot,                      &
#endif
         E_in, V_in, U_in,EyeRight_in, EyeConvD_in,clip0_in,clip1_in,      &
         clipx0_in, clipx1_in,clipy0_in, clipy1_in,clipz0_in, clipz1_in,   &
         pt_in,pt_r_in,cp_p_in, cp_n_in,                                   &
#ifdef SREND_SOLID_RENDERING
         Sd,S,Srgb,                                                        &
#endif
         mem_dum, memlen_dum)
         
      if(present(memlen)) memlen = memlen_dum
         
   end subroutine srend_render_load
! ********************************************************************


! ***** compose routines: image, flush, load *************************

! ********************************************************************
! srend_compose_image(): composits all data in B(nV)% then
! returns a PPM image of length memlen at location mem.
! ********************************************************************
   subroutine srend_compose_image(nV, mem, memlen)
      implicit NONE      
      INTEGER,intent(IN)                     :: nV 
      CHARACTER,dimension(:),intent(OUT)     :: mem
      INTEGER,optional,intent(OUT)           :: memlen
      integer                                :: memlen_dum
! --------- end var declarations -------------------------------------
#ifdef SREND_B_DYNAMIC_V
      if(.NOT. B_used) then
         if(present(memlen)) memlen = 0 ! error
         return
      end if
#endif
      if( .NOT. B(nV)%used ) then
         if(present(memlen)) memlen = 0 ! error
         return
      end if

      call srend_compose_base(-1,nV,mem,memlen_dum)
      B(nV)%n     = 0 ! reset
      
      if(present(memlen)) memlen = memlen_dum
      
   end subroutine srend_compose_image
! ********************************************************************


! ********************************************************************
! srend_compose_flush(): composits all data in B(nV)% then
! returns a CHAD=header+data of length memlen at location mem.
! ********************************************************************
   subroutine srend_compose_flush(nV, mem, memlen)
      implicit NONE
      INTEGER,intent(IN)                      :: nV
      CHARACTER, dimension(:),intent(OUT)     :: mem
      INTEGER, optional, intent(OUT)          :: memlen ! length of CHAD, not needed for some ops
      integer                                 :: memlen_dum ! internal use
! --------- end var declarations -------------------------------------
#ifdef SREND_B_DYNAMIC_V
      if(.NOT. B_used) then
         if(present(memlen)) memlen = 0 ! error
         return
      end if
#endif
      if( .NOT. B(nV)%used ) then
         if(present(memlen)) memlen = 0 ! error
         return
      end if

      call srend_compose_base(0,nV,mem,memlen_dum) ! no tiling, an srend_compose()
      B(nV)%n = 0 ! reset
      
      if(present(memlen)) memlen = memlen_dum

   end subroutine srend_compose_flush
! ********************************************************************


! ********************************************************************
! srend_compose_load(): loads a CHAD=header+data of length memlen from
! location mem into the B structure using information in the header.
! ********************************************************************
   subroutine srend_compose_load(mem, memlen)
      implicit NONE      
      CHARACTER,dimension(:),intent(IN)    :: mem ! this is the incoming
      INTEGER, optional, intent(OUT)       :: memlen ! this is the memory storage used after load
      integer                              :: memlen_dum ! internal use
      
      integer(int32)                       :: ip, ipacks
      integer(int32)                       :: nV, ix, ir, ic, im_Len
      integer(int32)                       :: iload, i, j

      real(real32),dimension(1:64)         :: real4_cast256

      type(B_header_type)                  :: Bh
! --------- end var declarations -------------------------------------

! initialize
#ifdef SREND_B_DYNAMIC_V
      if( .NOT. B_used) then
         allocate( B(1:SREND_MAX_V) )
         B_used = .TRUE.
      end if
#endif

      Bh = transfer(mem(1:SREND_BHEADER_LEN),Bh) ! first header with npacks info
      ipacks = Bh%nchad
      
      memlen_dum = 0
      ix = 0 ! index in mem, of data read and used
      do ip=1,ipacks
         Bh = transfer(mem(ix+1:ix+SREND_BHEADER_LEN),Bh) ! always read in yet again the first
         ix = ix + SREND_BHEADER_LEN ! advance index in mem(:)
         nV = Bh%nV ! find out the view index
         
         if(.NOT. B(nV)%used ) then ! make sure it exists
            if(.NOT. B(nV)%used ) then ! make sure it exists
               allocate( B(nV)%Bh(1:B_max_LOAD) ) !once
               allocate( B(nV)%cube(1:B_max_LOAD) )             !once
               B(nV)%used  = .TRUE.
               ! n=0 on creation
            end if
         end if
#ifdef SREND_B_DYNAMIC_V       
         i = size(B)
         if(nV > i) then
            j = max(nV,2*i)
            allocate(B_temp(1:j))
            B_temp(1:i) = B
            call move_alloc(B_temp,B) ! from B_temp to B
         end if
#endif
         B(nV)%n = B(nV)%n + 1
         iload = B(nV)%n
#ifdef SREND_B_DYNAMIC_LOAD
         i = size(B(nV)%cube)
         if(B(nV)%n > i) then
            j = max(B(nV)%n,2*i,B_max_LOAD)
            allocate(Bh_temp(1:j))
            allocate(cube_temp(1:j))
            Bh_temp(1:i) = B(nV)%Bh
            cube_temp(1:i) = B(nV)%cube
            B_max_LOAD = j
            call move_alloc(Bh_temp,B(nV)%Bh)
            call move_alloc(cube_temp,B(nV)%cube)
         end if
#endif
         B(nV)%Bh(iload) = Bh ! copy in
         memlen_dum = memlen_dum + SREND_BHEADER_LEN
         if( Bh%wh(2) - Bh%wh(1) >= 0) then ! if not, then nothing but the header
            if(allocated(B(nV)%cube(iload)%im)) deallocate(B(nV)%cube(iload)%im)
            allocate(B(nV)%cube(iload)%im(1:16,0:3,Bh%wh(1):Bh%wh(2),Bh%wh(3):Bh%wh(4)) )
            if( Bh%bit16 == 1) then
               im_len = Bh%plen ! to pass in
               call srend_unpack_16bit(Bh%RLE0,B(nV)%cube(iload)%im,             &
                                            Bh%wh(1),Bh%wh(2),Bh%wh(3),Bh%wh(4), & 
                                            mem(ix+1:),im_len)
               ix = ix + Bh%plen
               memlen_dum = memlen_dum + sizeof(B(nV)%cube(iload)%im)
            else
               do ir=Bh%wh(3),Bh%wh(4)
               do ic=Bh%wh(1),Bh%wh(2)
                  real4_cast256 = transfer(mem(ix+1:ix+256),real4_cast256)
                  do i=1,16
                  do j=0,3
                     B(nV)%cube(iload)%im(:,:,ic,ir) = real4_cast256(i+16*j)
                  end do
                  end do
                  ix = ix+256 ! memlen_dum
               end do
               end do
               memlen_dum = memlen_dum + sizeof(B(nV)%cube(iload)%im)
            end if
         end if
      end do
      
      if(present(memlen)) memlen = memlen_dum ! pass back storage size

   end subroutine srend_compose_load
! ********************************************************************
      
      

      
! rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
! rrrrrrr render rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
! rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
! the leading argument is SrendType:
! -1: image
!  0: flush
!  1: load
! ********************************************************************
   recursive subroutine srend_render( srendType,                      &
      nV, Vdim_in,XN_in,YN_in,ZN_in, Bd,                              &
      xoff_in,yoff_in,zoff_in, dt_in,                                 &
#ifdef SREND_WOODWARD_AMR
      xoffth,yoffth,zoffth,                                           &
      indxbq,indxbqm,indxbqf,vox,                                     &
#else
      bytedata,                                                       &
#endif
      Alpha_in,Beta_in, W_in,H_in, perspective_in,icotab,             &
#ifdef SREND_COTAB_PASSNAMELIST
      n_alpha_knot, alpha_knot, n_rgb_knot, rgb_knot,                 &
#endif
      E_in, V_in, U_in, EyeRight_in, EyeConvD_in,                     &
      clip0_in, clip1_in,                                             &
      clipx0_in,clipx1_in,clipy0_in,clipy1_in,clipz0_in,clipz1_in,    &
      pt_in, pt_r_in, cp_p_in, cp_n_in,                               & 
#ifdef SREND_SOLID_RENDERING
      Sd, S, Srgb,                                                    &
#endif
      mem, memlen)

#ifdef _OPENMP
      USE OMP_LIB
#endif
      implicit NONE
! ----------- incoming parameters ------------------------------------
! The "_in" suffix parameters are always copied to variable
! of the same name if values are changed
! operation: -1 image, 0 flush, 1 load -------------------------------
      INTEGER,intent(IN)            :: srendType
! ----- view parameters - - - - ---- - - - - --- - - - - - - - - - - -
      INTEGER,intent(IN)            :: nV
      INTEGER,intent(IN)            :: Vdim_in
! ----- data volume parameters - - - - - - - - - - - - - - - - - - - -
      INTEGER,intent(IN)            :: XN_in,YN_in,ZN_in
      INTEGER,intent(IN)            :: Bd ! boundary around incoming data
      INTEGER,intent(IN)            :: xoff_in,yoff_in,zoff_in ! offsets for cube
      REAL,intent(IN)               :: dt_in
      real(real32) SREND_SAVE       :: dt ! internal
#ifdef SREND_WOODWARD_AMR
      INTEGER,intent(IN)            :: xoffth,yoffth,zoffth ! offsets within indxbq vox
      INTEGER,intent(IN)            :: indxbq(5,-2:Vdim_in/4+3,-2:Vdim_in/4+3,-2:Vdim_in/4+3)
      INTEGER,intent(IN)            :: indxbqm(5,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6,-5:2*Vdim_in/4+6)
      INTEGER,intent(IN)            :: indxbqf(5,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12,-11:4*Vdim_in/4+12)
      CHARACTER,intent(IN)          :: vox(:,:)
#else
      CHARACTER,intent(IN)          :: bytedata(1-Bd:XN_in+Bd,1-Bd:YN_in+Bd,1-Bd:ZN_in+Bd)
#endif

! ----- image parameters - - - - - - - - - - - - - - - - - - - - - - -
      REAL,intent(IN)               :: Alpha_in, Beta_in
      INTEGER,intent(IN)            :: W_in, H_in
      INTEGER,intent(IN)            :: perspective_in
      INTEGER,intent(IN)            :: icotab
#ifdef SREND_COTAB_PASSNAMELIST
      INTEGER,intent(IN)            :: n_alpha_knot
      CHARACTER(LEN=256),intent(IN)      :: alpha_knot(1:n_alpha_knot)
      INTEGER,intent(IN)            :: n_rgb_knot
      CHARACTER(LEN=256),intent(IN)      :: rgb_knot(1:n_rgb_knot)
#endif

! view parameters
      REAL,intent(IN)               :: E_in(1:3)
      REAL, intent(IN)              :: V_in(1:3)
      REAL,intent(IN)               :: U_in(1:3)
      REAL,intent(IN)               :: EyeRight_in, EyeConvD_in
! clipping parameters
      REAL,intent(IN)               :: clip0_in
      real(real32) SREND_SAVE                  :: clip0 ! internal
      REAL,intent(IN)               :: clip1_in
      real(real32) SREND_SAVE                   :: clip1 ! internal
      REAL,intent(IN)               :: clipx0_in, clipx1_in
      real(real32) SREND_SAVE                   :: clipx0, clipx1 ! internal
      REAL,intent(IN)               :: clipy0_in, clipy1_in
      real(real32) SREND_SAVE                   :: clipy0, clipy1 ! internal
      REAL,intent(IN)               :: clipz0_in, clipz1_in
      real(real32) SREND_SAVE                   :: clipz0, clipz1 ! internal
      REAL,intent(IN)               :: pt_in(1:3)
      real(real32) SREND_SAVE                   :: pt(1:3) ! internal
      REAL,intent(IN)               :: pt_r_in
      real(real32) SREND_SAVE                   :: pt_r ! internal
      REAL,intent(IN)               :: cp_p_in(1:3), cp_n_in(1:3)
      real(real32) SREND_SAVE                  :: cp_p(1:3), cp_n(1:3) ! internal

#ifdef SREND_SOLID_RENDERING
      INTEGER,intent(IN)            :: Sd(0:1,1:3) ! 0=min:1=max,1=x 2=y 3=z
      CHARACTER,intent(IN)          :: S(Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(1,3))
      CHARACTER,intent(IN)          :: Srgb(1:3,Sd(0,1):Sd(1,1),Sd(0,2):Sd(1,2),Sd(0,3):Sd(0,3))
#endif
! results out
      CHARACTER,intent(INOUT)       :: mem(:)
      INTEGER,intent(INOUT)         :: memlen
! end: passed parameters --------------------------------------------

      integer(int32) :: i,j,k ! counters
#ifndef SREND_16BIT
      integer(int32) :: ii    ! counter
#endif

! ----------- target image array -------------------------------------
      real(real32),dimension(:,:,:,:),allocatable SREND_SAVE :: im
!DIR$ ATTRIBUTES ALIGN:64 :: im
! ------- color table ------------------------------------------------
! this is formed by renderer below on first pass and saved in
! B(nV)%cotab.  On following passes, the first threads copies
! B(nV)%cotab to cotab for all threads.
! color table (Alpha=0, R=1, G=2, B=3,index in 0:255)
      real(real32),dimension(0:255,0:3) SREND_SAVE :: cotab
!DIR$ ATTRIBUTES ALIGN:64 :: cotab

! ------------ variables for finding bounds ----------
      real(real32)                           :: a2,b2              ! alpha/2 and beta/2 in radians
      real(real32) SREND_SAVE                :: dW, dH             ! increments for W and H
      real(real32), dimension(1:3)           :: E0, V0, U0, R0     ! pre-processing vars to form view basis
      real(real32),dimension(1:3) SREND_SAVE :: V, U, R            ! polished basis, orthogonal and unit
!DIR$ ATTRIBUTES ALIGN:64 :: V, U, R
! computing arrays, checking bounding cubes
      real(real32)                           :: CubeCenter(1:3)    ! point in space, center of cube/volume
      real(real32)                           :: Ray_EC(1:3)        ! ray EP
      real(real32)                           :: Dist_EC            ! distance from E to C
      real(real32)                           :: Unit_EC(1:3)       ! unitized EC
      real(real32)                           :: Radius             ! radius of bounding sphere
      real(real32)                           :: Unit_EC_projU(1:3) ! unitized EC - Proj_U (EC)
#ifdef SREND_STEREO
      real(real32),dimension(1:3)            :: Ray_EC_stereo      ! stereo: ray EC
      real(real32)                           :: Dist_EC_stereo     ! stereo: distance EC
#endif
      real(real32)                           :: acos_arg           ! intermediate result, to check
      real(real32)                           :: aa,bb              ! lon/lat coordinate
      real(real32)                           :: gam                ! angle span from EC to touch bounding sphere
      real(real32)                           :: aax                ! expansion of gam with lat inc/dec
      integer(int32)                         :: ww0, ww1           ! intermediate bounds for w0, w1

      integer(int32) SREND_SAVE              :: w0,w1,h0,h1        ! coordinates render plane: (1:nx,1:ny,1:nz)
      integer(int32) SREND_SAVE              :: w4, h4             ! vars to adjust dims to multiple of 4, reused for dim 16packs
      integer(int32) SREND_SAVE              :: w4h4               ! number of 4x4=16 packs, used for allocatations
      integer(int32) SREND_SAVE              :: wh(1:4)            ! working bounds actually used: 1=w0 2=w1 3=h0 4=h1

      real(real32) SREND_SAVE                :: E_AMR_PASS(1:3) ! needed to be shared : E is altered by AMR 

! ------- for finding ray hits ---------------------------------------
! the saved variables here are set by 1 thread in setup, then the vars with SAVE are used by all threads for ray casting
      integer(int32), dimension(:,:), allocatable SREND_SAVE  :: rr                 ! horz pixel coord
      integer(int32), dimension(:,:), allocatable SREND_SAVE  :: cc                 ! vert pixel coord
      real(real32), dimension(:,:,:), allocatable SREND_SAVE  :: Q                  ! ray to cast
      real(real32), dimension(1:16)                           :: a_angle, b_angle   ! making Q casting vector
!DIR$ ATTRIBUTES ALIGN:64 :: rr, cc, Q, a_angle, b_angle
      real(real32), dimension(1:16)                           :: cosb, cosbcosa     ! making Q casting vector
      real(real32), dimension(1:16)                           :: cosbsina, sinb     ! making Q casting vector
!DIR$ ATTRIBUTES ALIGN:64 :: cosb, cosbcosa, cosbsina, sinb
      real(real32), dimension(:,:), allocatable SREND_SAVE    :: tmin, tmax         ! bounds of ray and pack in cube
      integer(int32), dimension(1:16)                         :: tmiss              ! flag
      real(real32), dimension(:), allocatable SREND_SAVE      :: tpackmax, tpackmin ! min/max of tmin/max by pack
      integer(int32),dimension(:),allocatable SREND_SAVE      :: itpackmax,itpackmin! min/max in dt increments
!DIR$ ATTRIBUTES ALIGN:64 :: tmin, tmax, tmiss, tpackmax, tpackmin, itpackmax, itpackmin
      real(real32),dimension(1:16)                            ::  t0                ! test ray intesections
      real(real32), dimension(1:16)                           :: x,y,z              ! cell coord
      integer(int32),dimension(:),allocatable SREND_SAVE      :: shiftW             ! shift/unshift across dateline
      integer(int32),dimension(:),allocatable SREND_SAVE      :: hit
!DIR$ ATTRIBUTES ALIGN:64 :: t0, x, y, z, shiftW, hit
      integer(int32) SREND_SAVE                               :: numhits            ! packs that actually hit cube
! optimizations to reduce arithmetic : save for OMP threads
      integer(int32),dimension(:),allocatable SREND_SAVE      :: im_cc, im_rr       ! mapping sampling results to target im() array
!DIR$ ATTRIBUTES ALIGN:64 :: im_cc,im_rr
#ifdef SREND_STEREO
      real(real32), dimension(1:16)                           :: dstQ,dstQ2         ! Dist common clip sphere
      real(real32), dimension(1:16)                           :: clip00, clip11     ! clip near/far stereo
      real(real32),dimension(1:3) SREND_SAVE                  :: Dist_stereo        ! vector shift right
      real(real32) SREND_SAVE                                 :: Dist_2_stereo      ! = dot(dst,dst)
      real(real32),dimension(1:16)                            :: dst_discr          ! used for stereo common clip r's
!DIR$ ATTRIBUTES ALIGN:64 :: dstQ,dstQ2, clip00, clip11, dst_discr
#endif
#ifdef SREND_TOFF
      real(real32),dimension(:,:),allocatable SREND_SAVE      :: toff               ! fix offset from clip0 and clip00
!DIR$ ATTRIBUTES ALIGN:64 :: toff
#endif
#ifdef SREND_POINT_RADIUS_CLIP
      real(real32),dimension(1:16)                            :: pta,ptb,ptc,ptdiscr,pt0,pt1
!DIR$ ATTRIBUTES ALIGN:64 :: pta,ptb,ptc,ptdiscr,pt0,pt1
#endif
#ifdef SREND_CLIPPING_PLANE
      real(real32) SREND_SAVE                                 :: n_dot_p            !n_dot_p = dot_product(cp_n,cp_p) ! do once in setup
      real(real32),dimension(1:16)                            :: n_dot_q  !q_dot_n(i) = dot_product(cp_n,(/Q(i,1,k),Q(i,2,k),Q(i,3,k)/) )
!DIR$ ATTRIBUTES ALIGN:64 :: n_dot_q
#endif
#if SREND_TEAM_RENDER==1
      integer(int32),dimension(:),allocatable SREND_SAVE      :: th_numhits ! threads in setting tmin and tmax, testing each ray in volume
      integer(int32),dimension(:,:),allocatable SREND_SAVE    :: th_hit
      integer(int32)                                          :: ith
      integer(int32),dimension(:,:),allocatable SREND_SAVE    :: th_wh
#endif
      integer(int32)                                          :: ic, ir             ! loop vars to set rr,cc for 16-pack
      real(real32) SREND_SAVE                                 :: x0,x1,y0,y1,z0,z1  ! volume dimensions used to set ray lengths before casting

! -------- computing variables: sampling -----------------------------
      integer(int32)                                          :: tt                 ! steps through ray pack
#ifdef SREND_TOFF
      real(real32),dimension(1:16)                            :: t                  ! ray increment, with stereo adjust
!DIR$ ATTRIBUTES ALIGN:64 :: t
#else
      real(real32)                                            :: t                  ! ray increment
#endif
      integer(int32), dimension(1:16)                         :: ix, iy, iz         ! hold values during casting
!DIR$ ATTRIBUTES ALIGN:64 :: ix, iy, iz
#ifdef SREND_SOLID_RENDERING
      integer(int32), dimension(1:16)                         :: ixx, iyy, izz      ! hold values during casting, for solid
!DIR$ ATTRIBUTES ALIGN:64 :: ixx, iyy, izz
#endif
      integer(int32), dimension(1:16)                         :: cotabi             ! RGBA values at a sample
!DIR$ ATTRIBUTES ALIGN:64 :: cotabi
      real(real32),dimension(1:16)                            :: dR,dG,dB,dA        ! tracer val
!DIR$ ATTRIBUTES ALIGN:64 :: dR,dG,dB, dA

#ifdef SREND_SOLID_RENDERING
      real(real32),dimension(1:16)                            :: s00,s10,s01,s11
!DIR$ ATTRIBUTES ALIGN:64 :: S00,S10,S01,S11
      integer(int32)                                          :: Sminmax_xyz(0:1,1:3) ! 0=min:1=max,1=x 2=y 3=z
#endif
      real(real32),dimension(1:16)                            :: q1Mtval            ! optimize
      real(real32),dimension(1:16)                            :: transparency       ! optimize
      real(real32),dimension(1:16)                            :: tval               ! scaling RGB  ! accumulated in i-pass from ray
      real(real32),dimension(1:16)                            :: ddA                ! alpha value
      real(real32),dimension(1:16)                            :: ddR                ! tracer value
      real(real32),dimension(1:16)                            :: ddG                ! tracer value
      real(real32),dimension(1:16)                            :: ddB                ! tracer value
      real(real32),dimension(1:16)                            :: tmask              ! mask out for t values outside of ray hit
!DIR$ ATTRIBUTES ALIGN:64 :: q1Mtval, transparency,tval,ddA,ddR,ddG,ddB,tmask
      real(real32),dimension(1:16)                            :: dx0,dy0,dz0,dx1,dy1,dz1 ! for trilinear inerpolation
      real(real32),dimension(1:16,1:8)                        :: dat8               ! pre-align data
!DIR$ ATTRIBUTES ALIGN:64 :: dx0,dy0,dz0,dx1,dy1,dz1,dat8
! ----- OMP threads - - -- --- - -- - - - - -- - --- - - - -  - -- ---
      integer(int32)                                          :: th                 ! thread number, 0:SREND_OMP_THREADS-1
      integer(int32)                                          :: nth                ! number of OMP threads calling
! partition work among threads, simply so using 16 packs
#if SREND_TEAM_RENDER==1
      integer(int32)                                          :: part, Wpart, Rpart ! to form partitions by rows
      integer(int32)                                          :: th0, th1           ! bounds for each thread's compositing
#endif
#ifdef SREND_TIMING
      integer(int64)                                          :: it1, it2, itc, it_start ! for testing .. collect statistics
      real(real32)                                            :: rc                 ! ticks/sec
#endif
#ifdef SREND_WOODWARD_AMR
       integer(int32),dimension(1:16)                         :: ibq,jbq,kbq        ! indices for Woodward AMR
!DIR$ ATTRIBUTES ALIGN:64 :: ibq,jbq,kbq
#endif
#ifndef SREND_16BIT
       real(real32),dimension(1:64)                           :: real4_cast256      ! vectors of 16*4 reals --> 64*4 char*1
#endif
! ------ behavior ---- type of operation for this call, local per thread
      logical                                                 :: load               ! set to T for a load call
      logical                                                 :: flush              ! set to T for a flush call
      logical                                                 :: image              ! set to T for flush2image
! --------------------------------------------------------------------
      type(B_header_type) SREND_SAVE                          :: Bh
!DIR$ ATTRIBUTES ALIGN:64 :: Bh

#ifdef SREND_FLOPS
      integer(int64)                                          :: flop ! performance count
#endif

! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!  var declarations done 
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

! initialize if B_DYNAMIC_V, then check each call to see if large enough
#ifdef SREND_B_DYNAMIC_V
#if SREND_TEAM_RENDER==1
!$OMP SINGLE
#endif
      if( .NOT. B_used) then
         allocate( B(1:SREND_MAX_V) )
         B_used = .TRUE.
      end if

      i = size(B)
      if(nV > i) then
         j = max(nV,2*i)
         allocate(B_temp(1:j))
         B_temp(1:i) = B
         call move_alloc(B_temp,B) ! from B_temp to B
      end if
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif
#endif
      
      

#ifdef SREND_TIMING
!$OMP MASTER
      call system_clock(it1,itc); rc = 1.0_real32 / itc
      it_start = it1
!$OMP END MASTER
#endif

#ifdef SREND_FLOPS
      flop=0
#endif


! ---- setup cotab
#if SREND_TEAM_RENDER==1 || !defined(_OPENMP)
!$OMP SINGLE
#ifndef SREND_COTAB_REBUILD_EACH_PASS
      if(.NOT. B(nV)%cotab_set) then
         call srend_set_cotab()
         B(nV)%cotab_set = .TRUE.
      else
         cotab = B(nV)%cotab  
      end if
#else
      call srend_set_cotab() ! only this if REBUILD_EACH_PASS
#endif 
!$OMP END SINGLE
#endif

#if SREND_TEAM_RENDER==0 && defined(_OPENMP)
#ifndef SREND_COTAB_REBUILD_EACH_PASS
      if(.NOT. B(nV)%cotab_set) then
!$OMP CRITICAL (color_table)
         if(.NOT. B(nV)%cotab_set ) then ! only 1 thread will do this
            call srend_set_cotab()
            B(nV)%cotab_set = .TRUE.
         else
            cotab = B(nV)%cotab 
         end if
!$OMP END CRITICAL (color_table)
      else
         cotab = B(nV)%cotab ! only this if REBUILD_EACH_PASS
      end if
#else
      call srend_set_cotab()
#endif
#endif


! set behavior, always local
      image = .FALSE. ; flush = .FALSE. ; load = .FALSE.
      if( srendType < 0 ) image = .TRUE.
      if( srendType ==0 ) flush = .TRUE.
      if( srendType > 0 ) load  = .TRUE.
      
      
! setup B structure for loading, if load==.TRUE.
#if SREND_TEAM_RENDER==1 || !defined(_OPENMP)
!$OMP SINGLE
      if(load .AND. .NOT. B(nV)%used ) then ! will always use this many
            allocate( B(nV)%Bh(1:B_max_LOAD) ) !once
            allocate( B(nV)%cube(1:B_max_LOAD) )             !once
            B(nV)%used  = .TRUE.
      end if
!$OMP END SINGLE
#else
      if(load .AND. .NOT. B(nV)%used ) then ! only enter if needed, just once per nV
!$OMP CRITICAL (B_structure)
         if(.NOT. B(nV)%used) then ! only 1 thread will do this
            allocate( B(nV)%Bh(1:B_max_LOAD) ) !once
            allocate( B(nV)%cube(1:B_max_LOAD) )             !once
            B(nV)%used  = .TRUE.
         end if
!$OMP END CRITICAL (B_structure)
      end if
#endif


      
#ifdef SREND_TIMING
!$OMP MASTER
      call system_clock(it2)
      print *,'RENDER block alloc nV,cotab: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
!$OMP END MASTER
#endif
      


! the first thread does this, all wait at end
#if SREND_TEAM_RENDER==1
!$OMP SINGLE 
      Bh%nV = nV ! save in Bheader
#else
      Bh%nV = nV ! save in Bheader
#endif
! SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
! initial setup
! SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
! --------- copy arguments into Bheader for passing----------------
! Here the "_in" arguments might be real(real64) or even integer(int64), so
! make them real(real32) and integer(int32) for those than matter for purposes
! of packing in Bheader and vector operations.
      Bh%Vdim = Vdim_in ! set Vdim
      dt = dt_in
      Bh%XN = XN_in; Bh%YN = YN_in; Bh%ZN = ZN_in
      Bh%perspective = perspective_in
      
! ordinary rendering, default  
      Bh%W = W_in; Bh%H = H_in
      Bh%Alpha = Alpha_in; Bh%Beta = Beta_in
#ifndef SREND_WOODWARD_AMR
      Bh%xoff = xoff_in
      Bh%yoff = yoff_in
      Bh%zoff = zoff_in
#else
      Bh%xoff = xoff_in + xoffth
      Bh%yoff = yoff_in + yoffth
      Bh%zoff = zoff_in + zoffth
#endif
      
! --------------- prep for stereo shift ------------------------------ 
      if(EyeRight_in /= 0.0_real32) then
! > make orthogonal basis
         R0(1) = V_in(2)*U_in(3) - V_in(3)*U_in(2) ! R = V x U
         R0(2) = V_in(3)*U_in(1) - V_in(1)*U_in(3)
         R0(3) = V_in(1)*U_in(2) - V_in(2)*U_in(1)
         U0(1) = R0(2)*V_in(3) - R0(3)*V_in(2) ! U = R x V
         U0(2) = R0(3)*V_in(1) - R0(1)*V_in(3)
         U0(3) = R0(1)*V_in(2) - R0(2)*V_in(1)
         V0 = V_in ! copy of V
! unitize vectors ! none should be zero
         V0 = V0 / sqrt( V0(1)**2 + V0(2)**2 + V0(3)**2 ) ! unitize
         R0 = R0 / sqrt( R0(1)**2 + R0(2)**2 + R0(3)**2 )
         U0 = U0 / sqrt( U0(1)**2 + U0(2)**2 + U0(3)**2 )
! shift eye
         E0 = E_in + EyeRight_in * R0
! EyeConvD shift to V0
         if( EyeConvD_in > 0.0_real32) then
            V0 = V0 - EyeRight_in*R0 / EyeConvD_in ! shift view vector to point along original V
            V0 = V0 / sqrt( V0(1)**2 + V0(2)**2 + V0(3)**2 ) ! unitize
         end if
      else ! not stereo
        E0 = E_in
        V0 = V_in
        U0 = U_in
      end if
      
      E_AMR_PASS = E0       ! this will be passed up for compositing
      Bh%E = E0 * Bh%Vdim ! for rendering

! load modified view variables
      Bh%V = V0
      Bh%U = U0
      
      
#ifdef SREND_STEREO
! distance from eye E0 to center eye position E_in
      Dist_stereo = (E_in - E0) * Bh%Vdim ! used in ray bound/intersection calcs
      Dist_2_stereo = dot_product(Dist_stereo,Dist_stereo)
#endif

! ---- coordinates, all tranformed the same: [x + xform_shift]*xform_vol_dim
      clipx0 = clipx0_in*Bh%Vdim
      clipx1 = clipx1_in*Bh%Vdim
      clipy0 = clipy0_in*Bh%Vdim
      clipy1 = clipy1_in*Bh%Vdim
      clipz0 = clipz0_in*Bh%Vdim
      clipz1 = clipz1_in*Bh%Vdim

! ---- distances, all transformed the same: x * xform_vol_dim
      clip0 = clip0_in * Bh%Vdim
      clip1 = clip1_in * Bh%Vdim

! point radius clip
#ifdef SREND_POINT_RADIUS_CLIP
      pt = pt_in*Bh%Vdim
      pt_r = pt_r_in*Bh%Vdim
#else
      pt = pt_in      ! unused, but supress compiler warnings
      pt_r = pt_r_in
#endif

#ifdef SREND_CLIPPING_PLANE
      cp_n = cp_n_in
      cp_p = cp_p_in * Bh%Vdim
      n_dot_p = dot_product(cp_n,cp_p - Bh%E) ! do once in setup
#else
      cp_n = cp_n_in ! unused, but supress compiler warnings
      cp_p = cp_p_in
#endif
         
      
! SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
! determine bounding array
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

! viewing increments, assume W and H are divisible by 4
      a2 = Bh%Alpha * SREND_PI/360.0_real32  ! half span
      b2 = Bh%Beta  * SREND_PI/360.0_real32  ! half span
      dW = a2 * 2.0_real32 /Bh%W
      dH = b2 * 2.0_real32 /Bh%H

! don't alter incoming parameters V,U: make orthogonal basis
      R(1) = Bh%V(2)*Bh%U(3) - Bh%V(3)*Bh%U(2) ! R = V x U
      R(2) = Bh%V(3)*Bh%U(1) - Bh%V(1)*Bh%U(3)
      R(3) = Bh%V(1)*Bh%U(2) - Bh%V(2)*Bh%U(1)
      U(1) = R(2)*Bh%V(3) - R(3)*Bh%V(2) ! U = R x V
      U(2) = R(3)*Bh%V(1) - R(1)*Bh%V(3)
      U(3) = R(1)*Bh%V(2) - R(2)*Bh%V(1)
      V = Bh%V ! copy of V
      V = V / sqrt( V(1)**2 + V(2)**2 + V(3)**2 ) ! unitize
      R = R / sqrt( R(1)**2 + R(2)**2 + R(3)**2 )
      U = U / sqrt( U(1)**2 + U(2)**2 + U(3)**2 )

! --------------------------------------------------------------------
! Set an initial sufficient spanning frustum before testing rays.
! This array is w0:w1,h0:h1
! set this array to something if missed: 1:4,1:4

      Radius = sqrt( 1.0_real32*(Bh%XN**2 + Bh%YN**2 + Bh%ZN**2) ) * .5_real32
      CubeCenter(1) = Bh%xoff + .5_real32*Bh%XN ! center of cubes
      CubeCenter(2) = Bh%yoff + .5_real32*Bh%YN
      CubeCenter(3) = Bh%zoff + .5_real32*Bh%ZN
        
      Ray_EC = CubeCenter - Bh%E
      Dist_EC = sqrt( Ray_EC(1)**2 + Ray_EC(2)**2 + Ray_EC(3)**2 )

#ifdef SREND_STEREO
      Ray_EC_stereo = CubeCenter - Bh%E
      Dist_EC_stereo = sqrt( Ray_EC_stereo(1)**2 + Ray_EC_stereo(2)**2 + Ray_EC_stereo(3)**2 )
#endif

! check if embedded, really just too close to know for sure not        
      if( Dist_EC <= Radius + 1.0_real32) then
         w0=1; w1=Bh%W; h0=1; h1=Bh%H
         goto 1666 ! done with this one
      end if
! near clip check
#ifdef SREND_STEREO
      if( Dist_EC_stereo + Radius < clip0) then
#else
      if( Dist_EC + Radius < clip0) then
#endif
         w0=1; w1=4; h0=1; h1=4 ! values to reject
         goto 1666
      end if
! far clip check
#ifdef SREND_STEREO
      if( Dist_EC_stereo - Radius > clip1) then
#else
      if( Dist_EC - Radius > clip1) then
#endif
         w0=1; w1=4; h0=1; h1=4 ! values to reject
         goto 1666
      end if
        
      Unit_EC = Ray_EC / Dist_EC ! make a unit vector, Dist_EC > 0 assured
      
      acos_arg = dot_product(Unit_EC,U)
      if( acos_arg >= 1.0_real32) acos_arg = 1.0_real32
      if( acos_arg <= -1.0_real32) acos_arg = -1.0_real32
      bb = SREND_PI*.5_real32 - acos( acos_arg )
!      bb = PI*.5 - acos( dot_product(Unit_EC,U) ) ! angle up to center! replaced by above
!      possible trouble with acos() due to numerical roundoff

      gam = asin( Radius / Dist_EC) ! angle off center to sphere surface
!     we know radius + 1.0 < Dist_EC and radius > 0
        
! now, setup span across for w0,W1
      if(bb + gam >= SREND_PI*.5_real32 - .05_real32) then ! North pole
         w0 = 1
         w1 = Bh%W
      else if(bb - gam <= -SREND_PI*.5_real32 + .05_real32) then ! South pole
         w0 = 1
         w1 = Bh%W
      else ! calculations for w0,w1
         Unit_EC_projU = Unit_EC - dot_product(Unit_EC,U)*U ! Unit_EC - projection to U
         Unit_EC_projU = Unit_EC_projU / sqrt( Unit_EC_projU(1)**2 + Unit_EC_projU(2)**2 + Unit_EC_projU(3)**2 )
!       fine, as Unit_EC cannot be even close to U=U
         
         acos_arg = dot_product(Unit_EC_projU,V)
         if( acos_arg >= 1.0_real32) acos_arg = 1.0_real32
         if( acos_arg <= -1.0_real32) acos_arg = -1.0_real32
         aa = acos( acos_arg )
!        aa = acos( dot_product(Unit_EC_projU,V) ) ! this is replaced by above
!        there was trouble - i.e. acos(1.00000) = NaN

         if( dot_product(Unit_EC_projU,R) > 0.0_real32) aa = -aa  

! calculation expansion due to elev/decl from equator
! we already know that the angle will not cause cos = 0
         aax = 1.0_real32/cos( min( abs(bb)+gam, b2) )

         ww0 = floor( (a2 -(aa+gam*aax)-.501_real32*dW)/dW )
         ww1 = ceiling( (a2-(aa-gam*aax)+.501_real32*dW)/dW )     
     
         if( Bh%Alpha < 360.0_real32) then
            w0 = max(1,ww0 )
            w1 = min(Bh%W,ww1 )
         else ! Alpha = 360.0
            if(ww1 - ww0 >= Bh%W) then
               w0 = 1
               w1 = Bh%W
            else if(ww1 > Bh%W) then
               w0 = ww0 - Bh%W
               w1 = ww1 - Bh%W   
               if( w0 <= -Bh%W/2) then
                  w0 = 1
                  w1 = Bh%W
               end if
            else
               w0 = ww0
               w1 = ww1
            end if
         end if

         if(w1-w0 > Bh%W .OR. w0 < -Bh%W/2 .OR. w1 > Bh%W) then
            w0=1
            w1=Bh%W
         end if

      end if ! set w0, w1

      if(w1-w0 < 0) then
         w0=1; w1=4; h0=1; h1=4 ! values to reject
         goto 1666
      end if
      
! set h0, h1
      h0 = max(1,  floor( (b2-(bb+gam)-.501_real32*dH)/dH) )
      h1 = min(Bh%H,ceiling( (b2-(bb-gam)+.501_real32*dH)/dH) )
        
      if(h1 < h0 ) then
         w0=1; w1=4; h0=1; h1=4 ! values to reject
         goto 1666
      end if
            
! expand to multiple of 4, extend right and down first
! assume: W and H are divisible by 4, really 8 for W
! but, ensure that in spherical case we also have multiples of 4
! on either side of the 0:1 splitting dateline
      h4 = mod(h1,4) 
      if(h4 /= 0) h1 = h1 + 4-h4
      h4 = mod(h0-1,4)
      if(h4 /= 0) h0 = h0 - h4

      w4 = mod(w1,4)
      if(w4 /= 0) w1 = w1 + 4-w4
      w4 = mod(Bh%W + w0-1,4)
      if(w4 /= 0) w0 = w0 - w4 

      if( 1+w1-w0 >= Bh%W) then ! spans anyway
         w0 = 1
         w1 = Bh%W
      end if

1666   continue ! jump here if no hits are possible

! this is for the smp mode where renderer calls the finisher just for processing and writing
! out the image from rendering one block.
      if( image ) then ! must be full dimensions
         w0 = 1; w1 = Bh%W; h0 = 1; h1 = Bh%H
      end if

! adjust w4h4 for allocation if needed
      if(w1-w0 > 0) then
         w4h4 = (w1-w0+1)*(h1-h0+1)/16 ! number of 4x4=16 packs
      else
         numhits = 0 ! for each shell
         wh(1) = 1000000!1
         wh(2) = -1000000!0
         wh(3) = 1000000!1
         wh(4) = -1000000!0
      end if
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif


#ifdef SREND_TIMING
!$OMP MASTER
      call system_clock(it2)
      print *,'RENDER bounds: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
!$OMP END MASTER
#endif
      
      
! determine thread id for load sharing scheme: id = th, num threads = nth
! all must do here before allocations below by thread for th_ vars
#if defined(_OPENMP)
      th = OMP_get_thread_num()
      nth = OMP_get_num_threads()
#else
      th = 0
      nth = 1
#endif


      
      
! SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
! find hits
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      if(w1-w0 <= 0 ) goto 2666 ! nothing to test
      
      
#if SREND_TEAM_RENDER==1
!$OMP SINGLE
#endif

! set volume boundaries : NX=NY=NZ=1
      x0 = Bh%xoff 
      x1 = x0 + Bh%XN
      y0 = Bh%yoff
      y1 = y0 + Bh%YN
      z0 = Bh%zoff
      z1 = z0 + Bh%ZN


! adjust for XYZ plane clipping
      if(clipx0 > x0) x0 = clipx0
      if(clipx1 < x1) x1 = clipx1
      if(clipy0 > y0) y0 = clipy0
      if(clipy1 < y1) y1 = clipy1
      if(clipz0 > z0) z0 = clipz0
      if(clipz1 < z1) z1 = clipz1

! will set during ray casting bounds checks to minimum span (wo:w1,h0:h1)
      wh(1) =  1000000000 ! w0
      wh(2) = -1000000000 ! w1
      wh(3) =  1000000000 ! h0
      wh(4) = -1000000000 ! h1

! reuse w4 and h4 to be number of 16-packs in each direction
      w4 = (1+w1-w0)/4
      h4 = (1+h1-h0)/4
      
      numhits = 0     ! zero all shells
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif


! check if clipping planes remove volume to render
      if( x1-x0<=0.0_real32 .OR. y1-y0<=0.0_real32 .OR. z1-z0<=0.0_real32) goto 2666 ! nothing to test

      
#if SREND_TEAM_RENDER==1
!$OMP SINGLE
#endif

! --------- allocate arrays based on rendering plane -----------------
! check and see if arrays are already large enough.  The unused ends
! of arrays will do no harm.
#if SREND_TEAM_RENDER==1
      if(allocated(rr)) then ! just test one of the following, all saved
         if( w4h4 > size(rr,DIM=2) ) then ! increase size if needed
            deallocate(rr)
            deallocate(cc)
            deallocate(Q)
            deallocate(tmin)
            deallocate(tmax)
            deallocate(tpackmin)
            deallocate(tpackmax)
            deallocate(itpackmin)
            deallocate(itpackmax)
            deallocate(shiftW)
            deallocate(hit)
            deallocate(im_cc)
            deallocate(im_rr)
            deallocate(th_numhits)
            deallocate(th_hit)
            deallocate(th_wh)
#ifdef SREND_TOFF
            deallocate(toff)
#endif
         end if
      end if

      if( .NOT. allocated(rr)) then
         allocate(    rr(1:16,w4h4) )
         allocate(    cc(1:16,w4h4) )
         allocate( Q(1:16,1:3,w4h4) )
         allocate(  tmin(1:16,w4h4) )
         allocate(  tmax(1:16,w4h4) )
         allocate(   tpackmax(w4h4) )
         allocate(   tpackmin(w4h4) )
         allocate(  itpackmax(w4h4) )
         allocate(  itpackmin(w4h4) )
         allocate(     shiftW(w4h4) )
         allocate(        hit(w4h4) )
         allocate(      im_cc(w4h4) )
         allocate(      im_rr(w4h4) )
         allocate( th_numhits(0:nth-1) )
         allocate(th_hit(w4h4,0:nth-1) )
         allocate(  th_wh(1:4,0:nth-1) )
#ifdef SREND_TOFF
         allocate(  toff(1:16,w4h4) )
#endif
      end if

! set bounds for all threads, will consolodate later with 1 thread
      do ith=0,nth-1
         th_wh(1,ith) = wh(1)
         th_wh(2,ith) = wh(2)
         th_wh(3,ith) = wh(3)
         th_wh(4,ith) = wh(4)
      end do
      
      th_numhits = 0  ! zero all threads and shells
#endif
! --- end SREND_TEAM_RENDER setup

! --- single thread
#if SREND_TEAM_RENDER==0
         allocate(    rr(1:16,w4h4) )
         allocate(    cc(1:16,w4h4) )
         allocate( Q(1:16,1:3,w4h4) )
         allocate(  tmin(1:16,w4h4) )
         allocate(  tmax(1:16,w4h4) )
         allocate(   tpackmax(w4h4) )
         allocate(   tpackmin(w4h4) )
         allocate(  itpackmax(w4h4) )
         allocate(  itpackmin(w4h4) )
         allocate(     shiftW(w4h4) )
         allocate(        hit(w4h4) )
         allocate(      im_cc(w4h4) )
         allocate(      im_rr(w4h4) )
#ifdef SREND_TOFF
         allocate(  toff(1:16,w4h4) )
#endif

#endif
! end single thread
      

      
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif


! --------- test rays, all threads --------------------
! entry test, same for all

#if SREND_TEAM_RENDER==1
      th0 = 1 + th * (w4h4 / nth)
      th1 = (th+1) * (w4h4 / nth) +  w4h4 - nth*(w4h4/nth)

do k=th0, th1 !>>>>>>>>>>> each thread takes a piece
#else
do k=1,w4h4
#endif

! ccccccccccccccccccccccccccccccccccccccccccccccccccccc
! We form 16-packs of pixels/rays

! load in indices for 4x4=16 pack
! rr(i,k) and cc(i,k) values are the im grid coordinates
            do ir = 0,3
               do ic = 0,3
                  rr(1+ic+4*ir,k) = ir + h0 + 4*((k-1)/w4)
                  cc(1+ic+4*ir,k) = ic + w0 + 4*mod((k-1),w4)
               end do
            end do

! here, we translate the cc pixel coordinates for
! spherical when cc < 1 for purposes of casting
! we translate these back before placing back in im() array,
! in this case.  Note that cc(i) will always be < 1 for
! all i = 1 to 16 for the entire pack.  
! But >>>> if w0=1 and w1 = W, don't unshift afterwards
! We just add W to shift here.
            shiftW(k) = 0
            if( cc(1,k) < 1) then ! all cc(:,k), and really <= -3
               shiftW(k) = Bh%W
               do i=1,16
                  cc(i,k) = cc(i,k) + shiftW(k)
               end do
            end if

! for mapping within rendering loop
            im_cc(k) =  (Bh%W + cc(1,k) - shiftW(k) + 3)/4 - Bh%W/4
            im_rr(k) =  (rr(1,k) + 3)/4

#ifdef SREND_TOFF
            toff(:,k) = 0.0_real32 ! will find the max value, move forward with additional clipping
#endif
            tmiss(:) = 1
            tmin(:,k) = 0.0_real32
            tmax(:,k) = 0.0_real32

SREND_VECTORIZE
do i=1,16 ! VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV

! make viewing vector Q
  a_angle(i) = (Bh%W/2 - cc(i,k))*dW + dW/2.0_real32
  b_angle(i) = (Bh%H/2 - rr(i,k))*dH + dH/2.0_real32

  cosb(i) = cos(b_angle(i))
  cosbcosa(i) = cosb(i)*cos(a_angle(i))
  cosbsina(i) = cosb(i)*sin(a_angle(i))
  sinb(i) = sin(b_angle(i))

! Q is unit here
  Q(i,1,k) = V(1)*cosbcosa(i) - R(1)*cosbsina(i) + U(1)*sinb(i)
  Q(i,2,k) = V(2)*cosbcosa(i) - R(2)*cosbsina(i) + U(2)*sinb(i)
  Q(i,3,k) = V(3)*cosbcosa(i) - R(3)*cosbsina(i) + U(3)*sinb(i)

! ------------------  remove sampling artifacts on clipping boundaries
#ifdef SREND_TOFF
! find offsets for toff: only one of these might be hit, unless on edge or corner--which is fine
! x
if( Bh%E(1) < clipx0 .AND. Q(i,1,k) > 0.0_real32) then
   t0(i) = (clipx0 - Bh%E(1))/Q(i,1,k)
   y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if (     y(i) >= y0    .AND. y(i) < y1  &
      .AND. z(i) >= z0    .AND. z(i) < z1) then
      toff(i,k) = t0(i)
   end if
end if

if( Bh%E(1) > clipx1 .AND. Q(i,1,k) < 0.0_real32) then
   t0(i) = (clipx1 - Bh%E(1))/Q(i,1,k)
   y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if (     y(i) >= y0    .AND. y(i) < y1  &
      .AND. z(i) >= z0    .AND. z(i) < z1) then
      toff(i,k) = t0(i)
   end if
end if


! y
if(Bh%E(2) < clipy0 .AND. Q(i,2,k) > 0.0_real32) then
   t0(i) = (clipy0 - Bh%E(2))/Q(i,2,k)
   x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
   z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if (     x(i) >= x0    .AND. x(i) < x1  &
      .AND. z(i) >= z0    .AND. z(i) < z1) then
      toff(i,k) = t0(i)
   end if
end if

if(Bh%E(2) > clipy1 .AND. Q(i,2,k) < 0.0_4) then
   t0(i) = (clipy1 - Bh%E(2))/Q(i,2,k)
   x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
   z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if (     x(i) >= x0    .AND. x(i) < x1  &
      .AND. z(i) >= z0    .AND. z(i) < z1) then
      toff(i,k) = t0(i)
   end if
end if


! z
if(Bh%E(3) < clipz0 .AND. Q(i,3,k) > 0.0_real32) then
   t0(i) = (clipz0 - Bh%E(3))/Q(i,3,k)
   x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
   y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   if (     x(i) >= x0    .AND. x(i) < x1  &
      .AND. y(i) >= y0    .AND. y(i) < y1) then
      toff(i,k) = t0(i)
   end if
end if

if(Bh%E(3) > clipz1 .AND. Q(i,3,k) < 0.0_real32) then
   t0(i) = (clipz1 - Bh%E(3))/Q(i,3,k)
   x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
   y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   if (     x(i) >= x0    .AND. x(i) < x1  &
      .AND. y(i) >= y0    .AND. y(i) < y1) then
      toff(i,k) = t0(i)
   end if
end if

#endif
! ------------------------ end: remove clipping artifacts ------------


! ---- find extent of sampling ray in this subvolume: [tmin,tmax] ----
! x
if (Q(i,1,k) /= 0.0_real32) then
   t0(i) = (x0 - Bh%E(1))/Q(i,1,k)
   y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if ( t0(i) >= 0.0_real32 .AND.        &
         y(i) >= y0    .AND. y(i) < y1  &
   .AND. z(i) >= z0    .AND. z(i) < z1) then
      if(Q(i,1,k)>0.0_real32) tmin(i,k)=t0(i)
      if(Q(i,1,k)<0.0_real32) tmax(i,k)=t0(i)
   end if
end if

if( Q(i,1,k) /= 0.0_4) then
t0(i) = (x1 - Bh%E(1) )/Q(i,1,k)
y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if ( t0(i) >= 0.0_real32 .AND.         &
         y(i) >= y0    .AND. y(i) < y1   &
   .AND. z(i) >= z0    .AND. z(i) < z1) then
      if(Q(i,1,k)<0.0_real32) tmin(i,k)=t0(i)
      if(Q(i,1,k)>0.0_real32) tmax(i,k)=t0(i)
   end if
end if

! y
if (Q(i,2,k) /= 0.0_real32) then
t0(i) = (y0 - Bh%E(2))/Q(i,2,k)
x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if ( t0(i) >= 0.0_real32 .AND.          &
         x(i) >= x0    .AND. x(i) < x1   &
   .AND. z(i) >= z0    .AND. z(i) < z1) then
      if(Q(i,2,k)>0.0_real32) tmin(i,k)=t0(i)
      if(Q(i,2,k)<0.0_real32) tmax(i,k)=t0(i)
   end if
end if

if (Q(i,2,k) /= 0.0_real32) then 
t0(i) = (y1 - Bh%E(2) )/Q(i,2,k)
x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
z(i) = t0(i)*Q(i,3,k) + Bh%E(3)
   if ( t0(i) >= 0.0_real32 .AND.            &
         x(i) >= x0    .AND. x(i) < x1  &
   .AND. z(i) >= z0    .AND. z(i) < z1) then
      if(Q(i,2,k)<0.0_real32) tmin(i,k)=t0(i)
      if(Q(i,2,k)>0.0_real32) tmax(i,k)=t0(i)
   end if
end if

! z
if (Q(i,3,k) /= 0.0_real32) then
t0(i) = (z0 - Bh%E(3))/Q(i,3,k)
x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   if ( t0(i) >= 0.0_real32 .AND.         &
         x(i) >= x0    .AND. x(i) < x1  &
   .AND. y(i) >= y0    .AND. y(i) < y1) then
      if(Q(i,3,k)>0.0_real32) tmin(i,k)=t0(i)
      if(Q(i,3,k)<0.0_real32) tmax(i,k)=t0(i)
   end if
end if

if (Q(i,3,k) /= 0.0_real32) then
t0(i) = (z1 - Bh%E(3) )/Q(i,3,k)
x(i) = t0(i)*Q(i,1,k) + Bh%E(1)
y(i) = t0(i)*Q(i,2,k) + Bh%E(2)
   if ( t0(i) >= 0.0_real32 .AND.          &
         x(i) >= x0    .AND. x(i) < x1   &
   .AND. y(i) >= y0    .AND. y(i) < y1) then
      if(Q(i,3,k)<0.0_real32) tmin(i,k)=t0(i)
      if(Q(i,3,k)>0.0_real32) tmax(i,k)=t0(i)
   end if
end if

if( tmin(i,k)*tmax(i,k)>0.0_real32) tmiss(i) = 0
! --------- end: find extent of sampling ray in this sub-volume ------


! --------------------------------------------------------------------
#ifdef SREND_STEREO
! stereo below
! assume that clip0 and clip1 > |dstereo| : if not, clip00 and clip11 resp. would NaN
! and so toff is NaN, and if tests fail so that errors are trapped
  dstQ(i) = Dist_stereo(1)*Q(i,1,k)+Dist_stereo(2)*Q(i,2,k)+ Dist_stereo(3)*Q(i,3,k)
  dstQ2(i) = dstQ(i) * dstQ(i)
  
  dst_discr(i) = clip0**2 - Dist_2_stereo + dstQ2(i)
  clip00(i) = sqrt( dst_discr(i) )  + dstQ(i) ! no safety, + no extra calc
  
  tmin(i,k) = max( tmin(i,k), clip00(i) ) ! near clipping
#ifdef SREND_TOFF
  toff(i,k) = max( toff(i,k), clip00(i) )
#endif

  dst_discr(i) = clip1**2 - Dist_2_stereo + dstQ2(i)
  clip11(i) = sqrt( dst_discr(i) ) + dstQ(i) ! no safety, + no extra calc
  
  tmax(i,k) = min( tmax(i,k), clip11(i) ) ! far clipping
#endif

#ifndef SREND_STEREO
#ifdef SREND_TOFF
  toff(i,k) = max( toff(i,k), clip0 )
#endif
  tmin(i,k) = max( tmin(i,k), clip0 ) ! near clipping
  tmax(i,k) = min( tmax(i,k), clip1 ) ! far clipping
#endif
! --------------------------------------------------------------------


! --------------------------------------------------------------------
#ifdef SREND_POINT_RADIUS_CLIP  
! first, skip useless calculation if [tmin,tmax] cannot hit sphere, just by distance
pta(i) = sqrt( (Bh%E(1)-pt(1))**2 + (Bh%E(2)-pt(2))**2 + (Bh%E(3)-pt(3))**2 )
if( tmax(i,k) < pta(i) - pt_r .OR. tmin(i,k) > pta(i) + pt_r) then ! can't be in sqphere
   tmiss(i) = 1
else
! solve quadratic: t = -b/2a += sqrt(b^2 - 4ac)/2a
   pta(i) = Q(i,1,k)**2 + Q(i,2,k)**2 + Q(i,3,k)**2 ! this vector Q should never be zero
   ptb(i) = 2.0_real32*(Q(i,1,k)*(Bh%E(1)-pt(1)) + &
                 Q(i,2,k)*(Bh%E(2)-pt(2)) + &
                 Q(i,3,k)*(Bh%E(3)-pt(3)) )
   ptc(i) = pt(1)**2+Bh%E(1)**2-2.0_real32*pt(1)*Bh%E(1) + &
            pt(2)**2+Bh%E(2)**2-2.0_real32*pt(2)*Bh%E(2) + &
            pt(3)**2+Bh%E(3)**2-2.0_real32*pt(3)*Bh%E(3) - pt_r**2
   ptdiscr(i) = ptb(i)**2 - 4.0_real32*pta(i)*ptc(i) ! discriminant b^2 - 4ac
   if( ptdiscr(i) > 0.0_real32) then 
      ptdiscr(i) = sqrt(ptdiscr(i))
      pt0(i) = (-ptb(i) - ptdiscr(i))*0.5_real32 / pta(i)
      pt1(i) = (-ptb(i) + ptdiscr(i))*0.5_real32 / pta(i)
      if( pt1(i) < tmin(i,k) .OR. tmax(i,k) < pt0(i)) then ! current interval not in sphere, missed
         tmiss(i) = 1
      else 
         if( pt1(i) < tmax(i,k) ) tmax(i,k) = pt1(i)
         if( tmin(i,k) < pt0(i) ) tmin(i,k) = pt0(i)
#ifdef SREND_TOFF
         toff(i,k) = max( toff(i,k), pt0(i) )
#endif
      end if
   else ! ray missed, so mark as missed
      tmiss(i) = 1
   end if
end if
#endif
! -------------------------------------------------------------------

! ---------- clipping plane -----------------------------------------
! point(cp_x,cp_y,cp_z)=cp_p(1:3) normal(cp_A,cp_B,cp_Z)=cp_n(1:z)
#ifdef SREND_CLIPPING_PLANE
!n_dot_p = dot_product(cp_n,cp_p) ! do once in setup
n_dot_q(i) = dot_product(cp_n,(/Q(i,1,k),Q(i,2,k),Q(i,3,k)/) )
t0(i)=0.0_real32 ! lies in plane
if(n_dot_q(i) /= 0.0_real32) t0(i) = n_dot_p / n_dot_q(i)
if(t0(i) > 0.0_real32 .AND. n_dot_q(i) > 0.0_real32 .AND. tmin(i,k) < t0(i) ) tmin(i,k) = t0(i) ! near clip
#ifdef SREND_TOFF
if(t0(i) > 0.0_real32 .AND. n_dot_q(i) > 0.0_real32 .AND. toff(i,k) > t0(i) ) toff(i,k) = t0(i) ! near clip
#endif
if(t0(i) > 0.0_real32 .AND. n_dot_q(i) < 0.0_real32 .AND. t0(i) < tmax(i,k) ) tmax(i,k) = t0(i) ! far clip
#endif
! --------------------------------------------------------------------


#ifdef SREND_TOFF
! _EPS added to toff() ensure a small bite into the furthest near clipping surface
  toff(i,k) = mod(toff(i,k) + SREND_DT_EPS, dt)
#endif

if(tmiss(i) == 1) then ! mark this sample as missing for 16-pack bounds
   tmin(i,k) =  1.0e30_real32
   tmax(i,k) = -1.0e20_real32
end if

end do ! i over 4x4=16 pack : iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii
! vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv



! we now find the span for t to iterate over the pack 
            tpackmax(k) = maxval(tmax(:,k))
            tpackmin(k) = minval(tmin(:,k))

            if( tpackmax(k) < tpackmin(k) ) goto 3666 ! no hits

            itpackmin(k) = floor(tpackmin(k)/dt)
            itpackmax(k) = ceiling(tpackmax(k)/dt)
            
! OK here, so find minimal bounds for wh casting array
! hit() contains indices to 16-packs which have at least 1
! ray hitting cube
! NOTE: we cast in 1:W, but we shift to wh(1):wh(2) to store
#if SREND_TEAM_RENDER==1
            th_numhits(th) = th_numhits(th) + 1   ! threaded
            th_hit(th_numhits(th),th) = k         ! threaded

            if(cc(1 ,k)-shiftW(k) < th_wh(1,th)) th_wh(1,th) = cc(1,k)-shiftW(k)
            if(cc(16,k)-shiftW(k) > th_wh(2,th)) th_wh(2,th) = cc(16,k)-shiftW(k)
            if(rr(1, k) < th_wh(3,th)) th_wh(3,th) = rr(1, k)
            if(rr(16,k) > th_wh(4,th)) th_wh(4,th) = rr(16,k)
#else
            numhits = numhits+1
            hit(numhits) = k
            if(cc(1 ,k)-shiftW(k) < wh(1)) wh(1) = cc(1,k)-shiftW(k)
            if(cc(16,k)-shiftW(k) > wh(2)) wh(2) = cc(16,k)-shiftW(k)
            if(rr(1, k) < wh(3)) wh(3) = rr(1, k)
            if(rr(16,k) > wh(4)) wh(4) = rr(16,k)
#endif

3666        continue ! no hits
      end do ! k k k k k k k k k k k k k k k k k k k k k k k k k k k k

! --------- end: tested all rays -------------------------------------


#if SREND_TEAM_RENDER==1
!$OMP BARRIER
!$OMP SINGLE
#endif

#if SREND_TEAM_RENDER==1
! first extract the hits for each shell from all threads, may not be well ordered for sampling
         do ith = 0,nth-1
            do k = 1,th_numhits(ith)
                  numhits = numhits+1
                  hit(numhits) = th_hit(k,ith)
            end do
         end do

! now, hit( numhits ) has the indices to indices for sampling, but these are not
! ordered.  Order these, but not if numhits==1
      if(numhits>1) call srend_heapsort( numhits, hit(1) ) ! heapsort

! setup bounds (w0:w1,h0:h1) from threads
      do ith=0,nth-1
         if( th_wh(1,ith) < wh(1)) wh(1) = th_wh(1,ith)
         if( th_wh(2,ith) > wh(2)) wh(2) = th_wh(2,ith)
         if( th_wh(3,ith) < wh(3)) wh(3) = th_wh(3,ith)
         if( th_wh(4,ith) > wh(4)) wh(4) = th_wh(4,ith)
      end do
#endif


! now, pack results from arrays 1:w4h4 into arrays 1:numhits
! This fills in unused gaps of these arrays to 1:numhits length
! by moving all entries forward.

         if( numhits > 0) then
            do k=1,numhits
               rr(:,k) = rr(:,hit(k))
               cc(:,k) = cc(:,hit(k))
               Q(:,:,k) = Q(:,:,hit(k))   
               tmin(:,k) = tmin(:,hit(k))
               tmax(:,k) = tmax(:,hit(k))
               itpackmin(k) = itpackmin(hit(k))
               itpackmax(k) = itpackmax(hit(k))
               shiftW(k) = shiftW(hit(k))
               im_cc(k) =  im_cc(hit(k)) ! mapper in sampling loop          
               im_rr(k) =  im_rr(hit(k))
#ifdef SREND_TOFF
               toff(:,k) = toff(:,hit(k))
#endif
            end do
         end if ! numhits > 0
         
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif


2666  continue ! place to jump if w1-w0<0, no ray testing was done


#if SREND_TEAM_RENDER==1
!$OMP SINGLE
#endif

! index im by 16 pack: (-3,-2,-1,0,1,2,3,4) -> (0,1)
      wh(1) = ( Bh%W + wh(1) + 3 )/4 - Bh%W/4
      wh(2) = wh(2)/4
      wh(3) = ( wh(3) + 3 )/4
      wh(4) = wh(4)/4

! allocate target rendering array, exit if nothing to render      
      if(image) then !n=0 here, always allocate target to full image extent
         wh(1) = 1; wh(2) = Bh%W/4; wh(3) = 1; wh(4) = Bh%H/4 ! dims must define image in any case
         if( allocated( im ) ) deallocate(im)
         allocate(im(1:16,0:3,wh(1):wh(2),wh(3):wh(4)) )
         im = 0.0_real32 ! zero array
      else if(flush) then !n=0 here
         if(numhits > 0) then
            if( allocated(im)) deallocate(im)
            allocate(im(1:16,0:3,wh(1):wh(2),wh(3):wh(4)) )
            im = 0.0_real32 ! zero array
         end if
         ! don't bother if numhits == 0
      else if( load ) then ! n >= 1
         if(numhits > 0) then
            if( allocated(im)) deallocate(im)
            allocate(im(1:16,0:3,wh(1):wh(2),wh(3):wh(4)) )
            im = 0.0_real32 ! zero array
         end if
         ! don't bother if numhits == 0
      end if ! no dud message data sends
      
      Bh%wh = wh ! set in header


! ssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssss
! sssssssss end setup ssssssssssssssssssssssssssssssssssssssssssssssss

! barrier here for all threads, necessary
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif


#ifdef SREND_TIMING
!$OMP MASTER
      call system_clock(it2)
      print *,'RENDER 16packs: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
!$OMP END MASTER
#endif



!amrbq = 0; amrbqm=0; amrbqf = 0
! RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
! Ray Casting loop : all threads do this
! RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
if( numhits > 0) then


! FFFFFFFFFFFFFFFFFFFFFFFF fluid casting FFFFFFFFFFFFFFFFFFFFFFFFFFFFF
#if SREND_TEAM_RENDER==1
! Here determine compositing work for thread th by 16pack over k
         part = numhits ! number of 16packs to do
         Wpart = part / nth ! number of 16packs per thread, coarse
         Rpart = part - Wpart * nth ! remainder, leftover 16packs
! we determine each th thread's span of 16packs to do: Wpart + (1 or 0)
         th0 = 1 + th*Wpart         !all get this many 16packs
         th1 = 1 + (th+1)*Wpart - 1
         if( th >=  Rpart) then ! add offset was Rpart was done
            th0 = th0 + Rpart
            th1 = th1 + Rpart      
         else ! th < Rpart
            th0 = th0 + th     ! offset done before
            th1 = th1 + th + 1 ! extra offset within
         end if
         

! ccccccccccc for each k cast a 16pack of rays over i and march them through by tt
         do k = th0, th1
#else
         do k=1,numhits
#endif
         
! init results for pack
            ddA = 0.0_real32
            ddR = 0.0_real32
            ddG = 0.0_real32
            ddB = 0.0_real32
            
#ifdef SREND_TOFF
            do tt = itpackmin(k)-1, itpackmax(k) 
#else
            do tt = itpackmin(k), itpackmax(k)
               t = tt*dt 
#endif


#ifdef SREND_EARLY_TERMINATION
               if( minval(ddA) > SREND_EARLY_TERMINATION ) exit ! done with tt do loop
#endif


SREND_VECTORIZE
               do i=1,16 ! VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV

! these are coordinates in grid plane: [offX,offX+N]
#ifdef SREND_TOFF
                  t(i) = dt*tt + toff(i,k)
                  x(i) = t(i)*Q(i,1,k) + Bh%E(1) + .5_real32
                  y(i) = t(i)*Q(i,2,k) + Bh%E(2) + .5_real32
                  z(i) = t(i)*Q(i,3,k) + Bh%E(3) + .5_real32
#ifdef SREND_FLOPS
      flop=flop+11
#endif
#else
                  x(i) = t*Q(i,1,k) + Bh%E(1) + .5_real32
                  y(i) = t*Q(i,2,k) + Bh%E(2) + .5_real32
                  z(i) = t*Q(i,3,k) + Bh%E(3) + .5_real32
#ifdef SREND_FLOPS
      flop=flop+9
#endif
#endif
                  ix(i) = floor(x(i))
                  iy(i) = floor(y(i))
                  iz(i) = floor(z(i))                  

! mask out bad values
                  tmask(i) = 1.0_real32
#ifdef SREND_TOFF
                  if( t(i) <= tmin(i,k)) tmask(i) = 0.0_real32
                  if( t(i) > tmax(i,k)) tmask(i) = 0.0_real32
#else
                  if( t <= tmin(i,k)) tmask(i) = 0.0_real32
                  if( t > tmax(i,k)) tmask(i) = 0.0_real32
#endif
                  if( tmask(i) == 0.0_real32) then
                     ix(i) = Bh%xoff+1 ! exists in data array, does not matter what
                     iy(i) = Bh%yoff+1 ! these are: iNYNp1 = iNY*N + 1
                     iz(i) = Bh%zoff+1
                  end if

! >>> trilinear on color table entry 
                  dx0(i) = x(i) - ix(i)
                  dy0(i) = y(i) - iy(i)
                  dz0(i) = z(i) - iz(i)
                  dx1(i) = 1.0_real32 - dx0(i)
                  dy1(i) = 1.0_real32 - dy0(i)
                  dz1(i) = 1.0_real32 - dz0(i)
#ifdef SREND_FLOPS
      flop=flop+6
#endif


! but ix,iy,iz is in [0:Vdim+1,0:Vdim+1,0:Vdim+1]
#ifndef SREND_WOODWARD_AMR
! data used is in indices [0:XN+1,0:YN+1,0:ZN+1] 
                  ix(i) = ix(i) - Bh%xoff
                  iy(i) = iy(i) - Bh%yoff
                  iz(i) = iz(i) - Bh%zoff
#else
! data is in indices [xoffth:xoffth+XN+1,yoffth:yoffth+YN+1,zoffth:zoffth+ZN+1], so subvolume xoff,yoff,zoff
                  ix(i) = ix(i) - xoff_in
                  iy(i) = iy(i) - yoff_in
                  iz(i) = iz(i) - zoff_in 
#endif
#ifdef SREND_FLOPS
      flop=flop+3
#endif

       

                  
               end do ! vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
               
               
! ---- Woodward AMR -----------------------------------------------------------
#ifdef SREND_WOODWARD_AMR
! wwwww This is the straightforward version looking up in vvoxels, only for all coarse in indxbq
!  dat8(i,1) = ichar( vox(nvar_off + mod(iz(i)+3,4)*16   + mod(iy(i)+3,4)*4   + mod(ix(i)+3,4)   +1, &
!                         (ix(i)+3)/4,   (iy(i)+3)/4,   (iz(i)+3)/4   ) )
#if 0
do i=1,16
  dat8(i,1) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
                         ISHFT(ix(i)+3,-2),   ISHFT(iy(i)+3,-2),   ISHFT(iz(i)+3,-2)   ) )
  dat8(i,2) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
                         ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+3,-2),   ISHFT(iz(i)+3,-2)   ) )
  dat8(i,3) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
                         ISHFT(ix(i)+3,-2),   ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+3,-2)   ) )
  dat8(i,4) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
                         ISHFT(ix(i)+3,-2),   ISHFT(iy(i)+3,-2),   ISHFT(iz(i)+1+3,-2) ) )
  dat8(i,5) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
                         ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+3,-2)   ) )
  dat8(i,6) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
                         ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+3,-2),   ISHFT(iz(i)+1+3,-2) ) )
  dat8(i,7) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
                         ISHFT(ix(i)+3,-2),   ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2) ) )
  dat8(i,8) = ichar( vox(nvar_off + ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
                         ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2) ) )
end do
#endif

! This is the AMR version using coarse, medium, and fine data bqs
do i=1,16
  if( tmask(i) == 0.0_real32) then
    dat8(i,1:8) = 0.0_real32
    CYCLE
  end if
  
  ibq(i) = ISHFT(ix(i)+3,-2)
  jbq(i) = ISHFT(iy(i)+3,-2)
  kbq(i) = ISHFT(iz(i)+3,-2)
!  if( indxbq(1,ISHFT(ix(i)+3,-2), ISHFT(iy(i)+3,-2), ISHFT(iz(i)+3,-2)) > 0) then
  if( indxbq(1,ibq(i), jbq(i), kbq(i)) > 0) then
!    amrbq = amrbq+1
    dat8(i,1) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
			  abs(indxbq(1,ibq(i),              jbq(i),              kbq(i)))   ) )
    dat8(i,2) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
			  abs(indxbq(1,ISHFT(ix(i)+1+3,-2), jbq(i),              kbq(i)))   ) )
    dat8(i,3) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
			  abs(indxbq(1,ibq(i),              ISHFT(iy(i)+1+3,-2), kbq(i)))   ) )
    dat8(i,4) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
			  abs(indxbq(1,ibq(i),              jbq(i),              ISHFT(iz(i)+1+3,-2))) ) )
    dat8(i,5) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
			  abs(indxbq(1,ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), kbq(i)))   ) )
    dat8(i,6) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
			  abs(indxbq(1,ISHFT(ix(i)+1+3,-2), jbq(i),              ISHFT(iz(i)+1+3,-2))) ) )
    dat8(i,7) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
			  abs(indxbq(1,ibq(i),              ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2))) ) )
    dat8(i,8) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
			  abs(indxbq(1,ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2))) ) )
  else
    x(i) = (ix(i) - .5_real32 + dx0(i))*2.0_real32
    y(i) = (iy(i) - .5_real32 + dy0(i))*2.0_real32
    z(i) = (iz(i) - .5_real32 + dz0(i))*2.0_real32
    ix(i) = floor(x(i))
    iy(i) = floor(y(i))
    iz(i) = floor(z(i))
    dx0(i) = x(i) - ix(i)
    dy0(i) = y(i) - iy(i)
    dz0(i) = z(i) - iz(i)

    ibq(i) = ISHFT(ix(i)+3,-2)
    jbq(i) = ISHFT(iy(i)+3,-2)
    kbq(i) = ISHFT(iz(i)+3,-2)
    
    if( indxbqm(1,ibq(i), jbq(i), kbq(i)) > 0) then
      dx1(i) = 1.0_real32 - dx0(i)
      dy1(i) = 1.0_real32 - dy0(i)
      dz1(i) = 1.0_real32 - dz0(i)
      dat8(i,1) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqm(1,ibq(i),              jbq(i),              kbq(i)))   ) )
      dat8(i,2) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqm(1,ISHFT(ix(i)+1+3,-2), jbq(i),              kbq(i)))   ) )
      dat8(i,3) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqm(1,ibq(i),              ISHFT(iy(i)+1+3,-2), kbq(i)))   ) )
      dat8(i,4) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqm(1,ibq(i),              jbq(i),              ISHFT(iz(i)+1+3,-2))) ) )
      dat8(i,5) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqm(1,ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), kbq(i)))   ) )
      dat8(i,6) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqm(1,ISHFT(ix(i)+1+3,-2), jbq(i),              ISHFT(iz(i)+1+3,-2))) ) )
      dat8(i,7) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqm(1,ibq(i),              ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2))) ) )
      dat8(i,8) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqm(1,ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2))) ) )
    else ! must be indxbqf > 0
!      amrbqf = amrbqf + 1
      x(i) = (ix(i) - .5_real32 + dx0(i))*2.0_real32
      y(i) = (iy(i) - .5_real32 + dy0(i))*2.0_real32
      z(i) = (iz(i) - .5_real32 + dz0(i))*2.0_real32
      ix(i) = floor(x(i))
      iy(i) = floor(y(i))
      iz(i) = floor(z(i))
      dx0(i) = x(i) - ix(i)
      dy0(i) = y(i) - iy(i)
      dz0(i) = z(i) - iz(i)

      dx1(i) = 1.0_real32 - dx0(i)
      dy1(i) = 1.0_real32 - dy0(i)
      dz1(i) = 1.0_real32 - dz0(i)
      
      ibq(i) = ISHFT(ix(i)+3,-2)
      jbq(i) = ISHFT(iy(i)+3,-2)
      kbq(i) = ISHFT(iz(i)+3,-2)
      
!     indxbqf(1,ibq(i), jbq(i), kbq(i)) > 0 !! this must be true
      dat8(i,1) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqf(1,ibq(i),              jbq(i),              kbq(i)))   ) )
      dat8(i,2) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqf(1,ISHFT(ix(i)+1+3,-2), jbq(i),              kbq(i)))   ) )
      dat8(i,3) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqf(1,ibq(i),              ISHFT(iy(i)+1+3,-2), kbq(i)))   ) )
      dat8(i,4) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqf(1,ibq(i),              jbq(i),              ISHFT(iz(i)+1+3,-2))) ) )
      dat8(i,5) = ichar( vox(ISHFT(IAND(iz(i)+3,3),4)   + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqf(1,ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), kbq(i)))   ) )
      dat8(i,6) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3,3),2)   + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqf(1,ISHFT(ix(i)+1+3,-2), jbq(i),              ISHFT(iz(i)+1+3,-2))) ) )
      dat8(i,7) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3,3)   +1, &
			    abs(indxbqf(1,ibq(i),              ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2))) ) )
      dat8(i,8) = ichar( vox(ISHFT(IAND(iz(i)+3+1,3),4) + ISHFT(IAND(iy(i)+3+1,3),2) + IAND(ix(i)+3+1,3) +1, &
			    abs(indxbqf(1,ISHFT(ix(i)+1+3,-2), ISHFT(iy(i)+1+3,-2), ISHFT(iz(i)+1+3,-2))) ) )
    end if
  end if
end do

! ---- end Woodward AMR ----------------------------------------------
#endif

#ifndef SREND_WOODWARD_AMR
! ---- NOT Woodward AMR
! obtain and lineup data for SIMD section, probably not useful to vectorize
SREND_VECTORIZE
               do i=1,16
                  dat8(i,1) = ichar(bytedata(ix(i)  ,iy(i)  ,iz(i)  ) )
                  dat8(i,2) = ichar(bytedata(ix(i)+1,iy(i)  ,iz(i)  ) )
                  dat8(i,3) = ichar(bytedata(ix(i)  ,iy(i)+1,iz(i)  ) )
                  dat8(i,4) = ichar(bytedata(ix(i)  ,iy(i)  ,iz(i)+1) )
                  dat8(i,5) = ichar(bytedata(ix(i)+1,iy(i)+1,iz(i)  ) )
                  dat8(i,6) = ichar(bytedata(ix(i)+1,iy(i)  ,iz(i)+1) )
                  dat8(i,7) = ichar(bytedata(ix(i)  ,iy(i)+1,iz(i)+1) )
                  dat8(i,8) = ichar(bytedata(ix(i)+1,iy(i)+1,iz(i)+1) )
               end do
! ---- end NOT Woodward AMR ------------------------------------------
#endif     
      
SREND_VECTORIZE
               do i=1,16 ! VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
! interpolate tval(i) from color table byte values on 8 corner points of cell 
                  tval(i)=                           &
                     dat8(i,1)*dx1(i)*dy1(i)*dz1(i)  &
                   + dat8(i,2)*dx0(i)*dy1(i)*dz1(i)  &
                   + dat8(i,3)*dx1(i)*dy0(i)*dz1(i)  &
                   + dat8(i,4)*dx1(i)*dy1(i)*dz0(i)  &
                   + dat8(i,5)*dx0(i)*dy0(i)*dz1(i)  &
                   + dat8(i,6)*dx0(i)*dy1(i)*dz0(i)  &
                   + dat8(i,7)*dx1(i)*dy0(i)*dz0(i)  &
                   + dat8(i,8)*dx0(i)*dy0(i)*dz0(i)
#ifdef SREND_FLOPS
      flop=flop+31
#endif

! optimize
                  tval(i) = tval(i)*tmask(i) ! will be 0 if the sample was outside of data array
                  if(tval(i) >= 255.0_real32) tval(i) = 249.999_real32 ! correction, ugly, but OK
                  cotabi(i) = floor(tval(i))
                  tval(i) = tval(i) - cotabi(i) ! reuse tval() var, the higher cotab value portion
                  q1Mtval(i) = 1.0_real32 - tval(i)    ! the lower cotab value portion
#ifdef SREND_FLOPS
      flop=flop+3
#endif
               end do ! vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv

! not SIMD here: interpolate RGBA values from color table cotab()
               do i=1,16
                  dR(i) =   (cotab(cotabi(i),1)*(q1Mtval(i)) + cotab(cotabi(i)+1,1)*tval(i) )
                  dG(i) =   (cotab(cotabi(i),2)*(q1Mtval(i)) + cotab(cotabi(i)+1,2)*tval(i) )
                  dB(i) =   (cotab(cotabi(i),3)*(q1Mtval(i)) + cotab(cotabi(i)+1,3)*tval(i) )
                  dA(i) = ( (cotab(cotabi(i),0)*(q1Mtval(i)) + cotab(cotabi(i)+1,0)*tval(i) ) )*tmask(i)
#ifdef SREND_FLOPS
      flop=flop+13
#endif
               end do

! ------ composite -----------------

!     front2back : eye to out
SREND_VECTORIZE
	      do i=1,16 ! VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
		  transparency(i) = 1.0_real32 - ddA(i)
		  ddR(i) = transparency(i)*dR(i)*dA(i) + ddR(i) ! replaced as these are premultiplied by alpha above
		  ddG(i) = transparency(i)*dG(i)*dA(i) + ddG(i)
		  ddB(i) = transparency(i)*dB(i)*dA(i) + ddB(i)
		  ddA(i) = transparency(i)*dA(i)       + ddA(i)
#ifdef SREND_FLOPS
      flop=flop+11
#endif
	      end do ! i : vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
	      
	      
! ---- end composite --------------- 

#ifdef SREND_SOLID_RENDERING
SREND_VECTORIZE
do i=1,16
  ixx(i) = floor(x(i)+.4999_real32)
  iyy(i) = floor(y(i)+.4999_real32)
  izz(i) = floor(z(i)+.4999_real32)
end do

if( minval(ixx) <= Sd(1,1) .AND. maxval(ixx) >= Sd(0,1) .AND. &
    minval(iyy) <= Sd(1,2) .AND. maxval(iyy) >= Sd(0,2) .AND. &
    minval(izz) <= Sd(1,3) .AND. maxval(izz) >= Sd(0,3) ) then
SREND_VECTORIZE
    do i=1,16
      if(ixx(i)< Sd(0,1) .OR. ixx(i)>Sd(1,1) .OR. &
         iyy(i)< Sd(0,2) .OR. iyy(i)>Sd(1,2) .OR. &
         izz(i)< Sd(0,3) .OR. izz(i)>Sd(1,3) ) then
	    ixx(i) = Sd(0,1)
	    iyy(i) = Sd(0,2)
	    izz(i) = Sd(0,3)
	    tmask(i) = 0.0_real32 ! if tmask is already 0, then a miss
      end if
    end do
SREND_VECTORIZE
    do i=1,16
	    dA(i) = tmask(i) * iachar(S(ixx(i),iyy(i),izz(i))) / 255.0_real32 ! always executed
	    dR(i)     = tmask(i) * iachar( Srgb(1,ixx(i),iyy(i),izz(i)) )/255.0_real32
            dG(i)     = tmask(i) * iachar( Srgb(2,ixx(i),iyy(i),izz(i)) )/255.0_real32
            dB(i)     = tmask(i) * iachar( Srgb(3,ixx(i),iyy(i),izz(i)) )/255.0_real32
    end do
SREND_VECTORIZE
    do i=1,16
	    transparency(i) = 1.0_real32 - ddA(i)
	    ddR(i) = transparency(i)*dR(i)*dA(i) + ddR(i) ! replaced, are premultiplied by alpha above
	    ddG(i) = transparency(i)*dG(i)*dA(i) + ddG(i)
	    ddB(i) = transparency(i)*dB(i)*dA(i) + ddB(i)
	    ddA(i) = transparency(i)*dA(i)       + ddA(i)
    end do ! solid rendering loop i
end if ! at least for one i within Sbounds
#endif

            end do ! tt : tttttttttttttttttttttttttttttttttttttttttttttttttt
            

! all map to same 4x4 16-pack, so use the first one      
!      im_cc(k) =  (Bh%W + cc(1,k) - shiftW(k) + 3)/4 - Bh%W/4
!      im_rr(k) =  (rr(1,k) + 3)/4
            im(:,0,im_cc(k),im_rr(k)) = ddA
            im(:,1,im_cc(k),im_rr(k)) = ddR
            im(:,2,im_cc(k),im_rr(k)) = ddG
            im(:,3,im_cc(k),im_rr(k)) = ddB

         end do ! k : over 16 packs of rays

! FFFFFFFFFFFFFFFFFFFFFFFF end: fluid casting FFFFFFFFFFFFFFFFFFFFFFFFFFFFF
   
! rrrrrrrrrrrrrr end ray casting lop rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
! rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
end if ! numhits > 0
! all threads must reach this so all data is in srendVC(nV)%im
#if SREND_TEAM_RENDER==1
!$OMP BARRIER
#endif

#ifdef SREND_FLOPS
      print *,"PPPPPPPPPPPP FLOPS=",flop,"th=",th
#endif

#ifdef SREND_TIMING
!$OMP MASTER
      call system_clock(it2)
      print *,'RENDER sampled: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
!$OMP END MASTER
#endif


#if SREND_TEAM_RENDER==1
!$OMP SINGLE
#endif
      Bh%E = E_AMR_PASS  ! divided by Vdim, adjusted back in compositing; set Bheader
      Bh%leader = 'CHAD' ! tag for CHAD/SCHAD
      Bh%header_len = SREND_BHEADER_LEN
      Bh%nchad = 1   ! number of packages, just this one for render_flush
      Bh%bit16 = 0
      Bh%RLE0 = 0
      Bh%plen = 0
      if( numhits>0) Bh%plen = sizeof( im )
      if( image ) then !n=0
         call srend_write_image(Bh,im,mem,memlen) ! use Bheader and im
#if defined(SREND_IM_AUTO_DEALLOCATION) && SREND_TEAM_RENDER==1
         deallocate( im )
#endif
      else if(load) then !n >= 1
!$OMP CRITICAL (B_structure)
         B(nV)%n = B(nV)%n + 1
#ifdef SREND_B_DYNAMIC_LOAD
         i = size(B(nV)%cube)
         if(B(nV)%n > i) then
            j = max(B(nV)%n,2*i,B_max_LOAD)
            allocate(Bh_temp(1:j))
            allocate(cube_temp(1:j))
            Bh_temp(1:i) = B(nV)%Bh
            cube_temp(1:i) = B(nV)%cube
            B_max_LOAD = j
            call move_alloc(Bh_temp,B(nV)%Bh)
            call move_alloc(cube_temp,B(nV)%cube)
         end if
#endif
         B(nV)%Bh(B(nV)%n) = Bh ! load header, im already there
         if(numhits>0) call move_alloc(im, B(nV)%cube(B(nV)%n)%im ) ! im is deallocated
!$OMP END CRITICAL (B_structure)
         memlen = Bh%plen+SREND_BHEADER_LEN
      else ! flush: Bheader+im(0) !n=0
#ifdef SREND_16BIT
         Bh%bit16 = 1
#ifdef SREND_RLE0
         Bh%RLE0 = 1
#endif
         if(numhits>0) then
            call srend_pack_16bit(Bh%RLE0,     im,                          &
                                       Bh%wh(1),Bh%wh(2),Bh%wh(3),Bh%wh(4), &
                                       mem(SREND_BHEADER_LEN+1:),memlen)
            Bh%plen = memlen
            memlen = memlen + SREND_BHEADER_LEN
            Bh%tlen = memlen
         else
            Bh%plen = 0
            memlen = SREND_BHEADER_LEN
            Bh%tlen = memlen
         end if
         mem(1:SREND_BHEADER_LEN) = transfer(Bh,mem(1:SREND_BHEADER_LEN))
#else
         ii = SREND_BHEADER_LEN
         if(numhits>0) then
            do ir=wh(3),wh(4)
            do ic=wh(1),wh(2)
              real4_cast256 = im(:,:,ic,ir) ! real4_cast <-- equiv --> char1_cast
              mem(ii+1:ii+256) = transfer(real4_cast256,mem(ii+1:ii+256))
              ii = ii+256 ! memlen
            end do
            end do
         end if
         memlen = ii
         Bh%tlen = memlen
         Bh%plen = Bh%tlen - SREND_BHEADER_LEN
         mem(1:SREND_BHEADER_LEN) = transfer(Bh,mem(1:SREND_BHEADER_LEN))
#endif
#if defined(SREND_IM_AUTO_DEALLOCATION) && SREND_TEAM_RENDER==1
         if(allocated(im)) deallocate( im )
#endif
      end if
#if SREND_TEAM_RENDER==1
!$OMP END SINGLE
#endif
! implicit barrier after SINGLE

#ifdef SREND_TIMING
!$OMP MASTER
      call system_clock(it2)
      print *,'RENDER DONE (flush, write): span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
!$OMP END MASTER
#endif


      return ! nothing left to do

! rrrrrrrrrrrrrrrr end rendering :ray casting rrrrrrrrrrrrrrrrrrrrrrrr


      CONTAINS
      

! ********************************************************************
! numerical recipes heapsort, just for integer(int32); must have n>1
   recursive SUBROUTINE srend_heapsort(N,RA)
      implicit none
      integer(int32),intent(INOUT) :: N, RA(N)
      integer(int32) :: L,RRA,IR,I,J
! ----------------
      L=N/2+1
      IR=N
      !The index L will be decremented from its initial value during the
      !"hiring" (heap creation) phase. Once it reaches 1, the index IR 
      !will be decremented from its initial value down to 1 during the
      !"retirement-and-promotion" (heap selection) phase.
4077  continue
      if(L > 1)then
         L=L-1
         RRA=RA(L)
      else
         RRA=RA(IR)
         RA(IR)=RA(1)
         IR=IR-1
         if(IR==1)then
            RA(1)=RRA
            return
         end if
      end if
      I=L
      J=L+L
4777  if(J<=IR)then
         if(J < IR)then
            if(RA(J) < RA(J+1))  J=J+1
         end if
         if(RRA < RA(J))then
            RA(I)=RA(J)
            I=J
            J=J+J
         else
            J=IR+1
         end if
         goto 4777
      end if
      RA(I)=RRA
      goto 4077
   END subroutine srend_heapsort
! ********************************************************************
      
      
! ********************************************************************
   recursive subroutine srend_set_cotab()
      implicit NONE

! color table data : to form color table : these are done once for all
! these are for hard-coded color table values, adjust as needed

#ifdef SREND_COTAB_HARDCODED
      character(LEN=64)              :: alpha_knot(1:SREND_COTAB_KEYS)
      character(LEN=64)              :: rgb_knot(1:SREND_COTAB_KEYS)
#endif

! IO : vars to read in color table knots and fill out color table
      integer(int32)                 :: cotab_pix,cotab_pix2
      real(real32)                   :: cotab_a,cotab_a2,cotab_r,cotab_r2
      real(real32)                   :: cotab_g,cotab_g2,cotab_b,cotab_b2
      
! These are for parsing color table file to form color table
#ifdef SREND_COTAB_FILENAME
      character (len=300)            :: cotab_line ! max length, way much here
      integer(int32)                 :: cotab_pos
      integer(int32)                 :: cotab_ferror ! test on read of file
      integer                        :: MYunit
#endif

! --------------------------------------------------------------------
! --------------------------------------------------------------------
! only first thread does color table setup on first pass then
! stores in B(nV)%cotab, then only first thread moves this stored
! color table back into cotab for following threads on later
! passes 
!
! It is known that an improper color table WILL break this code, but it
! will break on first pass in that case.  Indices must start at 0 and 
! end at 255 to build all 256 indices.  The indices cannot repeat, and 
! they must increase.

! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! color table setup 
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

#ifdef SREND_COTAB_HARDCODED
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! hard-coded color table for renderer; read in once
! offset_cotab : where to start reading tables
! the table reading/forming ends with a '255 ...' entry
! -------- write alpha knots here -------     
! LCSE original
      alpha_knot(1) = '0 0.0'
      alpha_knot(2) = '5 0.0'
      alpha_knot(3) = '36 0.0627'
      alpha_knot(4) = '72 0.148'
      alpha_knot(5) = '105 0.536'
      alpha_knot(6) = '133 1.0'
      alpha_knot(7) = '158 1.0'
      alpha_knot(8) = '196 0.328'
      alpha_knot(9) = '236 0.302'
      alpha_knot(10) = '244 0.105'
      alpha_knot(11) = '250 0.0'
      alpha_knot(12) = '255 0.0'
         rgb_knot(1) = '0 0.0 0.0 0.0'
         rgb_knot(2) = '5 0.0 0.0 0.0'
         rgb_knot(3) = '36 0.0 0.0 0.2509804'
         rgb_knot(4) = '71 0.0 0.1176471 0.3137255'
         rgb_knot(5) = '105 0.0 0.7843137 1.0'
         rgb_knot(6) = '131 1.0 1.0 1.0'
         rgb_knot(7) = '158 1.0 1.0 0.0'
         rgb_knot(8) = '196 1.0 0.0 0.0'
         rgb_knot(9) = '244 0.5019608 0.0 0.0'
         rgb_knot(10) = '255 0.5019608 0.0 0.0'
! linear blue-red
      alpha_knot(24) = '0 1.0'
      alpha_knot(25) = '255 1.0'
         rgb_knot(24) = '0 0.0 0.0 1.0'
         rgb_knot(25) = '255 1.0 0.0 0.0'
! upscale: for testing
      alpha_knot(30) = '0   0.0'
      alpha_knot(31) = '1   0.001'
      alpha_knot(32) = '2   0.01'
      alpha_knot(33) = '4   0.1'
      alpha_knot(34) = '8   0.2'
      alpha_knot(35) = '32  0.4'
      alpha_knot(36) = '64  0.5'
      alpha_knot(37) = '128 0.6'
      alpha_knot(38) = '192 0.7'
      alpha_knot(39) = '224 0.8'
      alpha_knot(40) = '255 1.0'
         rgb_knot(30) = '0   1.0 0.0 0.0'
         rgb_knot(31) = '32  0.5 0.5 0.0'
         rgb_knot(32) = '64  0.0 1.0 0.0'
         rgb_knot(33) = '128 0.0 0.5 0.5'
         rgb_knot(34) = '192 0.5 0.5 1.0'
         rgb_knot(35) = '255 0.9 0.9 1.0'
! red up, blue down
      alpha_knot(80) = '0 1.0'
      alpha_knot(81) = '32 0.2'
      alpha_knot(82) = '64 0.1'
      alpha_knot(83) = '128 0.0'
      alpha_knot(84) = '192 0.1'
      alpha_knot(85) = '224 0.2'
      alpha_knot(86) = '255 1.0'
         rgb_knot(80) = '0 0.5 0.5 1.0'
         rgb_knot(81) = '32 0.5 0.5 1.0'
         rgb_knot(82) = '64 0.2 0.2 1.0'
         rgb_knot(83) = '128 0.0 0.0 0.0'
         rgb_knot(84) = '192 1.0 0.2 0.2'
         rgb_knot(85) = '255 1.0 0.5 0.5'
! Woodward's Star AMR code lut: entropy
      alpha_knot(110) = '0 0.0'
      alpha_knot(111) = '18 0.1058824'
      alpha_knot(112) = '97 0.1294118'
      alpha_knot(113) = '131 0.2588235'
      alpha_knot(114) = '157 0.2156863'
      alpha_knot(115) = '165 0.0'
      alpha_knot(116) = '171 0.3803922'
      alpha_knot(117) = '184 0.5490196'
      alpha_knot(118) = '203 0.454902'
      alpha_knot(119) = '254 0.1686275'
      alpha_knot(120) = '255 0.1254902'
         rgb_knot(110) = '0 0.0 0.0 0.0'
         rgb_knot(111) = '18 0.0 0.0 0.2509804'
         rgb_knot(112) = '149 0.0 0.2352941 0.627451'
         rgb_knot(113) = '154 0.0 0.7843137 1.0'
         rgb_knot(114) = '159 1.0 1.0 1.0'
         rgb_knot(115) = '165 1.0 1.0 0.0'
         rgb_knot(116) = '175 1.0 0.0 0.0'
         rgb_knot(117) = '231 0.5019608 0.0 0.0'
         rgb_knot(118) = '255 0.5019608 0.0 0.0'      
! Woodward's Star AMR code lut: NegDivU
      alpha_knot(121) = '0 0.0'
      alpha_knot(122) = '7 0.254902'
      alpha_knot(123) = '61 0.3058824'
      alpha_knot(124) = '108 0.6078432'
      alpha_knot(125) = '135 0.4705882'
      alpha_knot(126) = '159 0.3098039'
      alpha_knot(127) = '192 0.1921569'
      alpha_knot(128) = '202 0.1333333'
      alpha_knot(129) = '254 0.09411765'
      alpha_knot(130) = '255 0.1254902'
         rgb_knot(121) = '0 0.0 0.0 0.0'
         rgb_knot(122) = '12 0.03921569 0.0 0.0'
         rgb_knot(123) = '26 1.0 0.0 0.0'
         rgb_knot(124) = '41 1.0 0.2509804 0.0'
         rgb_knot(125) = '79 1.0 1.0 0.0'
         rgb_knot(126) = '106 1.0 1.0 0.5019608'
         rgb_knot(127) = '131 1.0 1.0 1.0'
         rgb_knot(128) = '181 0.5019608 0.5019608 1.0'
         rgb_knot(129) = '195 0.0 0.0 1.0'
         rgb_knot(130) = '212 0.0 0.0 0.5019608'
         rgb_knot(131) = '255 0.5019608 0.0 0.0'
! Woodward's Star AMR code lut: FV demoSC08
      alpha_knot(132) = '0 0.0'
      alpha_knot(133) = '18 0.1058824'
      alpha_knot(134) = '96 0.2745098'
      alpha_knot(135) = '115 0.7843137'
      alpha_knot(136) = '133 1.0'
      alpha_knot(137) = '158 1.0'
      alpha_knot(138) = '184 0.5490196'
      alpha_knot(139) = '203 0.454902'
      alpha_knot(140) = '254 0.1686275'
      alpha_knot(141) = '255 0.1254902'
         rgb_knot(132) = '0 0.0 0.0 0.0'
         rgb_knot(133) = '18 0.0 0.0 0.2509804'
         rgb_knot(134) = '77 0.0 0.2352941 0.627451'
         rgb_knot(135) = '108 0.0 0.7843137 1.0'
         rgb_knot(136) = '131 1.0 1.0 1.0'
         rgb_knot(137) = '158 1.0 1.0 0.0'
         rgb_knot(138) = '196 1.0 0.0 0.0'
         rgb_knot(139) = '244 0.5019608 0.0 0.0'
         rgb_knot(140) = '255 0.5019608 0.0 0.0'
! Woodward's Star AMR code lut: Ur
      alpha_knot(142) = '0 1.0'
      alpha_knot(143) = '36 1.0'
      alpha_knot(144) = '70 0.372549'
      alpha_knot(145) = '107 0.254902'
      alpha_knot(146) = '128 0.05882353'
      alpha_knot(147) = '151 0.2705882'
      alpha_knot(148) = '194 0.5450981'
      alpha_knot(149) = '240 1.0'
      alpha_knot(150) = '255 1.0'
         rgb_knot(142) = '0 1.0 1.0 1.0'
         rgb_knot(143) = '76 0.0 1.0 1.0'
         rgb_knot(144) = '90 0.0 0.0 1.0'
         rgb_knot(145) = '103 0.0 0.0 0.5019608'
         rgb_knot(146) = '118 0.0 0.0 0.0'
         rgb_knot(147) = '128 0.25 0.25 0.25'
         rgb_knot(148) = '144 0.4156863 0.3568628 0.3568628'
         rgb_knot(149) = '163 1.0 0.0 0.0'
         rgb_knot(150) = '183 1.0 1.0 0.0'
         rgb_knot(151) = '208 1.0 1.0 0.0'
         rgb_knot(152) = '255 1.0 1.0 0.0'
! Woodward's Star AMR code lut: Vort
      alpha_knot(153) = '0 0.0'
      alpha_knot(154) = '8 0.03137255'
      alpha_knot(155) = '20 0.0627451'
      alpha_knot(156) = '32 0.7843137'
      alpha_knot(157) = '133 1.0'
      alpha_knot(158) = '153 0.8941177'
      alpha_knot(159) = '197 0.2431373'
      alpha_knot(160) = '241 0.2588235'
      alpha_knot(161) = '255 0.0'
         rgb_knot(153) = '0 0.0 0.0 0.0'
         rgb_knot(154) = '24 0.0 0.0 0.3921569'
         rgb_knot(155) = '41 0.0 0.5019608 0.5019608'
         rgb_knot(156) = '81 0.0 1.0 1.0'
         rgb_knot(157) = '115 1.0 1.0 1.0'
         rgb_knot(158) = '153 1.0 1.0 0.0'
         rgb_knot(159) = '197 1.0 0.0 0.0'
         rgb_knot(160) = '238 0.2509804 0.0 0.0'
         rgb_knot(161) = '255 0.0 0.0 0.0'
! all white blue
      alpha_knot(170) =   '0 1.0'
      alpha_knot(171) = '255 1.0'
         rgb_knot(170) =   '0 1. 1. 1.'
         rgb_knot(171) = '255 1. 1. 1.'
#endif


#if defined(SREND_COTAB_PASSNAMELIST) || defined(SREND_COTAB_HARDCODED)
! -------- create color tables -------------------

! read alpha knots and form table
      j = icotab
      read (alpha_knot(1),*) cotab_pix, cotab_a ! this must be 0
      cotab(cotab_pix,0) = cotab_a
5077    j = j+1
      read (alpha_knot(j),*) cotab_pix2, cotab_a2
      do i=cotab_pix+1,cotab_pix2
         cotab(i,0) =                                            &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_a        &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_a2
      end do
      cotab_pix = cotab_pix2
      cotab_a = cotab_a2
      if(cotab_pix /= 255) goto 5077 ! must have last line 255 # # #

! read RGB knots and form table
      j = icotab
      read (rgb_knot(j),*) cotab_pix, cotab_r, cotab_g, cotab_b ! 0
      cotab(cotab_pix,1) = cotab_r
      cotab(cotab_pix,2) = cotab_g
      cotab(cotab_pix,3) = cotab_b
5777    j = j + 1
      read (rgb_knot(j),*) cotab_pix2, cotab_r2, cotab_g2, cotab_b2
      do i=cotab_pix+1,cotab_pix2
         cotab(i,1) =                                             &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_r         &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_r2
         cotab(i,2) =                                             &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_g         &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_g2
         cotab(i,3) =                                             &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_b         &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_b2
      end do
      cotab_pix = cotab_pix2
      cotab_r = cotab_r2
      cotab_g = cotab_g2
      cotab_b = cotab_b2
      if(cotab_pix /= 255) goto 5777 ! must have last line 255 # # #

! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
#endif


#ifdef SREND_COTAB_FILENAME
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! read color table from: cotab.txt : example below for 
! Bh%nR=1, cotab_offset = (/8/) as passed to srend_render():
! ------------------
!  alpha 8   
!  0 0.0
!  128 1.0 1.0
!  255 1.0 1.0
!
!  rgb 8
!  0 0.0 0.0 0.0
!  128 1.0 1.0 1.0
!  255 1.0 1.0 1.0
! ------------------
      open(newunit=MYunit,file=SREND_COTAB_FILENAME,status='old', action='read')

! read alpha knots and form table
      rewind MYunit ! here for next Bh%nR if more than 1
5007  read(MYunit,'(A)') cotab_line
      if( cotab_line(1:5) == 'alpha') then
         read(cotab_line(6:300),*) cotab_pos
         if(cotab_pos /= icotab) goto 1001
      else
         goto 5007
      end if

      read (MYunit,*) cotab_pix, cotab_a ! cotab_pix must be 0
      cotab(cotab_pix,0) = cotab_a
5077    read (MYunit,*) cotab_pix2, cotab_a2
         do i=cotab_pix+1,cotab_pix2
      cotab(i,0) =                                        &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_a  &
         + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_a2
      end do
      cotab_pix = cotab_pix2
      cotab_a = cotab_a2
      if(cotab_pix /= 255) goto 5077 ! must have last line 255 # # #

! read RGB knots and form table
      rewind MYunit
5777  read(MYunit,'(A)') cotab_line
      if( cotab_line(1:3) == 'rgb') then
         read(cotab_line(4:300),*) cotab_pos
         if(cotab_pos /= icotab) goto 1002
      else
         goto 5777
      end if

      read (MYunit,*) cotab_pix, cotab_r, cotab_g, cotab_b ! 0
      cotab(cotab_pix,1) = cotab_r
      cotab(cotab_pix,2) = cotab_g
      cotab(cotab_pix,3) = cotab_b
5778    read (MYunit,*) cotab_pix2, cotab_r2, cotab_g2, cotab_b2
      do i=cotab_pix+1,cotab_pix2
         cotab(i,1) =                                             &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_r         &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_r2
         cotab(i,2) =                                             &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_g         &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_g2
         cotab(i,3) =                                             &
         1.0_real32*(cotab_pix2-i)/(cotab_pix2-cotab_pix)*cotab_b         &
               + 1.0_real32*(i-cotab_pix)/(cotab_pix2-cotab_pix)*cotab_b2
      end do
      cotab_pix = cotab_pix2
      cotab_r = cotab_r2
      cotab_g = cotab_g2
      cotab_b = cotab_b2
      if(cotab_pix /= 255) goto 5778 ! must have last line 255 # # #

      close(MYunit)
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
#endif

#ifndef SREND_COTAB_REBUILD_EACH_PASS
      if( .NOT. allocated( B(nV)%cotab) ) allocate( B(nV)%cotab(0:255,0:3) )
      B(nV)%cotab = cotab ! keep in cache, for AMR also Valpha which this table is scaled for
#endif

   end subroutine srend_set_cotab
! ********************************************************************


   end subroutine srend_render ! Srend_render subroutine
! rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
! rrrrrrr end: render rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
! rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr



! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! ccccccccccc compose cccccccccccccccccccccccccccccccccccccccccccccccc
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
   subroutine  srend_compose_base( SrendType, nV, mem, memlen)

      implicit NONE  

      integer(int32)                :: SrendType
      integer(int32)                :: nV
      character,dimension(:)        :: mem
      integer(int32),intent(OUT)    :: memlen

! counters
      integer(int32)                                   :: i,j,k, kc,ii,kim
#ifndef SREND_16BIT
      integer(int32)                                   :: ir, ic
#endif
      integer(int32)                                   :: xoff1,yoff1,zoff1 ! to calc offset and NX,NY,NZ after composite
      integer(int32)                                   :: xoff0,yoff0,zoff0
      
      real(real32), dimension(1:3)                     :: E_AMR_PASS
      integer(int32)                                   :: vdim_max, vdim_temp
      
! ------- target image array
      real(real32),dimension(:,:,:,:),allocatable      :: im
!DIR$ ATTRIBUTES ALIGN:64 :: im

! ------ sorting -----------------------------------------------------
      integer(int32)                                   :: NS  ! number entries
      integer(int32),dimension(1:16,0:B_max_LOAD)      :: wh  ! hold essential data from header
!DIR$ ATTRIBUTES ALIGN:64 :: wh
      integer(int32),dimension(1:B_max_LOAD)           :: L   ! list for sorting
      logical,dimension(0:2,0:2,0:2,0:B_max_LOAD)      :: oct ! T if a corner in octant, F otherwise
#ifdef SREND_OCTREE_SORT
      logical                                          :: n2sort = .TRUE. ! default, works if can be sorted
      integer(int32)                                   :: min_n, max_n, TreeTop, max_off, TreeTop_max
      integer(int32),allocatable                       :: Tree(:,:), Tree_temp(:,:)
!DIR$ ATTRIBUTES ALIGN:64 :: Tree, Tree_temp
      integer(int32)                                   :: ix, iy, iz
      integer(int32)                                   :: Tstack(0:31),Kstack(0:31),Pstack
!DIR$ ATTRIBUTES ALIGN:64 :: Tstack, Kstack
      integer(int32)                                   :: jj,kk
#endif
#ifdef SREND_COMPOSE_TIMING
      integer(int64)                                   :: it1, it2, itc, it_start ! for testing .. collect statistics
      real(real32)                                     :: rc
#endif
#ifndef SREND_16BIT
      real(real32),dimension(1:64)                     :: real4_cast256 ! vectors of 16*4 reals --> 64*4 char*1
#endif      
! ------ behavior ----------------------------------------------------
   logical                                             :: load   ! set to T for a load call
   logical                                             :: flush  ! set to T for a flush call
   logical                                             :: image  ! set to T for flush2image

   type(B_header_type)                                 :: Bh

!  declarations done -------------------------------------------------

! set behavior
      image = .FALSE. ; flush = .FALSE. ; load = .FALSE.
      if( srendType < 0 ) image = .TRUE.
      if( srendType ==0 ) flush = .TRUE.
!      if( srendType > 0 ) load  = .TRUE. ! this will not happen

#ifdef SREND_COMPOSE_TIMING
      call system_clock(it1,itc); rc = 1.0_real32 / itc
      it_start = it1
#endif      

! SSSSS Setup SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
! n>= 1, read in Bheader 1:n

! find maximum Vdim, scale all dimensions up to this max in next loop
      Vdim_max = 1
      do ii=1,B(nV)%n
         Bh = B(nV)%Bh(ii) ! load header to view data
         if(Bh%Vdim > Vdim_max) Vdim_max = Bh%Vdim
      end do

      xoff0 =   1000000000
      yoff0 =   1000000000
      zoff0 =   1000000000
      xoff1 =  -1000000000
      yoff1 =  -1000000000
      zoff1 =  -1000000000
      
      wh(1,0)  =   1000000000
      wh(2,0)  =  -1000000000
      wh(3,0)  =   1000000000
      wh(4,0)  =  -1000000000
      
      NS    = 0
      i     = 0
      do ii=1,B(nV)%n ! get data from headers

         Bh = B(nV)%Bh(ii) ! load header to view data
         
         Vdim_temp = Vdim_max / Bh%Vdim ! scale up for common units in order to sort
         Bh%xoff = Bh%xoff  * Vdim_temp ! xoff
         Bh%yoff = Bh%yoff  * Vdim_temp ! yoff
         Bh%zoff = Bh%zoff  * Vdim_temp ! zoff
         Bh%XN = Bh%XN  * Vdim_temp ! XN
         Bh%YN = Bh%YN  * Vdim_temp ! YN
         Bh%ZN = Bh%ZN  * Vdim_temp ! ZN
         
         if(Bh%xoff < xoff0) xoff0 = Bh%xoff ! lower offset
         if(Bh%yoff < yoff0) yoff0 = Bh%yoff
         if(Bh%zoff < zoff0) zoff0 = Bh%zoff
         
         if(Bh%xoff+Bh%XN > xoff1) xoff1 = Bh%xoff+Bh%XN ! upper offset
         if(Bh%yoff+Bh%YN > yoff1) yoff1 = Bh%yoff+Bh%YN
         if(Bh%zoff+Bh%ZN > zoff1) zoff1 = Bh%zoff+Bh%ZN
         
         
         if(Bh%wh(2)-Bh%wh(1) < 0) cycle ! a dud
         
         i = i+1
         NS = i ! data

         wh(1,i)  = Bh%wh(1) ! w0
         wh(2,i)  = Bh%wh(2) ! w1
         wh(3,i)  = Bh%wh(3) ! h0
         wh(4,i)  = Bh%wh(4) ! h1
         
         wh(5,i)  = Bh%xoff
         wh(6,i)  = Bh%yoff
         wh(7,i)  = Bh%zoff
         
         wh(8,i)  = Bh%XN
         wh(9,i)  = Bh%YN
         wh(10,i) = Bh%ZN

! Determine far corner offsets
         wh(11,i) = wh(5,i) + wh(8,i)  !xoff+XN , far faces
         wh(12,i) = wh(6,i) + wh(9,i)  !yoff+YN
         wh(13,i) = wh(7,i) + wh(10,i) !zoff+ZN
!        wh(14,i) ! unused
         wh(15,i) = ii ! tag for image
!        wh(16,i) !unused

         if(wh(1,i) < wh(1,0)) wh(1,0) = wh(1,i) ! find bounds of target plane
         if(wh(2,i) > wh(2,0)) wh(2,0) = wh(2,i)
         if(wh(3,i) < wh(3,0)) wh(3,0) = wh(3,i)
         if(wh(4,i) > wh(4,0)) wh(4,0) = wh(4,i)
      end do
      
! set for CHAD
      Bh%Vdim = Vdim_max ! load this for passing up
      !print *,"Bh%Vdim=",Bh%Vdim 

      Bh%xoff = xoff0
      Bh%yoff = yoff0
      Bh%zoff = zoff0
      Bh%XN = xoff1-xoff0
      Bh%YN = yoff1-yoff0
      Bh%ZN = zoff1-zoff0

! adjust E, just needed for sorting    
      E_AMR_PASS = Bh%E ! save unaltered for passing up, if needed
      Bh%E = Bh%E * Vdim_max ! expand for sorting
      
#ifdef SREND_COMPOSE_TIMING
      call system_clock(it2)
      print *,'COMPOSE read Bheader: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
#endif


! ----------- octree sort ---- attempt, n2 sort if conditions fail ---
#ifdef SREND_OCTREE_SORT
      n2sort = .FALSE. ! test if can oct-tree sort, else n2sort = .TRUE.
      min_n = 1000000000; max_n = 1
      do i=1,NS
         if( wh(8,i) /= wh(9,i) .OR. wh(8,i) /= wh(10,i) ) then ! ensure cube: XN=YN=ZN
            n2sort = .TRUE.
            EXIT
         end if
         if(  mod(wh(5,i),wh(8,i)) /= 0 & ! ensure offset is multiple of cube dimension XN=YN=ZN
         .OR. mod(wh(6,i),wh(8,i)) /= 0 & 
         .OR. mod(wh(7,i),wh(8,i)) /= 0 ) then
            n2sort = .TRUE.
            EXIT
         end if
         if( wh(8,i) < min_n) min_n = wh(8,i) ! finid minimum cube dim
         if( wh(8,i) > max_n) max_n = wh(8,i) ! find max cube dim
      end do
      if(n2sort) goto 6006
      
      do i=1,NS ! ensure that all block dimensions are multiples of min_n
         j = wh(8,i)/min_n
         if( mod(wh(8,i),min_n) /= 0 .OR. IAND(j,j-1)/=0) then ! !(XN is power of 2 multiple of min_n)
            n2sort = .TRUE.
            EXIT
         end if
      end do
      if(n2sort) goto 6006
      
! load tree with nodes 1:NS
! prepare tree
      TreeTop_max = 32+4*B(nV)%n ! a reasonable upper bound
      allocate( Tree(0:15,0:TreeTop_max) ) ! a sufficient estimate
      TreeTop = 0 ! index to top active index
      Tree    = 0 ! init
! set tree root
      Tree(9,0) = min_n
      max_off = max(xoff1-xoff0,yoff1-yoff0,zoff1-zoff0)
      do while( Tree(9,0) < max_off) ! set span of tree node
         Tree(9,0) = 2*Tree(9,0)
      end do
      Tree(10,0) = xoff0 ! offset of root
      Tree(11,0) = yoff0
      Tree(12,0) = zoff0
! insert items into tree
      do i=1,NS
         j = 0 ! start at root
6007     continue
         if(wh(8,i) == Tree(9,j)) then ! insert here
            Tree(8,j) = i ! other info not needed
            CYCLE
         end if
         ix = 0; iy = 0; iz = 0
         if( wh(5,i) >= Tree(10,j)+Tree(9,j)/2 ) ix = 1
         if( wh(6,i) >= Tree(11,j)+Tree(9,j)/2 ) iy = 1
         if( wh(7,i) >= Tree(12,j)+Tree(9,j)/2 ) iz = 1
         k = ix + 2*iy + 4*iz
         if( Tree(k,j) == 0) then ! insert new item
            TreeTop = TreeTop + 1 ! new node
            if(TreeTop > TreeTop_max) then ! resize by doubling
               allocate(Tree_temp(0:15,0:2*TreeTop_max))
               Tree_temp = 0
               Tree_temp(:,0:TreeTop_max) = Tree
               TreeTop_max = 2*TreeTop_max
               call move_alloc(Tree_temp,Tree)
            end if
            Tree(k,j) = TreeTop   ! link to new node
            Tree( 9,TreeTop) = Tree( 9,j)/2 ! width of new octants
            Tree(10,TreeTop) = Tree(10,j) + ix*Tree(9,TreeTop) ! offset of new actoant
            Tree(11,TreeTop) = Tree(11,j) + iy*Tree(9,TreeTop)
            Tree(12,TreeTop) = Tree(12,j) + iz*Tree(9,TreeTop)
            j = TreeTop
            goto 6007 ! try inserting again
         else ! move up tree
            j = Tree(k,j)
            goto 6007 ! try inserting again
         end if
      end do ! 1:NS have been inserted
      
! dump oct-tree to L
! prepare stack
      Tstack  = 0  ! init
      Pstack  = 1  ! start, the top
      Kstack  = 0  ! init
! now, dump tree into L(i)
      i = 0 ! counter up to NS
      j = 0 ! start at tree root
      k = 0 ! some index to octants in case 1 item at root
6077  continue
      if( Tree(8,j)/=0) then ! a leaf node
         i = i+1
         L(i) = Tree(8,j) ! the index
         Tree(Kstack(Pstack),Tstack(Pstack)) = 0 ! parent delink this node
         j = Tstack(Pstack)  ! descend to parent
         Pstack = Pstack - 1 ! parent below, descend
      end if
      ix=0; iy=0; iz=0;
      if(Bh%E(1) >= Tree(10,j)+Tree(9,j)/2) ix = 1 ! for order of processing octants
      if(Bh%E(2) >= Tree(11,j)+Tree(9,j)/2) iy = 1
      if(Bh%E(3) >= Tree(12,j)+Tree(9,j)/2) iz = 1
      do kk=0,1
      do jj=0,1
      do ii=0,1
         k = mod(ix+ii,2) + 2*mod(iy+jj,2) + 4*mod(iz+kk,2)
         if( Tree(k,j)/=0 ) then
           Pstack = Pstack + 1
           Tstack(Pstack) = j
           Kstack(Pstack) = k
           j = Tree(k,j)
           goto 6077
         end if
      end do
      end do
      end do
      Tree(Kstack(Pstack),Tstack(Pstack)) = 0 ! delink
      j = Tstack(Pstack)
      Pstack = Pstack - 1
      if(Pstack > 0) then
         goto 6077
      end if

#ifdef SREND_COMPOSE_TIMING
      call system_clock(it2)
      print *,'COMPOSE OCTREE sorted: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
#endif

      !n2sort = .TRUE. ! TESTING ONLY, to compare sort times; coment out for real runs

6006  continue ! jumped over tree sort because tree sort is not possible and n2sort==TRUE

if(n2sort) then
#endif
! ---- end if_SREND_OCTREE_SORT --------------------------------------

! ---- n2sort --------------------------------------------------------

      do i=1,NS    
! needed for n^2 sort; boolean TRUE if block is in an octant with respect to eye E
         oct(0,0,0,i) = .FALSE.
         oct(0,0,1,i) = .FALSE.
         oct(0,1,0,i) = .FALSE.
         oct(0,1,1,i) = .FALSE.
         oct(1,0,0,i) = .FALSE.
         oct(1,0,1,i) = .FALSE.
         oct(1,1,0,i) = .FALSE.
         oct(1,1,1,i) = .FALSE.
         
         if( wh( 5,i)<Bh%E(1) .AND. wh( 6,i)<Bh%E(2) .AND. wh( 7,i)<Bh%E(3) ) oct(0,0,0,i) = .TRUE.
         if( wh( 5,i)<Bh%E(1) .AND. wh( 6,i)<Bh%E(2) .AND. wh(13,i)>Bh%E(3) ) oct(0,0,1,i) = .TRUE.
         if( wh( 5,i)<Bh%E(1) .AND. wh(12,i)>Bh%E(2) .AND. wh( 7,i)<Bh%E(3) ) oct(0,1,0,i) = .TRUE.
         if( wh( 5,i)<Bh%E(1) .AND. wh(12,i)>Bh%E(2) .AND. wh(13,i)>Bh%E(3) ) oct(0,1,1,i) = .TRUE.
         if( wh(11,i)>Bh%E(1) .AND. wh( 6,i)<Bh%E(2) .AND. wh( 7,i)<Bh%E(3) ) oct(1,0,0,i) = .TRUE.
         if( wh(11,i)>Bh%E(1) .AND. wh( 6,i)<Bh%E(2) .AND. wh(13,i)>Bh%E(3) ) oct(1,0,1,i) = .TRUE.
         if( wh(11,i)>Bh%E(1) .AND. wh(12,i)>Bh%E(2) .AND. wh( 7,i)<Bh%E(3) ) oct(1,1,0,i) = .TRUE.
         if( wh(11,i)>Bh%E(1) .AND. wh(12,i)>Bh%E(2) .AND. wh(13,i)>Bh%E(3) ) oct(1,1,1,i) = .TRUE.

      end do

! prep for sorting: L() is the array of indices of blocks
      do i=1,NS
         L(i) = i
      end do
      
! ---- n2 sort : back2front algorithm
      do i=1,NS-1
         do j=i+1, NS
            ii=L(i)
            jj=L(j)
! The far corner of jj is "behind" the near corner of ii => swap.  This is not transitive.  
! A block which occludes nothing below rises to the top.  This is for back2front order, but 
! can be reversed for front2back
            if( oct(0,0,0,ii) .AND. wh(5,jj) < wh(11,ii) &
               .AND. wh(6,jj) < wh(12,ii) .AND. wh(7,jj) < wh(13,ii)  ) goto 7006 ! 000
            if( oct(0,0,1,ii) .AND. wh(5,jj) < wh(11,ii) &
               .AND. wh(6,jj) < wh(12,ii) .AND. wh(13,jj) > wh(7,ii)  ) goto 7006 ! 001
            if( oct(0,1,0,ii) .AND. wh(5,jj) < wh(11,ii) &
               .AND. wh(12,jj) > wh(6,ii) .AND. wh(7,jj) < wh(13,ii)  ) goto 7006 ! 010
            if( oct(0,1,1,ii) .AND. wh(5,jj) < wh(11,ii) &
               .AND. wh(12,jj) > wh(6,ii) .AND. wh(13,jj) > wh(7,ii)  ) goto 7006 ! 011
            if( oct(1,0,0,ii) .AND. wh(11,jj) > wh(5,ii) &
               .AND. wh(6,jj) < wh(12,ii) .AND. wh(7,jj) < wh(13,ii)  ) goto 7006 ! 100
            if( oct(1,0,1,ii) .AND. wh(11,jj) > wh(5,ii) &
               .AND. wh(6,jj) < wh(12,ii) .AND. wh(13,jj) > wh(7,ii)  ) goto 7006 ! 101
            if( oct(1,1,0,ii) .AND. wh(11,jj) > wh(5,ii) &
               .AND. wh(12,jj) > wh(6,ii) .AND. wh(7,jj) < wh(13,ii)  ) goto 7006 ! 110
            if( oct(1,1,1,ii) .AND. wh(11,jj) > wh(5,ii) &
               .AND. wh(12,jj) > wh(6,ii) .AND. wh(13,jj) > wh(7,ii)  ) goto 7006 ! 111
            cycle
      
7006        continue
            L(i) = jj
            L(j) = ii
         end do
      end do
      
! switch order to front2back for compositing in Srendfb.F90 
      do i = 1, NS / 2
         j = L(i)
         L(i) = L( NS - i + 1 )
         L( NS - i + 1 ) = j
      end do

#ifdef SREND_COMPOSE_TIMING
      call system_clock(it2)
      print *,'COMPOSE N2 sorted: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
#endif
     
#ifdef SREND_OCTREE_SORT
end if
#endif
! ------------ n2 sort ------ end ------------------------------------


! check bounds if too large, also set for finisher
      if(image) then ! finish out to image, use fill output dimensions 
            wh(1,0) = 1
            wh(2,0) = Bh%W/4 ! for 16pack
            wh(3,0) = 1
            wh(4,0) = Bh%H/4 ! for 16pack
      else if ( 1+wh(2,0)-wh(1,0) > Bh%W/4 ) then ! use output dimensions, flush here
         wh(1,0)  = 1
         wh(2,0)  = Bh%W/4 ! for 16pack
      end if
      
! allocate memory for target array
! this must be reallocated if fly-through
      if(NS > 0) then
         allocate( im(1:16,0:3,wh(1,0):wh(2,0),wh(3,0):wh(4,0)))
         im = 0.0_real32 ! allocate here
      end if

! fill Bh%W and Bh%H are put in header, so these don't match 16-pack dimensions
! load Bheader
      if( NS > 0) then
         Bh%wh(1) = wh(1,0)
         Bh%wh(2) = wh(2,0)
         Bh%wh(3) = wh(3,0)
         Bh%wh(4) = wh(4,0)
      else ! a dud
         Bh%wh(1) =  0
         Bh%wh(2) = -1
         Bh%wh(3) =  0
         Bh%wh(4) = -1
      end if
      
! save Bheader to srendVC(nV)%Bheader buffer for non-blocking send
      Bh%E = E_AMR_PASS

#ifdef SREND_COMPOSE_TIMING
      call system_clock(it2)
      print *,'COMPOSE alloc im(0): span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
#endif

! ccccccccccccccccccc compositing loop ccccccccccccccccccccccccccccccc

do kc = 1, NS
   k = L(kc) ! index to B(nV)%cube(kc)%im
   kim = wh(15,k)

         if( wh(1,k) < 1 .AND. wh(1,0) >= 1) then ! write each half
            do j = wh(3,k), wh(4,k), 1
               do i=wh(1,k),0,1 ! shifted: i -> W+i
SREND_VECTORIZE
                  do ii=1,16
                  im(ii,1,Bh%W/4+i,j) = (1.0_real32-im(ii,0,Bh%W/4+i,j)) &
                  * B(nV)%cube(kim)%im(ii,1,i,j) + im(ii,1,Bh%W/4+i,j) 

                  im(ii,2,Bh%W/4+i,j) = (1.0_real32-im(ii,0,Bh%W/4+i,j)) &
                  * B(nV)%cube(kim)%im(ii,2,i,j) + im(ii,2,Bh%W/4+i,j) 

                  im(ii,3,Bh%W/4+i,j) = (1.0_real32-im(ii,0,Bh%W/4+i,j)) &
                  * B(nV)%cube(kim)%im(ii,3,i,j) + im(ii,3,Bh%W/4+i,j) 

                  im(ii,0,Bh%W/4+i,j) = (1.0_real32-im(ii,0,Bh%W/4+i,j)) &
                  * B(nV)%cube(kim)%im(ii,0,i,j) + im(ii,0,Bh%W/4+i,j) 
                  end do ! ii
               end do ! i

               do i=1,wh(2,k),1 ! not shifted
SREND_VECTORIZE
                  do ii=1,16
                  im(ii,1,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,1,i,j) + im(ii,1,i,j) 

                  im(ii,2,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,2,i,j) + im(ii,2,i,j) 

                  im(ii,3,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,3,i,j) + im(ii,3,i,j) 

                  im(ii,0,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,0,i,j) + im(ii,0,i,j) 
                  end do ! ii
               end do ! i
            end do ! j

         else ! in range: 1:W,1:H or iw0(0):iw1(0),:

            do j = wh(3,k), wh(4,k), 1
               do i=wh(1,k),wh(2,k),1
SREND_VECTORIZE
                  do ii=1,16
                  im(ii,1,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,1,i,j) + im(ii,1,i,j) 

                  im(ii,2,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,2,i,j) + im(ii,2,i,j) 

                  im(ii,3,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,3,i,j) + im(ii,3,i,j) 

                  im(ii,0,i,j) = (1.0_real32-im(ii,0,i,j)) &
                  * B(nV)%cube(kim)%im(ii,0,i,j) + im(ii,0,i,j) 
                  end do
               end do ! i
            end do ! j

         end if ! in range
         
#ifdef SREND_IM_AUTO_DEALLOCATION
         deallocate( B(nV)%cube(kim)%im )
#endif
         
end do ! kc, k 

! ccccccccccccc end compositing loop ccccccccccccccccccccccccccccccccc


#ifdef SREND_COMPOSE_TIMING
      call system_clock(it2)
      print *,'COMPOSE composited: span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
#endif

      Bh%leader = 'CHAD' ! tag for CHAD/SCHAD
      Bh%header_len = SREND_BHEADER_LEN
      Bh%nchad = 1   ! number of packages, just this one for render_flush
      Bh%bit16 = 0
      Bh%RLE0 = 0
      Bh%plen = 0
      if( NS>0 ) Bh%plen = sizeof( im )
      Bh%tlen = Bh%plen+SREND_BHEADER_LEN

      if( image ) then !n=0
         call srend_write_image(Bh,im,mem,memlen) !
      else ! flush: Bheader+im(0) !n=0
#ifdef SREND_16BIT
         Bh%bit16 = 1
#ifdef SREND_RLE0
         Bh%RLE0 = 1
#endif
         if(NS>0) then
            call srend_pack_16bit(Bh%RLE0, im,                              &
                                       Bh%wh(1),Bh%wh(2),Bh%wh(3),Bh%wh(4), &
                                       mem(SREND_BHEADER_LEN+1:),memlen)
            Bh%plen = memlen
            memlen = memlen + SREND_BHEADER_LEN
            Bh%tlen = memlen
         else
            Bh%plen = 0
            memlen = SREND_BHEADER_LEN
            Bh%tlen = memlen
         end if
         mem(1:SREND_BHEADER_LEN) = transfer(Bh,mem(1:SREND_BHEADER_LEN))
#else
         ii = SREND_BHEADER_LEN
         if(NS>0) then
            do ir=Bh%wh(3),Bh%wh(4)
            do ic=Bh%wh(1),Bh%wh(2)
              real4_cast256 = im(:,:,ic,ir) ! 
              !mem(ii+1:ii+256) = char1_cast(:)
              mem(ii+1:ii+256) = transfer(real4_cast256,mem(ii+1:ii+256))
              ii = ii+256 ! memlen
            end do
            end do
         end if
         memlen = ii
         Bh%tlen = memlen
         Bh%plen = Bh%tlen - SREND_BHEADER_LEN
         mem(1:SREND_BHEADER_LEN) = transfer(Bh,mem(1:SREND_BHEADER_LEN))
#endif
      end if

      

#ifdef SREND_COMPOSE_TIMING
      call system_clock(it2)
      print *,'COMPOSE DONE (cast, write): span=',(it2-it1)*rc,'total=',(it2-it_start)*rc
      it1 = it2
#endif

      return

   end subroutine srend_compose_base! Srend subroutine
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! ccccccccccc end: compose ccccccccccccccccccccccccccccccccccccccccccc
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc



! WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
! WWWWWWWWWWWWWWWWWWWWW write to disk WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
! WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
   subroutine srend_write_image(inBh, im_in, mem, memlen)
      implicit NONE
      type(B_header_type)                 :: inBh
      real(real32),dimension(1:,0:,1:,1:) :: im_in
      character, dimension(:)             :: mem
      integer(int32)                      :: memlen
! --------------------------------------------------------------------
! global rendering parameters
      real(real32)                                     :: dW, dH      ! increments for W and H across viewing 
      real(real32),dimension(:,:,:),allocatable        :: im          ! scratch for finishing planes to image
      character, dimension(:,:,:),allocatable          :: ppm_buffer  ! scratch
      character, dimension(:,:,:),allocatable          :: ppm_buffer2 ! for perspective, extra padding for interp
      real(real32), dimension(:,:,:),allocatable       :: perspective ! for perspective
          
#if defined(SREND_IMAGE_BACKGROUND) && !defined(SREND_RGBA)
      real(real32),dimension(1:3)                      :: background_RGB
      real(real32)                                     :: transparency
#endif
#ifndef SREND_RGBA
      real(real32)                                     :: RGB(1:3)
#else
      real(real32)                                     :: RGB(0:3)
#endif
      character(LEN=80)                                :: header              ! file IO
      real(real32)                                     :: aa, bb, aaa, bbb    ! for rendering to perspective
      integer(int32)                                   :: i,j,k, ii,jj,kk     ! loop counters
      real(real32)                                     :: dWW, dHH            ! for arbitrary angle spans
      real(real32)                                     :: a_span, b_span      ! span, centered at spherical origin
      real(real32)                                     :: ddx0,ddy0,ddx1,ddy1 ! bilinear interpolation
#ifdef SREND_BMP
      type(BMP_header_type)                            :: BMP
      character                                        :: BMP_c124(1:124) ! for offset, avoid compiler warnings
#endif
      type(B_header_type)                              :: Bh
! --------------------------------------------------------------------

!
! PPM (Portable PixMap for imagery, not to be confused with Piecewise 
! Parabolic Method) is 24 bit RGB raster with a plain text header. 
! A provided file name with the extension ".ppm. selects this
! format.  An added fixed-length PPM header (here 20 bytes) is the 
! only difference between PPM and raw.
!
! If output images need to be converted to a different format,
! ImageMagick convert & mogrify are very fast.  Install ImageMagick;
! convert and mogrify commands will be installed.
! Example
!   convert all PPM images to PMG: mogrify -format png *.ppm
!
! ImageMagick's convert utility can tile out an image:
!   convert -crop 25%x25% d000008.ppm d00008.png
! This results in 16 images d000008-0.png to d000008-15.png, ordered
! by row with 0-based offset.
! --------------------------------------------------------------------

! The sampling data was collected with vectors of 16 real(real32) RGBA values, but we
! will process images by pixel rather than 16-packs
! remap im_in(1:16,0:3,...) to im(0:3,...)

      Bh = inBh

      allocate( im(0:3,Bh%W,Bh%H) )
      if( Bh%wh(2)-Bh%wh(1)>0) then
	do j = 1,Bh%H/4
	do i = 1,Bh%W/4
	do k=0,3
	do kk=1,16
	  im(k,(i-1)*4+ mod(kk-1,4)+1,(j-1)*4+ (kk-1)/4+1) = im_in(kk,k,i,j)
	end do
	end do
	end do
	end do
      else
#if defined(SREND_IMAGE_BACKGROUND) && !defined(SREND_RGBA)
        im = 1.0_real32
#else
        im = 0.0_real32
#endif
      end if
      
#ifdef SREND_RGBA
! un-premultiply alpha
      if( Bh%wh(2)-Bh%wh(1)>0) then
        do j = 1,Bh%H
        do i = 1,Bh%W
           if( im(0,i,j) < .001_real32 ) then
              im(0:3,i,j) = 0.0_real32
           else
              im(1:3,i,j) = im(1:3,i,j) / im(0,i,j) ! divide out alpha
           end if
        end do
        end do
      end if
#else
#ifdef SREND_IMAGE_BACKGROUND
      background_RGB = SREND_IMAGE_BACKGROUND
      if( Bh%wh(2)-Bh%wh(1)>0) then
        do j = 1,Bh%H
        do i = 1,Bh%W
           transparency = 1.0_real32 - im(0,i,j)
        do k=1,3
           im(k,i,j) = im(k,i,j) + transparency*background_RGB(k)
        end do
        end do
        end do
      end if
#endif
#endif
      

! the buffer for all, .ppm
#ifndef SREND_RGBA
      allocate( ppm_buffer(1:3,1:Bh%W,1:Bh%H) )
#else
      allocate( ppm_buffer(0:3,1:Bh%W,1:Bh%H) )
#endif
      
! WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
      if( Bh%perspective == 1) then
#ifndef SREND_RGBA
         allocate( ppm_buffer2(1:3,0:Bh%W+1,0:Bh%H+1) )
         allocate( perspective(1:3,1:Bh%W,  1:Bh%H  ) )
#else
         allocate( ppm_buffer2(0:3,0:Bh%W+1,0:Bh%H+1) )
         allocate( perspective(0:3,1:Bh%W,  1:Bh%H  ) )
#endif
         do j=1,Bh%H
            do i=1,Bh%W
#ifdef SREND_RGBA
               RGB(0) = 255.0_real32 * im(0,i,j)
#endif
               RGB(1) = 255.0_real32 * im(1,i,j)
               RGB(2) = 255.0_real32 * im(2,i,j)
               RGB(3) = 255.0_real32 * im(3,i,j)  
! turn real(real32) into bytes ... consistency
#ifdef SREND_RGBA
               ppm_buffer2(0,i,j) = char( floor(RGB(0)) ) ! A
#endif
               ppm_buffer2(1,i,j) = char( floor(RGB(1)) ) ! R
               ppm_buffer2(2,i,j) = char( floor(RGB(2)) ) ! G
               ppm_buffer2(3,i,j) = char( floor(RGB(3)) ) ! B
            end do
         end do
! pad top and bottom row of pixels in ppm_buffer2
#ifndef SREND_RGBA
         ppm_buffer2(1:3,1:Bh%W,0)         = ppm_buffer2(1:3,1:Bh%W,1)
         ppm_buffer2(1:3,1:Bh%W,Bh%H+1) = ppm_buffer2(1:3,1:Bh%W,Bh%H)
#else
         ppm_buffer2(0:3,1:Bh%W,0)         = ppm_buffer2(0:3,1:Bh%W,1)
         ppm_buffer2(0:3,1:Bh%W,Bh%H+1) = ppm_buffer2(0:3,1:Bh%W,Bh%H)
#endif
! pad sides, padded array corners are data corners
#ifndef SREND_RGBA
         ppm_buffer2(1:3,0,0:Bh%H+1)         = ppm_buffer2(1:3,1,0:Bh%H+1)
         ppm_buffer2(1:3,Bh%W+1,0:Bh%H+1) = ppm_buffer2(1:3,Bh%W,0:Bh%H+1)
#else
         ppm_buffer2(0:3,0,0:Bh%H+1)         = ppm_buffer2(0:3,1,0:Bh%H+1)
         ppm_buffer2(0:3,Bh%W+1,0:Bh%H+1) = ppm_buffer2(0:3,Bh%W,0:Bh%H+1)
#endif
! viewing increments, assume W and H are divisible by 4
         dW = Bh%Alpha * SREND_PI/180.0_real32 / Bh%W
         dH = Bh%Beta  * SREND_PI/180.0_real32 / Bh%H
! convert to radians, no shrink .. have padding
         a_span = Bh%Alpha * SREND_PI / 180.0_real32
         b_span = Bh%Beta  * SREND_PI / 180.0_real32

         dWW = 2.0_real32*tan(.5_real32*a_span)/Bh%W
         dHH = 2.0_real32*tan(.5_real32*b_span)/Bh%H
         do j=1,Bh%H
            bb = (Bh%H/2.0_real32 - j)*dHH + dHH/2.0_real32 ! angle coords, persp
            do i=1,Bh%W
               aa = (Bh%W/2.0_real32 - i)*dWW + dWW/2.0_real32 ! angle coords, persp
               aaa = atan(aa)               ! project
               bbb = atan( bb / sqrt(aa*aa + 1.0_real32) )
   ! interpolation: bilinear
               ddx0 = Bh%W/2.0_real32 - (aaa - dW/2.0_real32)/dW ! hit spherical pix
               ii = floor(ddx0)
               ddx0 = ddx0 - ii
               ddx1 = 1.0_real32 - ddx0
               
               ddy0 = Bh%H/2.0_real32 - (bbb - dH/2.0_real32)/dH
               jj = floor(ddy0)
               ddy0 = ddy0 - jj
               ddy1 = 1.0_real32 - ddy0
#ifndef SREND_RGBA
               do k=1,3
#else
               do k=0,3
#endif
                  perspective(k,i,j) =                         &
                  ichar(ppm_buffer2(k,ii,jj))    *ddx1*ddy1 +  &
                  ichar(ppm_buffer2(k,ii+1,jj))  *ddx0*ddy1 +  &
                  ichar(ppm_buffer2(k,ii,jj+1))  *ddx1*ddy0 +  &
                  ichar(ppm_buffer2(k,ii+1,jj+1))*ddx0*ddy0
               end do
            end do
         end do
            
         do j=1,Bh%H ! copy back to ppm_buffer for writing .ppm
            do i=1,Bh%W
#ifdef SREND_RGBA
               ppm_buffer(0,i,j) = char(floor(perspective(0,i,j)) )
#endif
               ppm_buffer(1,i,j) = char(floor(perspective(1,i,j)) )
               ppm_buffer(2,i,j) = char(floor(perspective(2,i,j)) )
               ppm_buffer(3,i,j) = char(floor(perspective(3,i,j)) )
            end do
         end do
! cccccccccccccc end spherical -> perspective projection ccccccccccccc
      else ! spherical

! char & byte are not vectorizable data types, but could be done
! in lower level fashion
         do j=1,Bh%H
         do i=1,Bh%W
#ifdef SREND_RGBA
            RGB(0) = 255.0_real32 * im(0,i,j)
#endif
            RGB(1) = 255.0_real32 * im(1,i,j)
            RGB(2) = 255.0_real32 * im(2,i,j)
            RGB(3) = 255.0_real32 * im(3,i,j)  

#ifdef SREND_RGBA
            ppm_buffer(0,i,j) = char( floor(RGB(0)) ) ! A
#endif
            ppm_buffer(1,i,j) = char( floor(RGB(1)) ) ! R
            ppm_buffer(2,i,j) = char( floor(RGB(2)) ) ! G
            ppm_buffer(3,i,j) = char( floor(RGB(3)) ) ! B
         end do
         end do
      end if ! perspective or spherical

! write .ppm 24-bit RGB image
#if !defined(SREND_RGBA) && !defined(SREND_BMP)
! write to imout
! a 24 bit raw .PPM (Portable Pixel Map) image file must have "'P6' width height 255" starting
! at byte 0, and the last item of info 255 must be followed by 1 white space, here a newline
! achar(10), then raw RGB bytes follow.  .PPM files can have comments and optional layouts, but
! srend always uses a 20 byte header with width and height both <= 99999.
      write(header,"('P6',I5,1X,I5,'   255')") Bh%W, Bh%H ! PPM header, 19 bytes exactly
      do i=1,19
        mem(i) = header(i:i)
      end do
      mem(20) = achar(10) ! PPM header above terminated by 1 newline (line feed: UNIX)
      ii=20
      do j=1,Bh%H
      do i=1,Bh%W
      do k=1,3
        ii = ii+1
        mem(ii) = ppm_buffer(k,i,j)
      end do
      end do
      end do
      
      memlen = ii! 20 + 3*Bh%W*Bh%H ! bytes for .ppm RGB image
#endif

! write .pam 32 bit RGBA image
#if defined(SREND_RGBA) && !defined(SREND_BMP)
! .pam RGBAformat : 32 bits: Imagemagick: convert im.pam im.png 
      write(header( 1: 3),"('P7',A1)") achar(10)                 ! 3= 1:3
      write(header( 4:15),"('WIDTH ',I5,A1)") Bh%W, achar(10)    !12= 4:15
      write(header(16:28),"('HEIGHT ',I5,A1)") Bh%H, achar(10)   !13=16:28
      write(header(29:41),"('DEPTH      4',A1)") achar(10)       !13=29:41
      write(header(42:54),"('MAXVAL   255',A1)") achar(10)       !13=42:54
      write(header(55:73),"('TUPLTYPE RGB_ALPHA',A1)") achar(10) !19=55:73
      write(header(74:80),"('ENDHDR',A1)") achar(10)             ! 7=74:80
      do i=1,80
        mem(i) = header(i:i)
      end do
      ii=80
      do j=1,Bh%H
      do i=1,Bh%W
      do k=1,3
        ii = ii+1
        mem(ii) = ppm_buffer(k,i,j) ! RGB
      end do
        ii = ii+1
        mem(ii) = ppm_buffer(0,i,j) ! alpha
      end do
      end do
      
      memlen = ii !80 + 4*Bh%W*Bh%H ! bytes for .pam RGBA image
#endif

! NOTE: 

! .bmp image, 24-bit
#if defined(SREND_BMP) && !defined(SREND_RGBA)
      BMP%bytes      = 54 + 3*Bh%W*Bh%H
      BMP%W          = Bh%W
      BMP%H          = Bh%H
      BMP%bits       = 24 ! =24 or 32
      BMP%data_bytes = 3*Bh%W*Bh%H
        BMP_c124 = transfer(BMP,BMP_c124) ! fix to avoid compiler warnings
        mem(1:54) = BMP_c124(3:56)
      ii=54
      do j=Bh%H,1,-1 ! bottom-2-Top
      do i=1,Bh%W
      do k=3,1,-1    ! BGR order
        ii = ii+1
        mem(ii) = ppm_buffer(k,i,j) ! BGR
      end do
      end do
      end do
      
      memlen = ii ! 54 + 3*Bh%W*Bh%H
#endif

! .bmp images, 32-bit
#if defined(SREND_BMP) && defined(SREND_RGBA)
      BMP%bytes      = 122 + 4*Bh%W*Bh%H
      BMP%W          = Bh%W
      BMP%H          = Bh%H
      BMP%bits       = 32 ! =24 or 32
      BMP%data_bytes = 4*Bh%W*Bh%H
      BMP%data_offset = 122      
      BMP%DIB_header_bytes = 108
      BMP%compression = 3
      BMP%mask_RGBA(1) =  16711680 !int(z'00ff0000') !R  16711680
      BMP%mask_RGBA(2) =     65280 !int(z'0000ff00') !G     65280
      BMP%mask_RGBA(3) =       255 !int(z'000000ff') !B       255
      BMP%mask_RGBA(4) = -16777216 !int(z'ff000000') !A -16777216
        BMP_c124 = transfer(BMP,BMP_c124) ! fix to avoid compiler warnings
        mem(1:122) = BMP_c124(3:124)
      ii=122
      do j=Bh%H,1,-1 ! bottom-2-Top
      do i=1,Bh%W
      do k=3,0,-1    ! BGRA order
        ii = ii+1
        mem(ii) = ppm_buffer(k,i,j) ! BGRA
      end do
      end do
      end do
      
      memlen = ii ! 122 + 4*Bh%W*Bh%H
#endif

      
      return

! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
! NOTE: The Portable Pix Map (PPM) format is directly read and 
! displayed by many browsers and desktops.  It is also a base format
! which ImageMagick and ffmpeg easily use; it amounts to a raw RGB 
! raster image with a header.
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!          ./ffmpeg -i %06d.ppm -b 1200k full.ogv
! This generates a smooth movie.
!   ffmpeg      : http://www.ffmpeg.org
!   -i %06d.ppm : input files ######.ppm, 000001.ppm, ...
!   -b 1200k    : bitrate 
!   target      : full.ogv
! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
   end subroutine srend_write_image
! wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww
! wwwwwwwwwwwwwww end : write to disk  wwwwwwwwwwwwwwwwwwwwwwwwwwwwwww
! wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww



! ********************************************************************
#ifdef SREND_16BIT
! ---- packaging messages with compression ---------------------------
! We must use real(real32) for sampling and compositing arithmetic, but 15
! bits are enough, and so we use 15 bits for MPI messaging and storage 
! in-between.  One can expect high extropy rendered plane data which 
! doesn't compress well, but there will be zeros surrounding data.  To 
! be simple and fast, we compress just consecutive zeros.

! integer(int16) im2(1:im2_alloc_len) is a common buffer we allocate once
! for all views.  We reallocate it whenever it is too short.
   recursive subroutine srend_pack_16bit(iRLE0,im,w0,w1,h0,h1,im2_out,im2_plen)
      implicit NONE
! the incoming data is in im(), all parameters known before call
! the result is in the common buffer im2(1:im2_alloc_len) with the
! parameter im2_plen, so the compressed result is: im2(1:im2_plen)
      integer(int16),intent(IN)                    :: iRLE0 ! 1 if compressed, o not
      integer(int32),intent(IN)                    :: h0,h1,w0,w1
      real(real32),intent(IN)                      :: im(1:16,0:3,w0:w1,h0:h1)
      character, dimension(:),intent(INOUT) :: im2_out
      integer(int32),intent(OUT)                   :: im2_plen ! a result needed for the data
      
      integer(int32)   :: ih,iw,irgb,i, ip  ! counters
      integer(int16)   :: int2_cast         ! convert data in integer(int16) to character(1:2)
      integer(int32)   :: zeros             ! number of zeros in integer(int16) to character(1:2)
      integer(int16)   :: int2_0cast
      
   if(iRLE0==0) then ! scale real(real32) data in [0.0,1.0] to [0,32767] for integer(int16) rep
      ip = 0
      do ih=h0,h1
      do iw=w0,w1
      do irgb=0,3
      do i=1,16
         int2_cast = INT2( 32767.0_real32 * max(0.0_real32,min(1.0_real32,im(i,irgb,iw,ih))))
         im2_out(ip+1:ip+2) = transfer(int2_cast,im2_out(ip+1:ip+2))
         ip = ip+2
      end do
      end do
      end do
      end do  ! im2(1:im2_len) is the unpacked data as integer(int16) in ARGB-1:16 order
      im2_plen = ip
   else ! scale real(real32) data in [0.0,1.0] to [0,32767] for integer(int16) rep, but
        ! also pack up to 32768 consecutive zeros as a negative integer(int16)
      zeros = 0
      ip = 0
      do ih=h0,h1
      do iw=w0,w1
      do irgb=0,3
      do i=1,16
         int2_cast = INT2( 32767.0_real32 * max(0.0_real32,min(1.0_real32,im(i,irgb,iw,ih)))) ! real(real32) to int2
         if( int2_cast == 0) then
            zeros = zeros - 1 ! another zero
         else
            if(zeros < 0) then ! write the zeros (neg int2), then the data
               int2_0cast = INT2(zeros)
               im2_out(ip+1:ip+2) = transfer(int2_0cast,im2_out(ip+1:ip+2))
               zeros = 0
               im2_out(ip+3:ip+4) = transfer(int2_cast,im2_out(ip+3:ip+4))
               ip = ip + 4
            else ! write the data
               im2_out(ip+1:ip+2) = transfer(int2_cast,im2_out(ip+1:ip+2))
               ip = ip + 2
            end if
         end if
         if(zeros == -32768) then ! max (min int2) number of zeros, so write
            int2_0cast = INT2(zeros)
            im2_out(ip+1:ip+2) = transfer(int2_0cast,im2_out(ip+1:ip+2))
            zeros = 0
            ip = ip+2
         end if
      end do
      end do
      end do
      end do
      if(zeros<0) then ! if the last data item was a zero, the number of zeros needs to be written out here
         int2_0cast = INT2(zeros)
         im2_out(ip+1:ip+2) = transfer(int2_0cast,im2_out(ip+1:ip+2))
         ip = ip+2
      end if
      im2_plen = ip
   end if
   end subroutine srend_pack_16bit
#endif
! ********************************************************************
   
   
! ********************************************************************
   subroutine srend_unpack_16bit(iRLE0,im,w0,w1,h0,h1,im2_in,im2_plen)
      implicit NONE
! The incoming data is im2(1:im2_plen) which is within im2(1:im2_alloc_len)
! The results are written into im(), and the parameters are already known before call.
      integer(int16),intent(IN)                 :: iRLE0 ! 1 if RLE0 packed, 0 if not
      integer(int32),intent(IN)                 :: h0,h1,w0,w1
      real(real32),intent(INOUT)                :: im(1:16,0:3,w0:w1,h0:h1) ! modified, the resulting data
      character,dimension(:),intent(IN)         :: im2_in ! this needs to be converted to integer(int16)
      integer(int32),intent(INOUT)              :: im2_plen
      integer(int32)                     :: ih,iw,irgb,i,ip
      integer(int16)                     :: int2_cast
      integer(int32)                     :: zeros
! ---- end var declarations ------------------------------------------

   if(iRLE0==0) then ! convert integer(int16) im2(1:im2_len) to real(real32) im()
      ip = 0
      do ih=h0,h1
      do iw=w0,w1
      do irgb=0,3
      do i=1,16
         int2_cast = transfer(im2_in(ip+1:ip+2),int2_cast)
         ip=ip+2
         im(i,irgb,iw,ih) = REAL(max(0.0_real32,min(1.0_real32,int2_cast/32767.0_real32)))
      end do
      end do
      end do
      end do
      im2_plen = 4 * 16*4*(w1-w0+1)*(h1-h0+1) !ip*2
   else ! convert integer(int16) im2(1:im2_len) to real(real32) im(), also unpack consecutive zeros
      zeros = 0
      ip = 0
      do ih=h0,h1
      do iw=w0,w1
      do irgb=0,3
      do i=1,16
         if( zeros < 0) then ! zeros to write in im, keep doing it till zeros=0
            im(i,irgb,iw,ih) = 0.0_real32
            zeros = zeros + 1
         else ! zeros==0, so read data item
            int2_cast = transfer(im2_in(ip+1:ip+2),int2_cast)
            ip = ip + 2
            if(int2_cast < 0) then ! write a zero
               zeros = int2_cast + 1 ! have already written 1 zero
               im(i,irgb,iw,ih) = 0.0_real32
            else ! write a non-zero item
               im(i,irgb,iw,ih) = REAL(max(0.0_real32,min(1.0_real32,int2_cast/32767.0_real32)))
            end if
         end if
      end do
      end do
      end do
      end do
      im2_plen = 4 * 16*4*(w1-w0+1)*(h1-h0+1)
   end if
   end subroutine srend_unpack_16bit
! ********************************************************************


! ********************************************************************
! ******* utilities: for external use ********************************
! ********************************************************************


! ********************************************************************
! returns number loaded B(nV)%n per view index nV
   INTEGER function srend_n(nV)
      implicit NONE
      INTEGER,intent(IN) :: nV
#ifdef SREND_B_DYNAMIC_V
      if(.NOT. B_used) then
         srend_n = 0
         return
      end if
#endif
      if(.NOT. B(nV)%used) then
         srend_n = 0 ! return value, handy for some logic to not return error
         return
      end if
      srend_n = B(nV)%n
   end function srend_n
! ********************************************************************


! ********************************************************************
! resets: B(nV)%cotab_set = .FALSE., for rebuilding color table
   subroutine srend_reset_cotab(nV)
      implicit NONE
      INTEGER, intent(IN)  :: nV
#ifdef SREND_B_DYNAMIC_V
      if(.NOT. B_used) return
#endif
      if(.NOT. B(nV)%used) return
      B(nV)%cotab_set = .FALSE.
   end subroutine srend_reset_cotab
! ********************************************************************


! ********************************************************************
   subroutine srend_CHAD_nV_set(mem, nV)
      implicit NONE
      CHARACTER,dimension(:),intent(INOUT)   :: mem
      INTEGER, intent(IN)                    :: nV
      type(B_header_type)                    :: Bh
! ---- end var declarations ------------------------------------------
      Bh = transfer(mem(1:SREND_BHEADER_LEN),Bh)   
      Bh%nV = nV
      mem(1:SREND_BHEADER_LEN) = transfer(Bh,mem(1:SREND_BHEADER_LEN))
   end subroutine srend_CHAD_nV_set
! ********************************************************************


! ********************************************************************
! prints CHAD,PPM,PAM,BMP info for data at mem
   subroutine srend_mem_info(mem)
      implicit NONE
      CHARACTER,dimension(:),intent(IN) :: mem ! data to examine
      
      integer(int32)    :: i, total_nchad, ip  ! CHAD/SCHAD info gathering
      integer(int32)    :: W, H, image_len     ! for PPM and PAM parsing
      character(LEN=80) :: s80                 ! examining header string

      type(B_header_type) :: Bh ! CHAD header ----
      
      type BMP24_header_type ! sufficient for getting BMP info, 24 and 32 bit
         SEQUENCE
         character(LEN=2) :: offset_2bytes_o_crap  ! force alignment to suppress compiler warnings
         character(LEN=2) :: ID = 'BM'             ! 1- 2 ='BM'
         integer(int32)   :: bytes                 ! 3- 6 ?=54+3*W*H or 54+4*W*H
         integer(int32)   :: zero4 = 0             ! 7-10 =0000 - reserved
         integer(int32)   :: data_offset = 54      !11-14 =54, size of header in bytes
         integer(int32)   :: DIB_header_bytes = 40 !15-18 =40, DIB header bytes
         integer(int32)   :: W                     !19-22 ?=W, Width of data
         integer(int32)   :: H                     !23-26 ?=H, Height of data
         integer(int16)   :: planes = 1            !27-28 =1
         integer(int16)   :: bits                  !29-30 ?=24 or 32
         integer(int32)   :: compression = 0       !31-34 =0 24-bit, =3 32-bit (none)
         integer(int32)   :: data_bytes            !35-38 ?=3*W*H or 4*W*H
         integer(int32)   :: Wres = 2834           !39-42 =2834, resolution in horz
         integer(int32)   :: Hres = 2834           !43-46 =2834, resolution in vert
         integer(int32)   :: palette = 0           !47-50 =0, none
         integer(int32)   :: colors = 0            !51-54 =0, means all
!BMP32   integer(int32)   :: mask_RGBA(1:4)        !55-70 0000FF00,00FF0000,FF000000,000000FF
!BMP32   character(len=4) :: color_space = 'BGRs'  !71-74 LCS color space, little endian
!BMP32   integer(int32)   :: zero12(1:12) = (/0,0,0,0,0,0, 0,0,0,0,0,0/) !75-122 unused
      end type BMP24_header_type
      
      type(BMP24_header_type) :: BMPh
      character               :: BMP_c56(1:56)
! --------------------------------------------------------------------
      s80(1:4) = transfer(mem(1:4),s80(1:4))!cast_4c(1:4) = mem(1:4)
      if( s80(1:4) =='CHAD') then
         Bh = transfer(mem(1:SREND_BHEADER_LEN),Bh)
         total_nchad = Bh%nchad
         print *,'INFO:CHAD/SCHAD: nchad= ',total_nchad,' tlen= ', Bh%tlen
         ip = 0
         do i=1, total_nchad
            Bh = transfer(mem(ip+1:ip+SREND_BHEADER_LEN),Bh)
            print *,'******** chad ',i,' **********************'
            print *,'INFO:leader=',Bh%leader,'tlen=',Bh%tlen
            print *,'INFO:nchad=',Bh%nchad,'plen=',Bh%plen
            print *,'INFO:,16bit=',Bh%bit16,'RLE0=',Bh%RLE0
            print *,'INFO:wh=',Bh%wh
            print *,'INFO:xyzoff=',Bh%xoff,Bh%yoff,Bh%zoff
            print *,'INFO:XYZN=',Bh%XN,Bh%YN,Bh%ZN
            print *,'INFO:Vdim=',Bh%Vdim,'W=',Bh%W,'H=',Bh%H
            print *,'INFO:nV=',Bh%nV,'perspective=',Bh%perspective
            print *,'INFO:alpha=',Bh%alpha,'beta=',Bh%beta
            print *,'INFO:E=',Bh%E
            print *,'INFO:V=',Bh%V
            print *,'INFO:U=',Bh%U
            print *,'INFO:header_len=',Bh%header_len
            ip = ip+BH%header_len + Bh%plen
         end do
      else if( s80(1:2) == 'P6') then
         s80(3:20) = transfer(mem(3:20),s80(3:20))
         read(s80(3:20),*) W, H ! PPM header, 19 bytes exactly
         image_len = 20+3*W*H
         print *,'******** PPM image **********************'
         print *,"INFO:W=",W,"H=",H,"total image len=",image_len
         print *,'INFO:header len=',20,'image data len=', 3*W*H
      else if( s80(1:2) == 'P7') then
         s80 = transfer(mem(1:80),s80)
         read(s80(10:14),*) W
         read(s80(23:27),*) H
         image_len = 80+4*W*H
         print *,'******** PAM image **********************'
         print *,"INFO:W=",W,"H=",H,"total image len=",image_len
         print *,'INFO:header len=',80,'image data len=', 4*W*H
      else if( s80(1:2) == 'BM') then
         BMP_c56(3:56) = mem(1:54)
         BMPh = transfer(BMP_c56,BMPh)
         print *,'******** BMP image **********************'
         if(BMPh%bits == 24) then
           print *,'INFO: BMP24 BGR'
         else
           print *,'INFO: BMP32 ABGR'
         end if
         print *,'INFO: total byte length =',BMPh%bytes
         print *,'INFO: header bytes =', BMPh%data_offset
         print *,'INFO: W =', BMPh%W
         print *,'INFO: H =', BMPh%H
         print *,'INFO: bits =', BMPh%bits
      else
         print *,"INFO: not an srend CHAD; also not a PPM,PAM, or BMP image"
      end if
   end subroutine srend_mem_info
! ********************************************************************

! ********************************************************************
! returns length of CHAD/SCHAD or image in bytes
   INTEGER function srend_memlen(mem)
      implicit NONE
      CHARACTER,dimension(1:),intent(IN) :: mem ! data to examine
  
      integer(int32)           :: memlen, W, H
      character(LEN=32)        :: s32
      type(B_header_type)      :: Bh
! ---- end variable declarations -------------------------------------
      s32(1:4) = transfer(mem(1:4),s32(1:4))
      if( s32(1:4) == 'CHAD') then
         Bh = transfer(mem(1:SREND_BHEADER_LEN),Bh)
         srend_memlen = Bh%tlen
         return
      else if( s32(1:2) == 'P6') then
         s32(3:20) = transfer(mem(3:20),s32(3:20))
         read(s32(3:20),*) W, H ! PPM header ONLY in srend, 20 bytes exactly
         srend_memlen = 20+3*W*H
         return
      else if( s32(1:2) == 'P7') then
         s32(10:27) = transfer(mem(10:27),s32(10:27))
         read(s32(10:14),*) W   ! PAM header ONLY in srend, 80 bytes exactly
         read(s32(23:27),*) H
         srend_memlen = 80+4*W*H
         return
      else if( s32(1:2) == 'BM') then
         memlen = transfer(mem(3:6),memlen)
         srend_memlen = memlen
         return
      else
         srend_memlen = 0
         return
      end if
   end function srend_memlen
! ********************************************************************


! ********************************************************************
   subroutine srend_image2disk(mem, fname)
      CHARACTER, dimension(:),intent(IN)   :: mem
      CHARACTER(len=*),intent(IN)          :: fname
      
      character(LEN=3)               :: ext, new_ext = 'nun' ! nothing, at first
      character(LEN=256)             :: fname_root, new_fname, new_fname2
      integer(int32)                 :: i
      character(LEN=2)               :: im_prefix
      integer                        :: MYunit
      
      i = index(fname,'.',BACK=.TRUE.)
      if(i>0 .AND. len(fname) == i+3) then
         new_ext = fname(i+1:i+3)
         fname_root = fname(1:i-1)
      else
         fname_root = trim(fname)
      end if
      
      im_prefix = transfer(mem(1:2),im_prefix)
      if( im_prefix == 'P6') then
         ext = 'ppm'
      else if( im_prefix == 'P7') then
         ext = 'pam'
      else if( im_prefix == 'BM') then
         ext = 'bmp'
      else
         return ! not an srend image
      end if
      
      new_fname = trim(fname_root) // '.' // ext
! write the file
      open(newunit=MYunit,status='unknown',file=trim(new_fname),access='stream')
      write (MYunit) mem(1:srend_memlen(mem))
      close (MYunit)
! ImageMagick must be installed for this to work
      if( new_ext /= 'nun' .AND. new_ext /= ext) then
         new_fname2 = trim(fname_root) // '.' // new_ext
         call system( 'convert ' // trim(new_fname) // ' ' // trim(new_fname2) )
         call system( 'rm ' // trim(new_fname) )
      end if

   end subroutine srend_image2disk
! ********************************************************************


! ********************************************************************
! chad1 = chad1+chad2: creates a SCHAD tacking chad2 after chad1, updatng chad1
! the user has to make sure that chad1 will hold chad1+chad2
   subroutine srend_CHAD_concat(chad1,chad2)
      implicit NONE
      CHARACTER,dimension(1:),intent(INOUT) :: chad1
      CHARACTER,dimension(1:),intent(IN)    :: chad2 ! data to examine
      integer(int32)          :: tlen1,tlen2,nchad2
      type(B_header_type)     :: Bh
! ---- end variable declarations -------------------------------------
      Bh = transfer(chad2(1:SREND_BHEADER_LEN),Bh)
      tlen2 = Bh%tlen
      nchad2 = Bh%nchad
      Bh = transfer(chad1(1:SREND_BHEADER_LEN),Bh)
      tlen1 = Bh%tlen
      Bh%tlen = tlen1 + tlen2
      Bh%nchad = Bh%nchad + nchad2
      chad1(1:SREND_BHEADER_LEN) = transfer(Bh,chad1(1:SREND_BHEADER_LEN))
      chad1(tlen1+1:tlen1+tlen2) = chad2(1:tlen2)
   end subroutine srend_CHAD_concat
! ********************************************************************


! ********************************************************************
! chad1 is an allocatable array, chad2 whatever
! chad1 is extended, as needed, to make room for chad2
   subroutine srend_CHAD_allocatable_concat(chad1, chad2)
      implicit NONE
      CHARACTER,dimension(:),allocatable,intent(INOUT) :: chad1
      CHARACTER,dimension(1:),optional,intent(IN)      :: chad2 ! data to examine
      integer(int32)                     :: tlen1,tlen2,nchad2
      character,dimension(:),allocatable :: temp_chad
      type(B_header_type) :: Bh
! ---- end variable declarations -------------------------------------
      Bh = transfer(chad2(1:SREND_BHEADER_LEN),Bh)
      tlen2 = Bh%tlen
      nchad2 = Bh%nchad
! if chad1 is not allocated, then allocate chad1 just large enough for chad2, then copy chad2 to chad1
      if( .NOT. allocated(chad1)) then
         allocate( chad1(1:tlen2))
         chad1(1:tlen2) = chad2(1:tlen2)
         return
      end if
! if chad1 is not a CHAD with leading 'CHAD'; can be forced with srend_CHAD_allocated_clear(chad1)
      Bh = transfer(chad1(1:SREND_BHEADER_LEN),Bh)
      if( Bh%leader /= 'CHAD') then ! chad1 is not an active CHAD
         if( sizeof(chad1) >= tlen2) then ! chad1 large enough, so opty in
            chad1(1:tlen2) = chad2(1:tlen2)
         else                             ! increase size of chad1 to fit chad2
            deallocate(chad1)
            allocate(chad1(1:tlen2))
            chad1(1:tlen2) = chad2(1:tlen2)
         end if
         return
      end if
! here there is a chad1
      tlen1 = Bh%tlen
      Bh%tlen = tlen1+tlen2           ! new len for chad1
      Bh%nchad = Bh%nchad + nchad2       ! new nchad for chad1
      if( sizeof(chad1) >= Bh%tlen) then
         chad1(1:SREND_BHEADER_LEN) = transfer(Bh,chad1(1:SREND_BHEADER_LEN)) ! copy new header in
         chad1(tlen1+1:tlen1+tlen2) = chad2(1:tlen2) ! tack on chad2
         return
      end if
!  sizeof(chad1) < tlen1 + tlen2, so extend chad1
      allocate(temp_chad(1:Bh%tlen))                     ! large enough temp_chad
      temp_chad(1:SREND_BHEADER_LEN) = transfer(Bh,temp_chad(1:SREND_BHEADER_LEN)) ! copy in header
      temp_chad(1:tlen1) = chad1(1:tlen1)             ! copy chad1 to temp_chad
      temp_chad(tlen1+1:tlen1+tlen2) = chad2(1:tlen2) ! tack chad2 after chad1 in temp
      deallocate(chad1)
      call move_alloc(temp_chad, chad1) ! moves temp_alloc to chad1 while creating chad1, deallocates temp_chad
   end subroutine srend_CHAD_allocatable_concat
! ********************************************************************

! ********************************************************************
   subroutine srend_CHAD_allocatable_clear(chad)
      implicit NONE
      CHARACTER,dimension(:),allocatable,intent(INOUT) :: chad
      type(B_header_type)     :: Bh
! ---- end variable declarations -------------------------------------
      if( .NOT. allocated(chad)) return
      Bh = transfer(chad(1:SREND_BHEADER_LEN),Bh)
      Bh%leader = 'blnk' ! clear
      chad(1:SREND_BHEADER_LEN) = transfer(Bh,chad(1:SREND_BHEADER_LEN))
   end subroutine srend_CHAD_allocatable_clear
! ********************************************************************

#ifdef SREND_MPI_SHORTCUTS
! ********************************************************************
! ******** MPI srend shortcuts ***************************************
! ********************************************************************
! These routines are not required for MPI or even as part of srend, 
! but they use srend to simplify certain kinds of MPI messaging 
! problems which commonly arise in applications.  Notably, each CHAD
! knows enough about itself to load data correctly to be used by the 
! receiving process.  These MPI shortcuts are enabled by defining
! SREND_MPI_SHORTCUTS then compiling with an MPI wrapper.

! uncomment one of these definitions for using srend_MPItest(MPIreq) 
! following srend_MPIisend(mem,MPItarg,MPItag,MPIreq)
!#define SREND_TEST_YIELDING call sleepqq(1) ! one millisecond, ifort only, requires: USE IFPORT
!#define SREND_TEST_YIELDING call sleep_u1() ! 1 microsecond, requires code below
!#define SREND_TEST_YIELDING call sleep_0()  ! release time slice, requires code below
#define SREND_TEST_YIELDING call sleep_n1() ! 1 nanosecond, requires code below

! The C code for sleep.c (lines between "#if 0" and "#endif")
#if 0
/* sleep.c   Ted Wetherbee  2014                     */
/* compile: gcc sleep.c -c || icc sleep.c            */
/* link with srend.o: sleep.o                        */
/*    sleep_u1: sleep for 1 micro second             */
/*    sleep_0 : yields time slice                    */
/*    sleep_n1: sleep for 1 nano second              */
#include <unistd.h>
#include <time.h>
int sleep_u1_     (){ usleep(1)            ; return 0;}
int sleep_0_      (){ sleep(0)             ; return 0;}
int sleep_n1_     (){ struct timespec tim,rem; 
                      tim.tv_sec = 0; tim.tv_nsec = 1; 
                      nanosleep(&tim, &rem); return 0;}
#endif
! --------------------------------------------------------------------

! these are global values for srend module, if MPI_SHORTCUTS defined
!   logical, save                    :: srend_MPI_INIT = .FALSE. 
!   integer, save                    :: srend_COMM               

! ********************************************************************      
   subroutine srend_comm_init() ! communicator for srend off MPI_COMM_WORLD
      include 'mpif.h'
#ifdef SREND_MPI_COMM_DUP
      integer :: MYer
#endif
      if(.NOT. srend_MPI_INIT) then
#ifdef SREND_MPI_COMM_DUP
         call MPI_comm_dup(MPI_COMM_WORLD, srend_COMM, MYer)
#else
         srend_COMM = MPI_COMM_WORLD
#endif
         srend_MPI_INIT = .TRUE.
      end if
   end subroutine srend_comm_init
! ********************************************************************

! ********************************************************************
! This is the ordinary MPI_send, length discovered from the CHAD
   subroutine srend_MPIsend(mem,MPItarg,MPItag)
      IMPLICIT NONE
      include 'mpif.h'
      CHARACTER,dimension(1:),intent(IN)   :: mem
      INTEGER,intent(IN)                   :: MPItarg, MPItag
      integer                   :: MYer
      type(B_header_type)       :: Bh
! --------------------------------------------------------------------
      Bh = transfer(mem(1:SREND_BHEADER_LEN),Bh)
      if( Bh%leader /='CHAD') then
         print *,"ERROR: not an srend CHAD"
         return
      end if
      if(.NOT. srend_MPI_INIT) call srend_comm_init()
      call MPI_send(mem,Bh%tlen,MPI_CHARACTER,MPItarg,MPItag,srend_COMM,MYer)
   end subroutine srend_MPIsend
! ********************************************************************


! ********************************************************************
! For non-blocking isend, an MPI tag is provided as an argument.
   subroutine srend_MPIisend(mem,MPItarg,MPItag,MPIreq)
      IMPLICIT NONE
      include 'mpif.h'
      CHARACTER,dimension(1:),intent(IN)   :: mem
      INTEGER,intent(IN)                   :: MPItarg, MPItag
      INTEGER,intent(INOUT)                :: MPIreq
      integer                   :: MYer
      type(B_header_type)       :: Bh
! --------------------------------------------------------------------
      Bh = transfer(mem(1:SREND_BHEADER_LEN),Bh)
      if( Bh%leader /='CHAD') then
         print *,"ERROR: not an srend CHAD"
         return
      end if
      if(.NOT. srend_MPI_INIT) call srend_comm_init()
      call MPI_isend(mem,Bh%tlen,MPI_CHARACTER,MPItarg,MPItag,srend_COMM,MPIreq,MYer)
   end subroutine srend_MPIisend
! ********************************************************************


! ********************************************************************
! This is called to complete an isend, and before the next isend
   subroutine srend_MPItest(MPIreq)
      !USE ifport ! only needed for ifort sleepqq(1)
      IMPLICIT NONE
      include 'mpif.h'
      INTEGER,intent(IN)                :: MPIreq
      integer                   :: MYer
      logical                   :: MPItestOK
! --------------------------------------------------------------------
      if(.NOT. srend_MPI_INIT) call srend_comm_init()
      MPItestOK = .FALSE.
      call MPI_test(MPIreq,MPItestOK,MPI_STATUS_IGNORE,MYer)
      do while(.NOT. MPItestOK)
        SREND_TEST_YIELDING
        call MPI_test(MPIreq,MPItestOK,MPI_STATUS_IGNORE,MYer)
      end do
   end subroutine srend_MPItest
! ********************************************************************

! ********************************************************************
! This receives a CHAD by MPItag and then loads it in the process
! srend "B" structure.
   subroutine srend_MPIrecv(MPItag)
      IMPLICIT NONE
      include 'mpif.h'
      INTEGER,intent(IN)                        :: MPItag                 
      character,dimension(:),allocatable   :: mem
      integer                              :: MYer
      integer                              :: MPIstat(MPI_STATUS_SIZE), MPIlen
! --------------------------------------------------------------------
      if(.NOT. srend_MPI_INIT) then
         call srend_comm_init()
      end if
      call MPI_probe(MPI_ANY_SOURCE, MPItag, srend_COMM, MPIstat, MYer)
      call MPI_get_count(MPIstat, MPI_CHARACTER, MPIlen,MYer)
      allocate( mem(1:MPIlen) )
      call MPI_recv(mem,MPIlen,MPI_CHARACTER,MPIstat(MPI_SOURCE),MPItag, &
                     srend_COMM, MPI_STATUS_IGNORE,MYer)
      call srend_compose_load(mem)
   end subroutine srend_MPIrecv
! ********************************************************************

! ****** end: srend MPI shortcuts ************************************
#endif




end module srend
! ********************************************************************
! ********************************************************************
! *********** end: module srend   ************************************
! ********************************************************************
! ********************************************************************




